# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""
The Abstract Location to Mask class.
This provides and interface to make all machines look the same.

CONSTRAINT - NO NODE OF ANY TYPE SHALL HAVE THE SAME NAME IN THE SAME GRAPH.
CONSTRAINT - MACHINE NAMES CANNOT HAVE A "-" IN THE NAME.
CONSTRAINT - BITMASKS MUST NOT BE MODIFIED, CREATED ONCE.
"""
import logging
import abc
import datetime
import enum
import hashlib
import re
from abc import ABCMeta
from collections import deque, defaultdict, OrderedDict
from enum import Enum
from functools import reduce
from typing import Optional, Callable, List
import networkx as nx
import numpy as np
import pandas as pd
from Octeres.SL.bitmaskgraphfilter import BitmaskGraphFilterContext
from Octeres.bitmask import EnhancedBitmaskLike, EnhancedBitmask
from Octeres.bitmask.bitmask_graph import validate_graph
from Octeres.bitmask.bitmask_graph_search import (
    is_bitmask_indexable,
    apply_bitmask_to_nxgraph,
    breadth_first_bitmask_search,
    bitmask_func_intersection_any,
    SearchStats,
    breadth_first_search,
)
from Octeres.data_generation import generate_names
from Octeres.util import superprint
from Octeres.bitmask import BitmaskSerializer

BitS = BitmaskSerializer()
log = logging.getLogger(__name__)


def get_pk_hash(name, ts, te):
    ts = pd.to_datetime(ts).to_pydatetime().isoformat()
    te = pd.to_datetime(te).to_pydatetime().isoformat()
    # don't need to include the class because ts+machine_name is enough
    pk_str = f"{ts}_{te}_{name}"
    hexdig = hashlib.shake_128(pk_str.encode("utf")).hexdigest(7)
    return int(hexdig, 16)


def convert_bitmask_to_hexmask(bitmask):
    hexmask = BitS.encode_bitmask_to_hex(bitmask)
    return hexmask


class AbstractLocationtoMaskError(Exception):
    pass


class Location_Error(Exception):
    pass


def get_bitmask(graph, root_name, location: str, debug=False) -> Optional[EnhancedBitmaskLike]:
    if location == root_name:
        root_dct = graph.nodes[root_name]
        bitmask = root_dct["bitmask"]
        if debug:
            superprint(f"searching for {location} => {bitmask}")
    elif location == "":
        bitmask = None
        if debug:
            superprint(f"searching for {location} => {bitmask}")
    else:
        bitmask = None
        for node_name, child_name in nx.bfs_edges(graph, source=root_name):
            if debug:
                superprint(f"searching edge {node_name}->{child_name}, evaluating child.")
            if location == child_name:
                child_dct = graph.nodes[child_name]
                bitmask = child_dct["bitmask"]
                break
        if debug:
            superprint(f"searching for {location} => {bitmask}")
    return bitmask


def get_bitmask_from_locations(
    ltm_obj,
    location_type,
    locations,
    bitmask_class: EnhancedBitmaskLike = EnhancedBitmask,
    debug=False,
):
    if locations in ["*", ["*"]]:
        bitmask = bitmask_class.ones(ltm_obj.bit_total)
    elif location_type == "hexmask":  # probably should use the map, but this is hardcoded.
        hexmask = locations
        bitmask = ltm_obj.BitS.decode_hex_to_bitmask(hexmask, limit=ltm_obj.bit_total)
    elif location_type == "bitmask":
        bitmask = locations
    elif locations is None:
        bitmask = ltm_obj.base_mask.copy()
    elif type(locations) == list:
        location_lst = locations
        root_name = ltm_obj.name
        graph = ltm_obj.graph
        # search stats vars
        bitmasks = []
        for location in location_lst:
            bitmask = get_bitmask(graph, root_name, location, debug=debug)
            if bitmask is None:
                raise Location_Error(f"Location {location} cannot be found for ltm " f"{ltm_obj.name}/{ltm_obj.get_pk()}.")
            else:
                bitmasks.append(bitmask)
        if len(bitmasks) == 1:
            bitmask = bitmasks[0].copy()  # we don't want to give away our node's bitmask.
        elif len(bitmasks) == 0:
            bitmask = ltm_obj.base_mask.copy()
        else:
            bitmask = bitmask_class.join_bitmasks(bitmasks, bitmask_class.logical_or)
            bitmask.flags.writeable = False  # immutable!
    else:
        raise Exception(f"Invalid location_type({location_type}) and locations({locations}).")
    return bitmask


def get_bitmask_from_locations_graph(
    graph,
    root_name,
    location_lst,
    bitmask_class: EnhancedBitmaskLike = EnhancedBitmask,
    debug=False,
):
    bitmasks = []
    for location in location_lst:
        bitmask = get_bitmask(graph, root_name, location, debug=debug)
        if bitmask is None:
            raise Location_Error(f"Location {location} cannot be found for graph {root_name}.")
        else:
            bitmasks.append(bitmask)
    if len(bitmasks) == 1:
        bitmask = bitmasks[0].copy()  # we don't want to give away our node's bitmask.
    else:
        bitmask = bitmask_class.join_bitmasks(bitmasks, bitmask_class.logical_or)
        bitmask.flags.writeable = False  # immutable!
    return bitmask


def get_ghost_nxgraph(root_name, root_obj_type, nodes, get_parents, get_obj_type, get_extra) -> nx.DiGraph:
    """
    built a nx graph using get_obj_type and get_parents
    get_extra must at least return {'is_schedulable': 1}
    """
    lookup_obj_type = {}
    lookup_parents = {}
    graph = nx.DiGraph()
    # pq = deque()
    pq = set()
    graph.add_node(root_name, obj_type=root_obj_type, **get_extra(root_name, root_name))
    for node_name in nodes:
        if node_name in pq:
            pq.remove(node_name)  # this is slow for some reason.  Slightly faster if we do the if check rather than try.
        obj_type = get_obj_type(root_name, node_name)
        lookup_obj_type[node_name] = obj_type
        graph.add_node(node_name, obj_type=obj_type, **get_extra(root_name, node_name))
        parent_names = get_parents(root_name, node_name)
        lookup_parents[node_name] = parent_names
        for parent_name in parent_names:
            graph.add_edge(parent_name, node_name)  # this will create a node even if it's not used.
            pq.add(parent_name)  # this is for a set.
            # if parent_name not in pq:
            #     pq.append(parent_name)
    # test to make sure order doesn't matter.
    # pq_lst = list(pq)
    # random.shuffle(pq_lst)
    # pq = set(pq_lst)
    while pq:
        node_name = pq.pop()
        try:
            obj_type = lookup_obj_type[node_name]
        except KeyError:
            obj_type = get_obj_type(root_name, node_name)
            lookup_obj_type[node_name] = obj_type
        graph.add_node(node_name, obj_type=obj_type, **get_extra(root_name, node_name))

        try:
            parent_names = lookup_parents[node_name]
        except KeyError:
            parent_names = get_parents(root_name, node_name)
            lookup_parents[node_name] = parent_names

        for parent_name in parent_names:
            # if parent_name not in graph.nodes:
            if not graph.has_node(parent_name):
                pq.add(parent_name)
                # if parent_name not in pq:
                #     pq.append(parent_name)
            graph.add_edge(parent_name, node_name)
    return graph


class Abstract_Location_to_Mask(metaclass=ABCMeta):
    """
    The abstract class for converting locations to masks and back.
    """

    # if you are going to make a ABC of this class, don't fill out machine_name.
    machine_name: Optional[str]
    name: Optional[str]
    obj_types: list  # usable object types for a given LTM.  Order may matter for sorting.
    graph: nx.DiGraph = None
    bit_total: int
    pk: int = None
    ppk: int = None
    range_start: datetime.datetime = None
    range_end: datetime.datetime = None
    skip_nodes = ()
    graph_only = False
    graph_sort_neighbors = None
    alias = None
    notes = None

    def __init_ltm__(self, ltm: dict):
        assert "location_to_mask_map" in ltm
        ltmb_dct = ltm["location_to_mask_base"]
        self.location_to_mask_base_id = ltmb_dct.get("location_to_mask_base_id", None)
        self.pk = self.location_to_mask_base_id
        self.ppk = ltmb_dct.get("previous_location_to_mask_base_id", None)
        try:
            self.name = ltmb_dct["name"]
        except KeyError:
            self.name = ltmb_dct["machine_name"]
        self.is_schedulable = ltmb_dct["is_schedulable"]
        self.obj_type = ltmb_dct["obj_type"]
        self.machine_name = ltmb_dct["machine_name"]
        self.range_start = ltmb_dct["range_start"]
        self.range_end = ltmb_dct["range_end"]
        self.bit_total = ltmb_dct["bit_total"]
        self.unit_a_type = ltmb_dct["unit_a_type"]
        self.unit_b_type = ltmb_dct["unit_b_type"]
        self.unit_c_type = ltmb_dct["unit_c_type"]
        self.unit_d_type = ltmb_dct["unit_d_type"]

    def __init__(
        self,
        ltm: dict = None,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        """
        given a ltm dictionary, create a ltm object.
        """
        self.BitS = BitS
        assert self.bit_total > 0, "LTM cannot have a bit_total < 1"
        self.bitmask_class = bitmask_class
        if bitmask_class is not None:
            self.base_mask = bitmask_class.zeros(self.bit_total)
            self.base_mask.flags.writeable = False
        else:
            self.base_mask = None
        if not hasattr(self, "name"):
            self.name = self.machine_name
        if ltm:
            self.__init_ltm__(ltm)
        self.graph = self.generate_bittree(bitmask_class=bitmask_class, debug=debug)
        # A fix for when one is defined and not the other.  Eventually switch to name.
        if self.name and not self.machine_name:
            self.machine_name = self.name
        elif self.machine_name and not self.name:
            self.name = self.machine_name

    def set_init(self, pk, range_start, range_end, validate=False):
        pk_a = int(self.pk) if self.pk is not None else self.pk
        pk_b = int(pk) if pk is not None else pk
        if validate:
            if pk_a is not None and pk_a != pk_b:
                raise Exception(f"pk mismatch on {self} {pk_a} <- {pk_b}")
            if self.range_start is not None and self.range_start != range_start:
                raise Exception(f"range_start mismatch on {self} {self.range_start} <- {range_start}")
        # if self.range_end is not None and self.range_end != range_end and range_end != pddt('2038-01-01'):
        #     raise Exception(f"range_end mismatch on {self} {self.range_end} <- {range_end}")
        self.pk = pk
        self.range_start = range_start
        self.range_end = range_end

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        self.class_reset()  # if we change to __init__ this can go.
        if hasattr(self, "names"):
            if hasattr(self, "names_extra"):
                extra_dct = self.names_extra
            else:
                extra_dct = {}
            nodes, parents_dct = generate_names(
                self.names,
                reverse=False,
                slice_range=True,
                skip_nodes=self.skip_nodes,
                extra_dct=extra_dct,
            )
            set_nodes = set(nodes)
            if len(nodes) > len(set_nodes):
                # FIXME: log a warning or something.  If you get the self.names wrong, the values can balloon to a very large list.
                nodes = list(set_nodes)
            self.parents_dct = parents_dct
            # This is odd, because self.get_parents may use the graph, from self.graph, we cannot have it exists or it can
            # be wrong.  We also want it to use the graph when it can after the graph is created.
            self.graph = None
            ghost_graph = get_ghost_nxgraph(
                self.machine_name,
                "root",
                nodes,
                self.get_parents,
                self.get_obj_type,
                self.get_extra,
            )
            if bitmask_class is not None:
                func_is_bitmask_indexable = (
                    is_bitmask_indexable if not hasattr(self, "is_bitmask_indexable") else self.is_bitmask_indexable
                )
                graph = apply_bitmask_to_nxgraph(
                    self.machine_name,
                    self.bit_total,
                    ghost_graph,
                    func_is_bitmask_indexable,
                    sort_neighbors=self.graph_sort_neighbors,
                    debug=debug,
                    bitmask_class=bitmask_class,
                )
                assert graph.nodes[self.machine_name]['bitmask'].length == self.bit_total
            else:
                graph = ghost_graph
        else:
            raise Exception(f"{self} cannot make bittree")
        if validate:
            validate_graph(graph)
        return graph

    def get_pk(self):
        return int(self.pk) if self.pk is not None else self.pk

    def set_pk(self, pk):
        """set primary key.  This is important when working with the database."""
        if self.pk and self.pk != pk:
            log.warning(f"{self} setting pk, doesn't match {self.pk} <- {pk}")
        self.pk = pk

    def get_ppk(self):
        return self.ppk

    def set_ppk(self, ppk):
        """set previous primary key.  This is important when working with the database."""
        if self.ppk and self.ppk != ppk:
            log.warning(f"{self} setting ppk, doesn't match {self.ppk} <- {ppk}")
        self.ppk = ppk

    def __repr__(self, *args, **kwargs):
        return f"<Location_to_Mask({self.name} {self.range_start}->{self.range_end})[{self.get_pk()}]>"

    @classmethod
    @abc.abstractmethod
    def get_parents(cls, machine_name: str, node_name: str) -> list:
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_obj_type(cls, machine_name: str, node_name: str) -> str:
        raise NotImplementedError()

    def get_is_schedulable(self, machine_name: str, node_name: str) -> int:  # 0 or 1
        if node_name == machine_name:
            result = 0
        else:
            result = 1
        return result

    def get_extra(self, machine_name: str, node_name: str) -> dict:
        return {"is_schedulable": self.get_is_schedulable(machine_name, node_name)}

    def get_locations_from_bitmask(self, bitmask, obj_type, debug=False):
        machine_name = self.name
        graph = self.graph
        bitmask_any = bitmask
        bitmask_subset = bitmask
        fc = BitmaskGraphFilterContext(
            self.name,
            self.graph,
            bitmask_any,
            None,
            bitmask_subset,
            None,
            obj_type,
            None,
        )
        ag = None
        node_name_lst = breadth_first_bitmask_search(
            machine_name,
            graph,
            bitmask,
            bitmask_func_intersection_any,
            fc.filter_func_subset_obj_type,
            animated_graph=ag,
            debug=debug,
        )
        return node_name_lst

    def convert_bitmask_to(self, bitmask, obj_type=None, debug=False) -> dict:
        """given a hexmask, convert it to all the possible locations that are a subset of the bitmask
        returns
        {
            location_type: list[unit_value],
        }
        """
        location_dct = defaultdict(list)
        bitmask_any = None
        bitmask_all = None
        bitmask_subset = bitmask
        fc = BitmaskGraphFilterContext(
            self.name,
            self.graph,
            bitmask_any,
            bitmask_all,
            bitmask_subset,
            None,
            obj_type,
            None,
        )
        node_lst = breadth_first_bitmask_search(
            self.name,
            self.graph,
            bitmask,
            bitmask_func_intersection_any,
            fc.filter_func_and_filter_node_names,
            debug=debug,
        )
        for node in node_lst:
            obj_type = self.get_obj_type(self.name, node)
            location_dct[obj_type].append(node)
        location_dct = dict(location_dct)
        return location_dct

    def convert_hexmask_to(self, hexmask, debug=False) -> dict:
        """given a hexmask, convert it to all the possible locations that are a subset of the hexmask
        returns
        {
            location_type: list[unit_value],
        }
        """
        bitmask = self.BitS.decode_hex_to_bitmask(hexmask, limit=self.bit_total)
        return self.convert_bitmask_to(bitmask, debug=debug)

    def get_dict(self):
        dct = dict(
            name=self.name,
            machine_name=self.name,
            pk=self.pk,
            ppk=self.ppk,
            range_start=self.range_start,
            range_end=self.range_end,
            bit_total=self.bit_total,
        )
        return dct

    @classmethod
    def class_reset(cls):
        pass

    @classmethod
    # @abc.abstractmethod
    def generate_tree(cls, level_idx_to_node_names_dct):
        raise NotImplementedError()

    def linkable(self, other):
        """given another ltm, compare the two and see if they can be linked or put one after the other in time."""
        error_lst = []
        if self.name != other.name:
            error_lst.append(f"ltm name({self.name} -> {other.name}) must match to link")
        if other.range_start and other.range_start < self.range_start:
            # in this case it would move the range start of the current ltm earlier and may cross others.
            error_lst.append(
                f"The current {self.range_start=} cannot be greater than the new {other.range_start}. "
                f"This may allow a merge and cause and overlap."
            )
        if other.range_start and self.range_end != datetime.datetime(2038, 1, 1) and self.range_end > other.range_start:
            # 2038 is important because it's a default for ending an ltm, and it can be lowered.
            error_lst.append(
                f"The current {self.range_end=} cannot be greater than the new {other.range_start}. "
                f"This may allow a merge and cause and overlap."
            )
        return error_lst

    # # TODO
    # def difference(self, other):
    #     """ will replace ltm_difference"""
    #     pass
    #
    # # TODO
    # def mergeable(self, other):
    #     """ will replace ltm_update_or_mergeable"""
    #     pass


class Obj_Type_BGQ(Enum):
    ROOT = "root"
    NODE = "node"
    NODEBOARD = "nodeboard"
    NODEBOARDS = "nodeboards"
    MIDPLANE = "midplane"


class ABC_Location_to_Mask_Graph(Abstract_Location_to_Mask, metaclass=abc.ABCMeta):

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if self.graph:
            nodes = list(self.graph.predecessors(node_name))
        else:
            nodes = super().get_parents(machine_name, node_name)
        return nodes

    def get_children(self, machine_name: str, node_name: str) -> list:
        if self.graph:
            nodes = list(self.graph.successors(node_name))
        else:
            raise NotImplementedError()
        return nodes

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if self.graph:
            obj_type = self.graph.nodes[node_name]["obj_type"]
        else:
            obj_type = super().get_obj_type(machine_name, node_name)
        return obj_type

    def is_bitmask_indexable(self, machine_name, node_name) -> bool:
        return is_bitmask_indexable(machine_name, node_name)

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        return super().generate_bittree(validate=validate, bitmask_class=bitmask_class, debug=debug)


class ABC_Location_to_Mask_BGQ(ABC_Location_to_Mask_Graph, metaclass=abc.ABCMeta):
    obj_types = [
        Obj_Type_BGQ.ROOT.value,
        Obj_Type_BGQ.MIDPLANE.value,
        Obj_Type_BGQ.NODEBOARDS.value,
        Obj_Type_BGQ.NODEBOARD.value,
    ]
    obj_type_regex_lookup = OrderedDict(
        [
            (
                Obj_Type_BGQ.MIDPLANE.value,
                re.compile("^R(?P<row>[0-9a-fA-F])(?P<column>[0-9a-fA-F])-M(?P<midplane>[0-9a-fA-F])$"),
            ),
            (
                Obj_Type_BGQ.NODEBOARD.value,
                re.compile("^R(?P<row>[0-9a-fA-F])(?P<column>[0-9a-fA-F])-M(?P<midplane>[0-9a-fA-F])-N(?P<nodeboard>[0-9]{2})$"),
            ),
            (
                Obj_Type_BGQ.NODEBOARDS.value,
                re.compile(
                    "^(R(?P<row>[0-9a-fA-F])(?P<column>[0-9a-fA-F])-M(?P<midplane>[0-9a-fA-F])-N(?P<nodeboard>[0-9]{2}),{0,1})+$"
                ),
            ),
        ]
    )
    name_to_bitmask_dct = {}

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if node_name == machine_name:
            return []
        bitmask = self.name_to_bitmask_dct[node_name]
        # the largest blocks parents is the machine_name.
        if bitmask.all():
            return [machine_name]
        parents_and_size = []
        parents_sizes_dct = defaultdict(list)
        # FIXME, this can be done better...
        for other_node_name, other_bitmask in self.name_to_bitmask_dct.items():
            #  let's not find ourselves.          let's not find nodes that look exactly like me.
            # if node_name != other_node_name and other_node_name != machine_name and not (bitmask == other_bitmask).all():
            # trutha = not (bitmask == other_bitmask).all()
            # truthb = not bitmask.equal_all(other_bitmask)
            # assert trutha == truthb
            if (
                node_name != other_node_name and other_node_name != machine_name and bitmask.bit_count != other_bitmask.bit_count
            ):  # and not bitmask.equal_all(other_bitmask):
                truth1 = self.bitmask_class.logical_is_subset(other_bitmask, bitmask)
                if truth1:
                    # we have too many parents...
                    parents_sizes_dct[other_bitmask.bit_count].append(other_node_name)
                    parents_and_size.append(other_bitmask.bit_count)
        smallest_size = reduce(lambda left, right: left if left <= right else right, parents_and_size)
        parents = parents_sizes_dct[smallest_size]
        return parents

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if machine_name == node_name:
            obj_type = Obj_Type_BGQ.ROOT.value
        else:
            obj_type = None
            for key, regex in self.obj_type_regex_lookup.items():
                match = regex.match(node_name)
                if match:
                    obj_type = key
                    break
            if obj_type is None:
                obj_type = "cobalt_location"
        return obj_type

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        if bitmask_class != EnhancedBitmask:
            name_to_bitmask_dct = {}
            for key, value in self.name_to_bitmask_dct.items():
                idx_lst = EnhancedBitmask.get_idx_lst_from_bitmask(value)
                name_to_bitmask_dct[key] = bitmask_class.get_bitmask_from_idx_lst(value.length, idx_lst)
        else:
            name_to_bitmask_dct = self.name_to_bitmask_dct
        graph = graph_from_name_to_bitmask_dct(self, name_to_bitmask_dct)
        return graph


def _add_ghost_node_recurse(_graph: nx.DiGraph, _machine_name, _node_name, get_parents, get_obj_type):
    # superprint(f"adding ghost node: {_node_name}")
    parents = get_parents(_machine_name, _node_name)
    obj_type = get_obj_type(_machine_name, _node_name)
    _graph.add_node(_node_name, obj_type=obj_type)
    for parent_name in parents:
        if parent_name not in _graph.nodes:
            _add_ghost_node_recurse(_graph, _machine_name, parent_name, get_parents, get_obj_type)
        _graph.add_edge(parent_name, _node_name)


def add_ghost_node(graph: nx.DiGraph, machine_name, node_name, get_parents, get_obj_type) -> None:
    # recursively go up to the parents and make sure you have them all created to the root.
    # on the tail we must add the edges and only the tail because
    _add_ghost_node_recurse(graph, machine_name, node_name, get_parents, get_obj_type)


def graph_from_name_to_bitmask_dct(ltm_obj, name_to_bitmask_dct):
    machine_name = ltm_obj.machine_name
    base_mask = name_to_bitmask_dct[machine_name]
    graph: nx.DiGraph = nx.DiGraph()
    graph.add_node(machine_name, obj_type="root", bitmask=base_mask, level_idx=0)
    for node, bitmask in name_to_bitmask_dct.items():
        add_ghost_node(graph, machine_name, node, ltm_obj.get_parents, ltm_obj.get_obj_type)
    for node_name, node in graph.nodes(data=True):
        bitmask = name_to_bitmask_dct[node_name]
        if "bitmask" not in node:
            bitmask.flags.writeable = False
            node["bitmask"] = bitmask
        node["obj_type"] = ltm_obj.get_obj_type(machine_name, node_name)
        node["is_schedulable"] = ltm_obj.get_is_schedulable(machine_name, node_name)
        if hasattr(bitmask, "bit_count"):
            if bitmask.bit_count == 1:
                # int = (array([0]),) - [0] -> array([0]) - [0] -> int64 -> int()
                node["idx"] = bitmask.bit_idx_first  # int(np.where(bitmask == 1)[0][0])
        else:
            if sum([i for i in bitmask if i == 1]) == 1:
                node["idx"] = int(np.where(bitmask == 1)[0][0])
    return graph


def match_nid(node_name):
    result = False
    try:
        int(node_name)
    except ValueError:
        if ("-" in node_name) or ("," in node_name):
            try:
                int(node_name.replace("-", "").replace(",", ""))
            except ValueError:
                pass
            else:
                result = True
    else:
        result = True
    return result


def split_cray_nids(nidstr):
    """given a cray provided nid list, split it into a integer list, uncompressed"""
    nidlst = []
    nidgroups = nidstr.split(",")
    for nidgroup in nidgroups:
        if nidgroup.find("-") != -1:
            a, b = nidgroup.split("-", 1)
            a = int(a)
            b = int(b)
            if a > b:
                tmp = b
                b = a
                a = tmp
            b = b + 1  # need one more for inclusive
        else:
            a = int(nidgroup)
            b = a + 1
        for nid in range(a, b):
            nidlst.append(nid)
    nidlst = [str(nid) for nid in sorted(list(set(nidlst)))]
    return nidlst


def join_cray_nids(nid_str_lst):
    """given a nid_int_lst, return a nid_str"""
    if nid_str_lst == []:
        nid_str = ""
    else:
        nid_int_lst = list(map(int, nid_str_lst))
        nid_int_lst = sorted(nid_int_lst)

        def _join_group(lr):
            lr = filter(lambda i: i is not None, lr)
            return "-".join(map(str, lr))

        nid_groupings = []
        # handle the head
        last = nid_int_lst[0]
        left_right = [last, None]
        for nid_int in nid_int_lst[1:]:
            # if it equals, we have order
            if (last + 1) == nid_int:
                # don't append it keep going
                left_right[1] = nid_int
                last = nid_int
            else:
                # not incremented by one
                nid_groupings.append(_join_group(left_right))
                last = nid_int
                left_right = [last, None]
        # handle the tail
        nid_groupings.append(_join_group(left_right))
        nid_str = ",".join(nid_groupings)
    return nid_str


class Obj_Type_Cray(Enum):
    ROOT = "root"
    HOSTNAME = "hostname"
    IP = "ip"
    NID = "nid"
    CNAME = "cname"
    CABINET = "cabinet"
    CHASSIS = "chassis"
    SLOT = "slot"
    BLADE = "blade"
    ROW = "row"
    RACK = "rack"


class NameGeneratorCrayXC40:
    def __init__(self, max_idx, initial_idx=0):
        self.initial_idx = initial_idx
        self.hostname_bitmask_idx = initial_idx
        self.hostname_bitmask_idx_reverse = max_idx
        self.nid_bitmask_idx = initial_idx
        self.nid_bitmask_idx_reverse = max_idx
        self.hostname_lookup = {}
        self.nid_lookup = {}
        self.max_idx = max_idx

    def reset(self):
        max_idx = self.max_idx
        self.hostname_bitmask_idx = self.initial_idx
        self.hostname_bitmask_idx_reverse = max_idx
        self.nid_bitmask_idx = self.initial_idx
        self.nid_bitmask_idx_reverse = max_idx
        self.hostname_lookup = {}
        self.nid_lookup = {}
        self.max_idx = max_idx

    def generate_hostname(self, name_dct, generate_parent=False):
        row = name_dct["row"]
        cabinet = name_dct["cabinet"]
        chassis = name_dct["chassis"]
        slot = name_dct["slot"]
        cname = name_dct["cname"]
        key = f"{row}__{cabinet}__{chassis}__{slot}__{cname}"
        try:
            hostname = self.hostname_lookup[key]
        except KeyError:
            hostname = f"nid{self.hostname_bitmask_idx:0>5}"
            self.hostname_lookup[key] = hostname
        # this can be called twice, if it is, the second time needs to not increment
        if not generate_parent:
            assert self.hostname_bitmask_idx < (self.max_idx + self.initial_idx)
            # if self.hostname_bitmask_idx >= self.max_idx:
            #     print(f"eek! {hostname} {self.hostname_bitmask_idx}, {self.max_idx}")
            self.hostname_bitmask_idx += 1
        return hostname

    def generate_nid(self, name_dct, generate_parent=False):
        row = name_dct["row"]
        cabinet = name_dct["cabinet"]
        chassis = name_dct["chassis"]
        slot = name_dct["slot"]
        cname = name_dct["cname"]
        key = f"{row}__{cabinet}__{chassis}__{slot}__{cname}"
        try:
            nid = self.nid_lookup[key]
        except KeyError:
            nid = f"{self.nid_bitmask_idx}"
            self.nid_lookup[key] = nid
        if not generate_parent:
            assert self.nid_bitmask_idx < (self.max_idx + self.initial_idx)
            # if self.nid_bitmask_idx >= self.max_idx:
            #     print(f"eek! {nid} {self.nid_bitmask_idx}, {self.max_idx}")
            self.nid_bitmask_idx += 1
        return nid

    def generate_nid_reverse(self, name_dct, generate_parent=False):
        if not generate_parent:
            assert self.nid_bitmask_idx_reverse >= self.initial_idx
            self.nid_bitmask_idx_reverse -= 1
        row = name_dct["row"]
        cabinet = name_dct["cabinet"]
        chassis = name_dct["chassis"]
        slot = name_dct["slot"]
        cname = name_dct["cname"]
        key = f"{row}__{cabinet}__{chassis}__{slot}__{cname}"
        try:
            nid = self.nid_lookup[key]
        except KeyError:
            nid = f"{self.nid_bitmask_idx_reverse}"
            self.nid_lookup[key] = nid
        return nid

    def generate_hostname_reverse(self, name_dct, generate_parent=False):
        if not generate_parent:
            assert self.hostname_bitmask_idx_reverse >= self.initial_idx
            self.hostname_bitmask_idx_reverse -= 1
        row = name_dct["row"]
        cabinet = name_dct["cabinet"]
        chassis = name_dct["chassis"]
        slot = name_dct["slot"]
        cname = name_dct["cname"]
        key = f"{row}__{cabinet}__{chassis}__{slot}__{cname}"
        try:
            hostname = self.hostname_lookup[key]
        except KeyError:
            hostname = f"nid{self.hostname_bitmask_idx_reverse:0>5}"
            self.hostname_lookup[key] = hostname
        return hostname


class ABC_Location_to_Mask_Cray(ABC_Location_to_Mask_Graph, metaclass=abc.ABCMeta):
    machine_name: Optional[str]
    obj_types = [
        Obj_Type_Cray.ROOT.value,
        Obj_Type_Cray.CNAME.value,
        Obj_Type_Cray.HOSTNAME.value,
        Obj_Type_Cray.NID.value,
    ]
    # 'c', CABINET, '-', `ROW`, 'c', CHASSIS, 's', SLOT, 'n', NODE
    re_row = re.compile("^row(?P<row>[0-9]+)$")
    re_cabinet = re.compile("^c(?P<cabinet>[0-9]+)-(?P<row>[0-9]+)$")
    re_chassis = re.compile("^c(?P<cabinet>[0-9]+)-(?P<row>[0-9]+)c(?P<chassis>[0-9]+)$")
    re_slot = re.compile("^c(?P<cabinet>[0-9]+)-(?P<row>[0-9]+)c(?P<chassis>[0-9]+)s(?P<slot>[0-9]+)$")
    re_cname = re.compile("^c(?P<cabinet>[0-9]+)-(?P<row>[0-9]+)c(?P<chassis>[0-9]+)s(?P<slot>[0-9]+)n(?P<node>[0-9])$")
    re_hostname = re.compile("^nid[0-9]{5}")
    # re_nid = re.compile("^[0-9][,-0-9]*$")
    re_ip = re.compile("^[0-9]+.[0-9]+.[0-9]+.[0-9]+$")

    hostname_to_cname_dct = {}
    skip_nodes = []

    @classmethod
    def hostname_to_cname(cls, node_name):
        if cls.hostname_to_cname_dct:
            gen_cname = cls.hostname_to_cname_dct[node_name]
        else:
            hostname_to_cname_dct = {}
            cname_count = 4
            slot_count = 16
            chassis_count = 3
            cabinet_count = 12
            row_count = 3
            bitmask_idx = 0
            for row in range(row_count):
                for cabinet in range(cabinet_count):
                    for chassis in range(chassis_count):
                        for slot in range(slot_count):
                            for node in range(cname_count):
                                # gen_nid = f"nid{bitmask_idx:0>5}"
                                # gen_cname = f"c{cabinet}-{row}c{chassis}s{slot}n{node}"
                                nid_int: int = bitmask_idx
                                # nid_str = str(nid_int)
                                hostname = f"nid{nid_int:0>5}"
                                cname = f"c{cabinet}-{row}c{chassis}s{slot}n{node}"
                                # ip = f"10.{chassis}.{slot}.{node}"
                                hostname_to_cname_dct[hostname] = cname
                                bitmask_idx += 1
            cls.hostname_to_cname_dct = hostname_to_cname_dct
            gen_cname = cls.hostname_to_cname_dct[node_name]
        return gen_cname

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if self.graph:
            parents = super().get_parents(machine_name, node_name)
        elif machine_name == node_name:
            parents = []
        else:
            try:
                # Only exists if generate_names is called and the output is saved on the cls.
                parents = self.parents_dct[node_name]
            except (KeyError, AttributeError):
                obj_type = self.get_obj_type(machine_name, node_name)
                if obj_type == Obj_Type_Cray.NID.value:
                    parents = [f"nid{node_name:0>5}"]
                elif obj_type == Obj_Type_Cray.HOSTNAME.value:
                    cname = self.hostname_to_cname(node_name)
                    if cname:
                        parents = [cname]
                    else:
                        raise Exception(f"Unknown node_name to parents, {node_name=} {machine_name=}")
                elif obj_type == Obj_Type_Cray.CNAME.value:
                    # parents = [machine_name]
                    match = re.match(self.re_cname, node_name)
                    match_dct = match.groupdict()
                    row = match_dct["row"]
                    cabinet = match_dct["cabinet"]
                    chassis = match_dct["chassis"]
                    slot = match_dct["slot"]
                    parents = [f"c{cabinet}-{row}c{chassis}s{slot}"]
                elif obj_type == Obj_Type_Cray.SLOT.value:
                    match = re.match(self.re_slot, node_name)
                    match_dct = match.groupdict()
                    row = match_dct["row"]
                    cabinet = match_dct["cabinet"]
                    chassis = match_dct["chassis"]
                    parents = [f"c{cabinet}-{row}c{chassis}"]
                elif obj_type == Obj_Type_Cray.CHASSIS.value:
                    match = re.match(self.re_chassis, node_name)
                    match_dct = match.groupdict()
                    row = match_dct["row"]
                    cabinet = match_dct["cabinet"]
                    parents = [f"c{cabinet}-{row}"]
                elif obj_type == Obj_Type_Cray.CABINET.value:
                    match = re.match(self.re_cabinet, node_name)
                    match_dct = match.groupdict()
                    row = match_dct["row"]
                    parents = [f"row{row}"]
                elif obj_type == Obj_Type_Cray.ROW.value:
                    parents = [machine_name]
                else:
                    raise Exception(f"Unknown node_name to parents, {node_name=} {machine_name=} {obj_type=}")
        return parents

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if self.graph:
            obj_type = super().get_obj_type(machine_name, node_name)
        elif machine_name == node_name:
            obj_type = Obj_Type_Cray.ROOT.value
        else:
            if re.match(self.re_hostname, node_name) is not None:
                obj_type = Obj_Type_Cray.HOSTNAME.value
            elif match_nid(node_name):  # re.match(self.re_nid, node_name) is not None:
                obj_type = Obj_Type_Cray.NID.value
            elif re.match(self.re_ip, node_name) is not None:
                obj_type = Obj_Type_Cray.IP.value
            elif re.match(self.re_cname, node_name) is not None:
                obj_type = Obj_Type_Cray.CNAME.value
            elif re.match(self.re_slot, node_name) is not None:
                obj_type = Obj_Type_Cray.SLOT.value
            elif re.match(self.re_chassis, node_name) is not None:
                obj_type = Obj_Type_Cray.CHASSIS.value
            elif re.match(self.re_row, node_name) is not None:
                obj_type = Obj_Type_Cray.ROW.value
            elif re.match(self.re_cabinet, node_name) is not None:
                obj_type = Obj_Type_Cray.CABINET.value
            else:
                raise Exception(f"Unknown obj_type, {node_name=}, {machine_name=}")
        return obj_type


LOCATION_TO_MASK_BASE = "location_to_mask_base"
LOCATION_TO_MASK_BASE_ID = "location_to_mask_base_id"
PREVIOUS_LOCATION_TO_MASK_BASE_ID = "previous_location_to_mask_base_id"
LOCATION_TO_MASK_MAP = "location_to_mask_map"
LOCATION_TO_MASK_BRANCH = "location_to_mask_branch"
LOCATION_TO_MASK_EDGE = "location_to_mask_edge"


def get_ltmb_dct(
    name=None,
    machine_name: str = None,
    location_to_mask_base_id=None,
    previous_location_to_mask_base_id=None,
    range_start=None,
    range_end=None,
    bit_total=0,
    unit_a_type="unit_a",
    unit_b_type="unit_b",
    unit_c_type="unit_c",
    unit_d_type="unit_d",
    inserted_timestamp=None,  # not used
    updated_timestamp=None,  # not used
    is_schedulable=0,
    obj_type="",
    pk=None,  # not used, but need to be in place to catch them.
    pk_column=None,  # not used, but need to be in place to catch them.
    ppk=None,  # not used, but need to be in place to catch them.
    ppk_column=None,  # not used, but need to be in place to catch them.
):
    dct = dict(
        name=name if name else (machine_name if machine_name else None),
        machine_name=machine_name,
        location_to_mask_base_id=location_to_mask_base_id,
        previous_location_to_mask_base_id=(previous_location_to_mask_base_id if previous_location_to_mask_base_id else None),
        range_start=range_start,
        range_end=range_end,
        bit_total=bit_total,
        unit_a_type=unit_a_type,
        unit_b_type=unit_b_type,
        unit_c_type=unit_c_type,
        unit_d_type=unit_d_type,
        is_schedulable=is_schedulable,
        obj_type=obj_type,
    )
    return dct


def get_ltmm_dct(
    name: Optional[str] = None,
    idx: int = 0,
    bit_total: int = 0,
    unit_a: Optional[str] = None,
    unit_b: Optional[str] = None,
    unit_c: Optional[str] = None,
    unit_d: Optional[str] = None,
    unit_classification: Optional[str] = None,
    is_schedulable=0,
    obj_type="",
    location_to_mask_map_id=None,
    location_to_mask_base_id=None,
):
    dct = dict(
        name=name,
        idx=idx,
        bit_total=bit_total,
        unit_a=str(unit_a),
        unit_b=str(unit_b),
        unit_c=str(unit_c),
        unit_d=str(unit_d),
        unit_classification=unit_classification,
        is_schedulable=is_schedulable,
        obj_type=obj_type,
        location_to_mask_map_id=location_to_mask_map_id,
        location_to_mask_base_id=location_to_mask_base_id,
    )
    return dct


def get_ltmbr_dct(
    name=None,
    hexmask=None,
    bitmask=None,
    bit_total=0,
    is_schedulable=0,
    obj_type="",
    location_to_mask_branch_id=None,
    location_to_mask_base_id=None,
):
    new_hexmask: Optional[str]
    new_bitmask: Optional[EnhancedBitmaskLike]
    new_bit_total: int
    # we somehow have a bitmask and hexmask, lets check them
    if hexmask is not None and bitmask is not None:
        new_bitmask = BitS.decode_hex_to_bitmask(hexmask, limit=bitmask.length)
        if not (new_bitmask == bitmask).all():
            raise Exception("cannot set a hexmask and bitmask together")
        else:
            new_hexmask = hexmask
            new_bit_total = bitmask.length
    else:
        # we have a hexmask, no bitmask
        if hexmask is not None and bitmask is None:
            new_bitmask = BitS.decode_hex_to_bitmask(hexmask, limit=bit_total)
            new_hexmask = hexmask
            new_bit_total = new_bitmask.length
        # we have a bitmask, no hexmask
        elif hexmask is None and bitmask is not None:
            new_hexmask = BitS.encode_bitmask_to_hex(bitmask)
            new_bitmask = bitmask
            new_bit_total = bitmask.length
        # we don't have either
        # elif hexmask is None and bitmask is None:
        else:
            new_hexmask = None
            new_bitmask = None
            new_bit_total = bit_total
            if bit_total:
                raise Exception("bit_total cannot be set without a bitmask or hexmask")
    dct = dict(
        name=name,
        bit_total=new_bit_total,
        is_schedulable=is_schedulable,
        obj_type=obj_type,
        location_to_mask_branch_id=location_to_mask_branch_id,
        location_to_mask_base_id=location_to_mask_base_id,
        bitmask=new_bitmask,
        hexmask=new_hexmask,
    )
    if bit_total:
        assert dct["bit_total"] == bit_total
    return dct


def get_ltme_dct(
    parent: str = None,
    child: str = None,
    location_to_mask_edge_id=None,
    location_to_mask_base_id=None,
):
    dct = dict(
        parent=parent,
        child=child,
        location_to_mask_edge_id=location_to_mask_edge_id,
        location_to_mask_base_id=location_to_mask_base_id,
    )
    return dct


def get_ltm_from_graph(
    graph: nx.DiGraph,
    bit_total,
    machine_name: str,
    range_start: Optional[datetime.datetime],
    range_end: datetime.datetime,
    get_parents,
    get_obj_type,
    get_is_schedulable,
    bitmask_class: EnhancedBitmaskLike = EnhancedBitmask,
    debug=False,
):
    stat = SearchStats()
    ltmc = LTMCreator(
        graph,
        machine_name,
        range_start,
        range_end,
        bit_total,
        get_parents,
        get_obj_type,
        get_is_schedulable,
        bitmask_class,
    )
    breadth_first_search(
        machine_name,
        graph,
        stat,
        ltmc.search_func,
        tail_func=ltmc.tail_func,
        stop_when_found=False,
        debug=debug,
    )
    ltm_dct = {
        LOCATION_TO_MASK_BASE: ltmc.ltmb,
        LOCATION_TO_MASK_BRANCH: ltmc.ltmbr_lst,
        LOCATION_TO_MASK_MAP: ltmc.ltmm_lst,
        LOCATION_TO_MASK_EDGE: ltmc.ltme_lst,
    }
    return ltm_dct


class LTMCreator:
    def __init__(
        self,
        graph: nx.DiGraph,
        machine_name: str,
        range_start: Optional[datetime.datetime],
        range_end: datetime.datetime,
        bit_total: int,
        get_parents: Callable[[str, str], List],
        get_obj_type: Callable[[str, str], str],
        get_is_schedulable: Callable[[str, str], int],
        bitmask_class: type[EnhancedBitmaskLike],
    ):
        self.graph = graph
        self.name = machine_name
        self.machine_name = machine_name
        self.range_start: Optional[datetime.datetime] = range_start
        self.range_end: datetime.datetime = range_end
        self.bit_total = bit_total
        self.ltmb = {}
        self.ltmbr_lst = []
        self.ltmm_lst = []
        self.ltme_lst = []
        self._node_lookup = {}
        self.get_parents = get_parents
        self.get_obj_type = get_obj_type
        self.get_is_schedulable = get_is_schedulable
        self.bitmask_class = bitmask_class

    def search_func(
        self,
        parent_name: Optional[str],
        node_name: str,
        results: List[str],
        debug=False,
    ) -> bool:
        node_parents = self.graph.in_edges(node_name)
        node_children = self.graph.edges(node_name)
        node_dct = self.graph.nodes[node_name]
        bitmask = node_dct["bitmask"]
        if node_name not in self._node_lookup:
            obj_type = self.get_obj_type(self.machine_name, node_name)
            # assert obj_type == node_dct['obj_type']
            if not node_parents:  # bitmask.bit_count == bitmask.length:
                # bitmask is not used..
                dct = get_ltmb_dct(
                    name=node_name,
                    machine_name=self.machine_name,
                    range_start=self.range_start,
                    range_end=self.range_end,
                    previous_location_to_mask_base_id=None,
                    location_to_mask_base_id=None,
                    bit_total=self.bit_total,
                    unit_a_type="a",
                    unit_b_type="b",
                    unit_c_type="c",
                    unit_d_type="d",
                    is_schedulable=self.get_is_schedulable(self.machine_name, node_name),
                    obj_type=obj_type,
                )
                del dct["location_to_mask_base_id"]
                del dct["previous_location_to_mask_base_id"]
                self.ltmb = dct
            elif node_parents and node_children or bitmask.bit_count > 1:
                dct = get_ltmbr_dct(
                    name=node_name,
                    bitmask=node_dct["bitmask"],
                    bit_total=self.bit_total,
                    is_schedulable=self.get_is_schedulable(self.machine_name, node_name),
                    obj_type=obj_type,
                )
                self.ltmbr_lst.append(dct)
            else:  # bitmask.bit_count == 1 # not node.children:
                idx_lst = self.bitmask_class.get_idx_lst_from_bitmask(node_dct["bitmask"])
                if len(idx_lst) == 1:
                    idx = idx_lst.pop()
                    dct = get_ltmm_dct(
                        name=node_name,
                        idx=idx,
                        bit_total=self.bit_total,
                        unit_a="",
                        unit_b="",
                        unit_c="",
                        unit_d="",
                        unit_classification="compute",
                        is_schedulable=self.get_is_schedulable(self.machine_name, node_name),
                        obj_type=obj_type,
                    )
                    self.ltmm_lst.append(dct)
                elif len(idx_lst) == 0:
                    # skipped
                    # slot c0-0c0s3 on theta has no children because they are service nodes.
                    # skip it.
                    pass
                else:
                    raise Exception(f"A leaf {node_name} cannot have more than one bit set.")

            self._node_lookup[node_name] = node_dct
        else:
            pass  # don't create two of the same dct from the same node

        if parent_name is not None:
            dct_edge = get_ltme_dct(parent_name, node_name)
            self.ltme_lst.append(dct_edge)
        return False

    def tail_func(self, node_name: str, results: List[str], debug=False):
        pass
