# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import re
from typing import Optional
from pandas import to_datetime as pddt
from Octeres.SL.location_to_mask_graph import XLocation_to_Mask_None
from Octeres.SL.abstract_location_to_mask import ABC_Location_to_Mask_Graph
from Octeres.SL.obj_type import Obj_Type
from Octeres.data_generation import NameGeneratorIDX

_, MACHINE_NAME = __name__.rsplit("ltm_", 1)


class Location_to_Mask(ABC_Location_to_Mask_Graph):
    """very close to thetagpu, get_parents and get_obj_type are slightly different, along with the regexes."""

    machine_name: Optional[str]
    re_node = re.compile(r"^salad(?P<node>[0-9]{2})$")
    re_gpu = re.compile(r"^salad(?P<node>[0-9]{2})-gpu(?P<gpu>[0-9])$")
    re_slot = re.compile(r"^n(?P<node>[0-9]{2})g(?P<gpu>[0-9])s(?P<slot>[0-9]+)$")
    re_idx = re.compile(r"^\d+$")

    def get_is_schedulable(self, machine_name: str, node_name: str) -> int:  # 0 or 1
        if machine_name == node_name:
            result = 0
        else:
            try:
                int(node_name)
            except ValueError:
                result = 1
            else:
                # _idx is not schedulable.
                result = 0
        return result

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if machine_name == node_name:
            parents = []
        else:
            try:
                parents = self.parents_dct[node_name]
            except KeyError:
                if match := re.match(self.re_slot, node_name):
                    match_dct = match.groupdict()
                    node = match_dct["node"]
                    gpu = match_dct["gpu"]
                    parents = [f"{machine_name}{node}-gpu{gpu}"]
                elif re.match(self.re_gpu, node_name):
                    head, tail = node_name.rsplit("-", 1)
                    parents = [head]
                elif re.match(self.re_node, node_name):
                    parents = [machine_name]
                else:
                    raise Exception(f"Unknown node_name to parents, {node_name=} {machine_name=}")
        return parents

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if machine_name == node_name:
            obj_type = Obj_Type.ROOT.value
        else:
            if re.match(self.re_slot, node_name) is not None:
                obj_type = Obj_Type.SLOT.value
            elif re.match(self.re_gpu, node_name) is not None:
                obj_type = Obj_Type.GPU.value
            elif re.match(self.re_node, node_name) is not None:
                obj_type = Obj_Type.NODE.value
            else:
                raise Exception(f"Unknown obj_type, {node_name=}, {machine_name=}")
        return obj_type


class Location_to_Mask_Node(Location_to_Mask):
    machine_name = MACHINE_NAME
    bit_total = 24
    obj_types = [Obj_Type.ROOT.value, Obj_Type.NODE.value]
    ng = NameGeneratorIDX(24)
    names = [
        (
            (Obj_Type.NODE.value, (1, 25, 1), "salad{node:0>2}"),
            (Obj_Type.NODE_IDX.value, (0, 1, 1), ng.generate_idx),
        ),
    ]

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        try:
            obj_type = super().get_obj_type(machine_name, node_name)
        except Exception:
            if re.match(self.re_idx, node_name) is not None:
                obj_type = Obj_Type.NODE_IDX.value
            else:
                raise Exception(f"Unknown obj_type, {node_name=}, {machine_name=}")
        return obj_type

    @classmethod
    def class_reset(cls):
        cls.ng.reset()


class Location_to_Mask_GPU(Location_to_Mask):
    machine_name = MACHINE_NAME
    bit_total = 192
    ng = NameGeneratorIDX(192)
    obj_types = [Obj_Type.ROOT.value, Obj_Type.NODE.value, Obj_Type.GPU.value]
    names = [
        (
            (Obj_Type.NODE.value, (1, 25, 1), "salad{node:0>2}"),
            (Obj_Type.GPU.value, (0, 8, 1), "salad{node:0>2}-gpu{gpu}"),
            (Obj_Type.GPU_IDX.value, (0, 1, 1), ng.generate_idx),
        ),
    ]

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        try:
            obj_type = super().get_obj_type(machine_name, node_name)
        except Exception:
            if re.match(self.re_idx, node_name) is not None:
                obj_type = Obj_Type.GPU_IDX.value
            else:
                raise Exception(f"Unknown obj_type, {node_name=}, {machine_name=}")
        return obj_type

    @classmethod
    def class_reset(cls):
        cls.ng.reset()


class Location_to_Mask_Slot(Location_to_Mask):
    machine_name = MACHINE_NAME
    bit_total = 1344
    obj_types = [
        Obj_Type.ROOT.value,
        Obj_Type.NODE.value,
        Obj_Type.GPU.value,
        Obj_Type.SLOT.value,
    ]
    ng = NameGeneratorIDX(1344)
    names = [
        (
            (Obj_Type.NODE.value, (1, 25, 1), "salad{node:0>2}"),
            (Obj_Type.GPU.value, (0, 8, 1), "salad{node:0>2}-gpu{gpu}"),
            (Obj_Type.SLOT.value, (0, 7, 1), "n{node:0>2}g{gpu}s{slot}"),
            (Obj_Type.SLOT_IDX.value, (0, 1, 1), ng.generate_idx),
        ),
    ]

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        try:
            obj_type = super().get_obj_type(machine_name, node_name)
        except Exception:
            if re.match(self.re_idx, node_name) is not None:
                obj_type = Obj_Type.SLOT_IDX.value
            else:
                raise Exception(f"Unknown obj_type, {node_name=}, {machine_name=}")
        return obj_type

    @classmethod
    def class_reset(cls):
        cls.ng.reset()


ltm_lst = [
    {
        "ts": pddt("1970-01-01"),
        "te": pddt("2020-05-29"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2020-05-29"),
        "te": pddt("2020-07-20"),
        "ltm_class": Location_to_Mask_Node,
        "pk": 744,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2020-07-20"),
        "te": pddt("2020-07-21"),
        "ltm_class": Location_to_Mask_GPU,
        "pk": 745,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2020-07-21"),
        "te": pddt("2038-01-01"),
        "ltm_class": Location_to_Mask_Slot,
        "pk": 746,
        "name": MACHINE_NAME,
    },
]
