# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import re
from enum import Enum
from pandas import to_datetime as pddt

from Octeres.bitmask import eb, EnhancedBitmaskLike, EnhancedBitmask
from Octeres.SL.location_to_mask_graph import (
    XLocation_to_Mask_None,
    XLocation_to_Mask_DoesNotExist,
)
from Octeres.SL.abstract_location_to_mask import (
    get_pk_hash,
    graph_from_name_to_bitmask_dct,
    ABC_Location_to_Mask_Graph,
)
from Octeres.SL.abstract_location_to_mask import (
    NameGeneratorCrayXC40,
    ABC_Location_to_Mask_Cray,
    Obj_Type_Cray,
)

_, MACHINE_NAME = __name__.rsplit("ltm_", 1)


class Location_to_Mask_Cray_192(ABC_Location_to_Mask_Cray):
    machine_name = MACHINE_NAME
    bit_total = 192
    obj_types = [
        Obj_Type_Cray.ROOT.value,
        Obj_Type_Cray.CNAME.value,
        Obj_Type_Cray.HOSTNAME.value,
        Obj_Type_Cray.NID.value,
    ]
    ng = NameGeneratorCrayXC40(bit_total)
    names = [
        (
            (Obj_Type_Cray.ROW.value, (0, 1, 1), "row{row}"),
            (Obj_Type_Cray.CABINET.value, (0, 1, 1), "c{cabinet}-{row}"),
            (Obj_Type_Cray.CHASSIS.value, (0, 1, 1), "c{cabinet}-{row}c{chassis}"),
            (Obj_Type_Cray.SLOT.value, (0, 48, 1), "c{cabinet}-{row}c{chassis}s{slot}"),
            (
                Obj_Type_Cray.CNAME.value,
                (0, 4, 1),
                "c{cabinet}-{row}c{chassis}s{slot}n{cname}",
            ),
            (Obj_Type_Cray.HOSTNAME.value, (0, 1, 1), ng.generate_hostname),
            (Obj_Type_Cray.NID.value, (0, 1, 1), ng.generate_nid),
        ),
    ]

    @classmethod
    def class_reset(cls):
        cls.ng.reset()


class Location_to_Mask_Cray_200(ABC_Location_to_Mask_Cray):
    machine_name = MACHINE_NAME
    bit_total = 200
    obj_types = [
        Obj_Type_Cray.ROOT.value,
        Obj_Type_Cray.SLOT.value,
        Obj_Type_Cray.CNAME.value,
    ]
    names = (
        (Obj_Type_Cray.SLOT.value, (0, 50, 1), "c0-0c0s{slot}"),
        (Obj_Type_Cray.CNAME.value, (0, 4, 1), "c0-0c0s{slot}n{cname}"),
    )


class Location_to_Mask_2(ABC_Location_to_Mask_Cray):
    machine_name = MACHINE_NAME
    bit_total = 2
    obj_types = [
        Obj_Type_Cray.ROOT.value,
        Obj_Type_Cray.CNAME.value,
        Obj_Type_Cray.HOSTNAME.value,
        Obj_Type_Cray.NID.value,
    ]
    ng = NameGeneratorCrayXC40(bit_total)
    names = [
        (
            (Obj_Type_Cray.ROW.value, (0, 1, 1), "row{row}"),
            (Obj_Type_Cray.CABINET.value, (0, 1, 1), "c{cabinet}-{row}"),
            (Obj_Type_Cray.CHASSIS.value, (0, 1, 1), "c{cabinet}-{row}c{chassis}"),
            (Obj_Type_Cray.SLOT.value, (0, 1, 1), "c{cabinet}-{row}c{chassis}s{slot}"),
            (
                Obj_Type_Cray.CNAME.value,
                (0, 2, 1),
                "c{cabinet}-{row}c{chassis}s{slot}n{cname}",
            ),
            (Obj_Type_Cray.HOSTNAME.value, (0, 1, 1), ng.generate_hostname),
            (Obj_Type_Cray.NID.value, (0, 1, 1), ng.generate_nid),
        )
    ]

    @classmethod
    def class_reset(cls):
        cls.ng.reset()


class Location_to_Mask_4_inc(ABC_Location_to_Mask_Cray):
    machine_name = MACHINE_NAME
    bit_total = 4
    obj_types = [
        Obj_Type_Cray.ROOT.value,
        Obj_Type_Cray.CNAME.value,
        Obj_Type_Cray.HOSTNAME.value,
        Obj_Type_Cray.NID.value,
    ]
    ng = NameGeneratorCrayXC40(bit_total)
    names = [
        (
            (Obj_Type_Cray.ROW.value, (0, 1, 1), "row{row}"),
            (Obj_Type_Cray.CABINET.value, (0, 1, 1), "c{cabinet}-{row}"),
            (Obj_Type_Cray.CHASSIS.value, (0, 1, 1), "c{cabinet}-{row}c{chassis}"),
            (Obj_Type_Cray.SLOT.value, (0, 1, 1), "c{cabinet}-{row}c{chassis}s{slot}"),
            (
                Obj_Type_Cray.CNAME.value,
                (0, 4, 1),
                "c{cabinet}-{row}c{chassis}s{slot}n{cname}",
            ),
            (Obj_Type_Cray.HOSTNAME.value, (0, 1, 1), ng.generate_hostname),
            (Obj_Type_Cray.NID.value, (0, 1, 1), ng.generate_nid),
        )
    ]
    hostname_to_cname_dct = {
        "nid00000": "c0-0c0s0n0",
        "nid00001": "c0-0c0s0n1",
        "nid00002": "c0-0c0s0n2",
        "nid00003": "c0-0c0s0n3",
    }

    @classmethod
    def class_reset(cls):
        cls.ng.reset()

    @classmethod
    def hostname_to_cname(cls, node_name):
        gen_cname = cls.hostname_to_cname_dct[node_name]
        return gen_cname


class Location_to_Mask_4_dec(ABC_Location_to_Mask_Cray):
    machine_name = MACHINE_NAME
    bit_total = 4
    obj_types = [
        Obj_Type_Cray.ROOT.value,
        Obj_Type_Cray.CNAME.value,
        Obj_Type_Cray.HOSTNAME.value,
        Obj_Type_Cray.NID.value,
    ]
    ng = NameGeneratorCrayXC40(bit_total)
    names = [
        (
            (Obj_Type_Cray.ROW.value, (0, 1, 1), "row{row}"),
            (Obj_Type_Cray.CABINET.value, (0, 1, 1), "c{cabinet}-{row}"),
            (Obj_Type_Cray.CHASSIS.value, (0, 1, 1), "c{cabinet}-{row}c{chassis}"),
            (Obj_Type_Cray.SLOT.value, (0, 1, 1), "c{cabinet}-{row}c{chassis}s{slot}"),
            (
                Obj_Type_Cray.CNAME.value,
                (0, 4, 1),
                "c{cabinet}-{row}c{chassis}s{slot}n{cname}",
            ),
            (Obj_Type_Cray.HOSTNAME.value, (0, 1, 1), ng.generate_hostname_reverse),
            (Obj_Type_Cray.NID.value, (0, 1, 1), ng.generate_nid_reverse),
        )
    ]
    hostname_to_cname_dct = {
        "nid00000": "c0-0c0s0n3",
        "nid00001": "c0-0c0s0n2",
        "nid00002": "c0-0c0s0n1",
        "nid00003": "c0-0c0s0n0",
    }

    @classmethod
    def class_reset(cls):
        cls.ng.reset()


class Location_to_Mask_8(ABC_Location_to_Mask_Cray):
    machine_name = MACHINE_NAME
    bit_total = 8
    obj_types = [
        Obj_Type_Cray.ROOT.value,
        Obj_Type_Cray.CNAME.value,
        Obj_Type_Cray.HOSTNAME.value,
        Obj_Type_Cray.NID.value,
    ]
    ng = NameGeneratorCrayXC40(bit_total)
    names = [
        (
            (Obj_Type_Cray.ROW.value, (0, 1, 1), "row{row}"),
            (Obj_Type_Cray.CABINET.value, (0, 1, 1), "c{cabinet}-{row}"),
            (Obj_Type_Cray.CHASSIS.value, (0, 1, 1), "c{cabinet}-{row}c{chassis}"),
            (Obj_Type_Cray.SLOT.value, (0, 2, 1), "c{cabinet}-{row}c{chassis}s{slot}"),
            (
                Obj_Type_Cray.CNAME.value,
                (0, 4, 1),
                "c{cabinet}-{row}c{chassis}s{slot}n{cname}",
            ),
            (Obj_Type_Cray.HOSTNAME.value, (0, 1, 1), ng.generate_hostname),
            (Obj_Type_Cray.NID.value, (0, 1, 1), ng.generate_nid),
        )
    ]

    @classmethod
    def class_reset(cls):
        cls.ng.reset()


class Location_to_Mask_8_flat(ABC_Location_to_Mask_Cray):
    machine_name = MACHINE_NAME
    bit_total = 8
    obj_types = [
        Obj_Type_Cray.ROOT.value,
        Obj_Type_Cray.CNAME.value,
        Obj_Type_Cray.NID.value,
        Obj_Type_Cray.HOSTNAME.value,
    ]
    name_to_bitmask_dct = {
        "0": eb.array([1, 0, 0, 0, 0, 0, 0, 0]),
        "1": eb.array([0, 1, 0, 0, 0, 0, 0, 0]),
        "2": eb.array([0, 0, 1, 0, 0, 0, 0, 0]),
        "3": eb.array([0, 0, 0, 1, 0, 0, 0, 0]),
        "4": eb.array([0, 0, 0, 0, 1, 0, 0, 0]),
        "5": eb.array([0, 0, 0, 0, 0, 1, 0, 0]),
        "6": eb.array([0, 0, 0, 0, 0, 0, 1, 0]),
        "7": eb.array([0, 0, 0, 0, 0, 0, 0, 1]),
        "c0-0": eb.array([1, 1, 1, 1, 1, 1, 1, 1]),
        "c0-0c0": eb.array([1, 1, 1, 1, 1, 1, 1, 1]),
        "c0-0c0s0": eb.array([1, 1, 1, 1, 0, 0, 0, 0]),
        "c0-0c0s0n0": eb.array([1, 0, 0, 0, 0, 0, 0, 0]),
        "c0-0c0s0n1": eb.array([0, 1, 0, 0, 0, 0, 0, 0]),
        "c0-0c0s0n2": eb.array([0, 0, 1, 0, 0, 0, 0, 0]),
        "c0-0c0s0n3": eb.array([0, 0, 0, 1, 0, 0, 0, 0]),
        "c0-0c0s1": eb.array([0, 0, 0, 0, 1, 1, 1, 1]),
        "c0-0c0s1n0": eb.array([0, 0, 0, 0, 1, 0, 0, 0]),
        "c0-0c0s1n1": eb.array([0, 0, 0, 0, 0, 1, 0, 0]),
        "c0-0c0s1n2": eb.array([0, 0, 0, 0, 0, 0, 1, 0]),
        "c0-0c0s1n3": eb.array([0, 0, 0, 0, 0, 0, 0, 1]),
        "nid00000": eb.array([1, 0, 0, 0, 0, 0, 0, 0]),
        "nid00001": eb.array([0, 1, 0, 0, 0, 0, 0, 0]),
        "nid00002": eb.array([0, 0, 1, 0, 0, 0, 0, 0]),
        "nid00003": eb.array([0, 0, 0, 1, 0, 0, 0, 0]),
        "nid00004": eb.array([0, 0, 0, 0, 1, 0, 0, 0]),
        "nid00005": eb.array([0, 0, 0, 0, 0, 1, 0, 0]),
        "nid00006": eb.array([0, 0, 0, 0, 0, 0, 1, 0]),
        "nid00007": eb.array([0, 0, 0, 0, 0, 0, 0, 1]),
        "row0": eb.array([1, 1, 1, 1, 1, 1, 1, 1]),
        "test": eb.array([1, 1, 1, 1, 1, 1, 1, 1]),
    }

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        if bitmask_class != EnhancedBitmask:
            name_to_bitmask_dct = {}
            for key, value in self.name_to_bitmask_dct.items():
                idx_lst = EnhancedBitmask.get_idx_lst_from_bitmask(value)
                name_to_bitmask_dct[key] = bitmask_class.get_bitmask_from_idx_lst(value.length, idx_lst)
        else:
            name_to_bitmask_dct = self.name_to_bitmask_dct
        graph = graph_from_name_to_bitmask_dct(self, name_to_bitmask_dct)
        return graph


class Location_to_Mask_16(ABC_Location_to_Mask_Cray):
    machine_name = MACHINE_NAME
    bit_total = 16
    obj_types = [
        Obj_Type_Cray.ROOT.value,
        Obj_Type_Cray.CNAME.value,
        Obj_Type_Cray.NID.value,
        Obj_Type_Cray.HOSTNAME.value,
    ]
    ng = NameGeneratorCrayXC40(bit_total)
    names = [
        (
            (Obj_Type_Cray.ROW.value, (0, 1, 1), "row{row}"),
            (Obj_Type_Cray.CABINET.value, (0, 1, 1), "c{cabinet}-{row}"),
            (Obj_Type_Cray.CHASSIS.value, (0, 1, 1), "c{cabinet}-{row}c{chassis}"),
            (Obj_Type_Cray.SLOT.value, (0, 4, 1), "c{cabinet}-{row}c{chassis}s{slot}"),
            (
                Obj_Type_Cray.CNAME.value,
                (0, 4, 1),
                "c{cabinet}-{row}c{chassis}s{slot}n{cname}",
            ),
            (Obj_Type_Cray.HOSTNAME.value, (0, 1, 1), ng.generate_hostname),
            (Obj_Type_Cray.NID.value, (0, 1, 1), ng.generate_nid),
        ),
    ]

    @classmethod
    def class_reset(cls):
        cls.ng.reset()


class Obj_Type_3Level(Enum):
    ROOT = "root"
    LEVEL_3 = "level_3"
    LEVEL_2 = "level_2"
    LEVEL_1 = "level_1"


class Location_to_Mask_3Level(ABC_Location_to_Mask_Graph):
    machine_name = MACHINE_NAME
    obj_types = [
        Obj_Type_3Level.ROOT.value,
        Obj_Type_3Level.LEVEL_1.value,
        Obj_Type_3Level.LEVEL_2.value,
        Obj_Type_3Level.LEVEL_3.value,
    ]
    names = [
        (
            (Obj_Type_3Level.LEVEL_1.value, (0, 2, 1), "test{level_1:0>2}"),
            (Obj_Type_3Level.LEVEL_2.value, (0, 7, 1), "test{level_1:0>2}-{level_2}"),
            (
                Obj_Type_3Level.LEVEL_3.value,
                (0, 8, 1),
                "x{level_1}y{level_2}z{level_3}",
            ),
        ),
    ]
    bit_total = 112
    pattern_l2 = re.compile(f"{machine_name}(?P<l1>[0-9]+)-(?P<l2>[0-9]+)")
    pattern_l3 = re.compile("x(?P<l1>[0-9]+)y(?P<l2>[0-9]+)z(?P<l3>[0-9]+)")

    def get_is_schedulable(self, machine_name: str, node_name: str) -> int:  # 0 or 1
        obj_type = self.get_obj_type(machine_name, node_name)
        if obj_type in [
            Obj_Type_3Level.LEVEL_1.value,
            Obj_Type_3Level.LEVEL_2.value,
            Obj_Type_3Level.LEVEL_3.value,
        ]:
            result = 1
        else:
            result = 0
        return result

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if machine_name == node_name:
            parents = []
        else:
            obj_type = self.get_obj_type(machine_name, node_name)
            if obj_type == Obj_Type_3Level.LEVEL_3.value:
                match = re.match(self.pattern_l3, node_name)
                if match:
                    match_dct = match.groupdict()
                    level_1_idx = int(match_dct["l1"])
                    level_2_idx = int(match_dct["l2"])
                    parents = [f"{machine_name}{level_1_idx:02d}-{level_2_idx}"]
                else:
                    raise Exception(f"Unknown node_name to parents, {node_name=} {machine_name=}")
            elif obj_type == Obj_Type_3Level.LEVEL_2.value:
                match = re.match(self.pattern_l2, node_name)
                if match:
                    match_dct = match.groupdict()
                    level_1_idx = int(match_dct["l1"])
                    parents = [f"{machine_name}{level_1_idx:02d}"]
                else:
                    raise Exception(f"Unknown node_name to parents, {node_name=} {machine_name=}")
            else:
                parents = [machine_name]
        return parents

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        # this ltm has odd names
        # fixme: this should be tested against the real data found for the test.
        if machine_name == node_name:
            obj_type = Obj_Type_3Level.ROOT.value
        else:
            if "z" in node_name:
                obj_type = Obj_Type_3Level.LEVEL_3.value
            elif "-" in node_name:
                obj_type = Obj_Type_3Level.LEVEL_2.value
            else:
                obj_type = Obj_Type_3Level.LEVEL_1.value
        return obj_type


ltm_lst = [
    {
        "ts": pddt("1970-01-01"),
        "te": pddt("2011-07-01"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2011-07-01 00:00:00"),
        "te": pddt("2011-07-01 00:02:00"),
        "ltm_class": Location_to_Mask_Cray_192,
        "name": MACHINE_NAME,
        "alias": "dup_a",
    },
    {
        "ts": pddt("2011-07-01 00:02:00"),
        "te": pddt("2011-07-01 00:04:00"),
        "ltm_class": Location_to_Mask_Cray_192,
        "name": MACHINE_NAME,
        "alias": "dup_b",
    },
    {
        "ts": pddt("2011-07-01 00:04:00"),
        "te": pddt("2011-07-01 00:06:00"),
        "ltm_class": Location_to_Mask_Cray_192,
        "name": MACHINE_NAME,
        "alias": "dup_c",
    },
    {
        "ts": pddt("2011-07-01 00:06:00"),
        "te": pddt("2011-07-01 00:08:00"),
        "ltm_class": Location_to_Mask_Cray_192,
        "name": MACHINE_NAME,
        "alias": "dup_d",
    },
    {
        "ts": pddt("2011-07-01 00:08:00"),
        "te": pddt("2011-07-01 00:09:00"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2011-07-01 00:09:00"),
        "te": pddt("2011-07-01 00:10:00"),
        "ltm_class": Location_to_Mask_Cray_192,
        "name": MACHINE_NAME,
        "alias": "dup_e",
    },
    {
        "ts": pddt("2011-07-01 00:10:00"),
        "te": pddt("2011-07-01 00:12:00"),
        "ltm_class": Location_to_Mask_Cray_192,
        "name": MACHINE_NAME,
        "alias": "dup_f",
    },
    {
        "ts": pddt("2011-07-01 00:12:00"),
        "te": pddt("2011-07-01 00:14:00"),
        "ltm_class": Location_to_Mask_Cray_200,
        "name": MACHINE_NAME,
        "alias": "dup_g",
    },
    {
        "ts": pddt("2011-07-01 00:14:00"),
        "te": pddt("2012-01-01"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2012-01-01"),
        "te": pddt("2012-01-04"),
        "ltm_class": Location_to_Mask_16,
        "name": MACHINE_NAME,
        "alias": "future_16a",
        "bit_total": 16,
    },
    {
        "ts": pddt("2012-01-04"),
        "te": pddt("2012-01-08"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "future_8a",
        "bit_total": 8,
    },
    {
        "ts": pddt("2012-01-08"),
        "te": pddt("2012-01-12"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "future_8b",
        "bit_total": 8,
    },  # might need to combine these.
    {
        "ts": pddt("2012-01-12"),
        "te": pddt("2012-01-16"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "future_8c",
        "bit_total": 8,
    },
    {
        "ts": pddt("2012-01-16"),
        "te": pddt("2012-01-20"),
        "ltm_class": Location_to_Mask_16,
        "name": MACHINE_NAME,
        "alias": "future_16b",
        "bit_total": 16,
    },
    # this window is necessary for tests inserting events into the future of the above ltm into the below time range.
    {
        "ts": pddt("2012-01-20"),
        "te": pddt("2012-02-01"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2012-02-01"),
        "te": pddt("2012-02-04"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "sl_8a",
        "bit_total": 8,
    },
    {
        "ts": pddt("2012-02-04"),
        "te": pddt("2012-02-08"),
        "ltm_class": Location_to_Mask_4_inc,
        "name": MACHINE_NAME,
        "alias": "sl_4a",
        "bit_total": 4,
    },
    {
        "ts": pddt("2012-02-08"),
        "te": pddt("2012-02-12"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "sl_8b",
        "bit_total": 8,
    },
    {
        "ts": pddt("2012-02-12"),
        "te": pddt("2012-02-16"),
        "ltm_class": Location_to_Mask_2,
        "name": MACHINE_NAME,
        "alias": "sl_2a",
        "bit_total": 2,
    },
    {
        "ts": pddt("2012-02-16"),
        "te": pddt("2012-02-20"),
        "ltm_class": Location_to_Mask_4_inc,
        "name": MACHINE_NAME,
        "alias": "sl_4b",
        "bit_total": 4,
    },
    {
        "ts": pddt("2012-02-20"),
        "te": pddt("2012-02-24"),
        "ltm_class": Location_to_Mask_4_dec,
        "name": MACHINE_NAME,
        "alias": "sl_4c",
        "bit_total": 4,
    },
    {
        "ts": pddt("2012-02-24"),
        "te": pddt("2012-02-28"),
        "ltm_class": Location_to_Mask_4_inc,
        "name": MACHINE_NAME,
        "alias": "sl_4d",
        "bit_total": 4,
    },
    {
        "ts": pddt("2012-02-28"),
        "te": pddt("2012-03-03"),
        "ltm_class": Location_to_Mask_4_inc,
        "name": MACHINE_NAME,
        "alias": "sl_4e",
        "bit_total": 4,
    },
    {
        "ts": pddt("2012-03-03"),
        "te": pddt("2012-03-07"),
        "ltm_class": Location_to_Mask_2,
        "name": MACHINE_NAME,
        "alias": "sl_2b",
        "bit_total": 2,
    },
    # gap
    {
        "ts": pddt("2012-03-07"),
        "te": pddt("2015-01-01"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2015-01-01"),
        "te": pddt("2015-01-03"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_a",
        "notes": "A outer left touching",
    },
    {
        "ts": pddt("2015-01-03"),
        "te": pddt("2015-01-06"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_b",
        "notes": "B exact left, right",
    },
    {
        "ts": pddt("2015-01-06"),
        "te": pddt("2015-01-08"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_c",
        "notes": "C outer right touching",
    },
    # these collide with other ltms above
    {
        "ts": pddt("2015-01-04"),
        "te": pddt("2015-01-05"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_d",
        "notes": "D inside",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-02"),
        "te": pddt("2015-01-07"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_e",
        "notes": "E span",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-01"),
        "te": pddt("2015-01-04"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_f",
        "notes": "F left to middle",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-05"),
        "te": pddt("2015-01-08"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_g",
        "notes": "G right into middle",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-03"),
        "te": pddt("2015-01-04"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_h",
        "notes": "H inner left touching",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-05"),
        "te": pddt("2015-01-06"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_i",
        "notes": "I inner right touching",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-01"),
        "te": pddt("2015-01-02"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_j",
        "notes": "J outer left not touching",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-07"),
        "te": pddt("2015-01-08"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_11_k",
        "notes": "K outer right not touching",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-04"),
        "te": pddt("2015-01-04 00:00:00.000001"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_14_l",
        "notes": "L in mid, almost equal rs and re",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-05"),
        "te": pddt("2015-01-04"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_14_m",
        "notes": "M in middle, range_start > range_end",
        "test_only": 1,
    },
    {
        "ts": pddt("2015-01-04"),
        "te": pddt("2015-01-04"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "test_14_n",
        "notes": "N in middle, equal start and end",
        "test_only": 1,
    },
    # gap
    {
        "ts": pddt("2015-01-08"),
        "te": pddt("2016-01-01"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2016-01-01"),
        "te": pddt("2016-02-01"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestLocationtoMask:ltm_a",
        "notes": "ltm_a",
    },
    {
        "ts": pddt("2016-02-01"),
        "te": pddt("2016-03-01"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestLocationtoMask:ltm_b",
        "notes": "next to ltm_a",
    },
    # collides
    {
        "ts": pddt("2016-01-01"),
        "te": pddt("2016-03-01"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestLocationtoMask:ltm_c",
        "notes": "overlaps ltm_a and ltm_b",
        "test_only": 1,
    },
    {
        "ts": pddt("2016-03-01"),
        "te": pddt("2016-07-01"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestFullData",
        "notes": "",
    },
    {
        "ts": pddt("2016-07-01"),
        "te": pddt("2017-01-01"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestEvents",
        "notes": "",
    },
    {
        "ts": pddt("2017-01-01"),
        "te": pddt("2017-08-31"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestEvents",
        "notes": "",
    },
    {
        "ts": pddt("2017-08-31"),
        "te": pddt("2017-09-21"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestEvents",
        "notes": "",
    },
    {
        "ts": pddt("2017-09-21"),
        "te": pddt("2017-12-05"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestEvents",
        "notes": "",
    },
    {
        "ts": pddt("2017-12-05"),
        "te": pddt("2018-01-01"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestEvents",
        "notes": "",
    },
    {
        "ts": pddt("2018-01-01"),
        "te": pddt("2018-01-17"),
        "ltm_class": Location_to_Mask_8_flat,
        "name": MACHINE_NAME,
        "alias": "testflat8",
    },
    {
        "ts": pddt("2018-01-17"),
        "te": pddt("2018-02-16"),
        "ltm_class": Location_to_Mask_16,
        "name": MACHINE_NAME,
        "alias": "TestFullDataCollisions",
    },
    # halved in size from ^^^ to vvv
    {
        "ts": pddt("2018-02-16"),
        "te": pddt("2018-03-16"),
        "ltm_class": Location_to_Mask_8,
        "name": MACHINE_NAME,
        "alias": "TestFullDataCollisions",
    },
    {
        "ts": pddt("2018-03-16"),
        "te": pddt("2018-04-01"),
        "ltm_class": Location_to_Mask_Cray_192,
        "name": MACHINE_NAME,
        "alias": "TestLTM",
    },
    # gap
    {
        "ts": pddt("2018-04-01"),
        "te": pddt("2018-07-01"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2018-07-01"),
        "te": pddt("2038-01-01"),
        "ltm_class": Location_to_Mask_3Level,
        "name": MACHINE_NAME,
        "alias": "level_3",
    },
]

for ltm_module_dct in ltm_lst:
    if "pk" not in ltm_module_dct and ltm_module_dct["ltm_class"] not in [
        XLocation_to_Mask_DoesNotExist,
        XLocation_to_Mask_None,
    ]:
        pk = get_pk_hash(ltm_module_dct["name"], ltm_module_dct["ts"], ltm_module_dct["te"])
        ltm_module_dct["pk"] = pk
