# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

# cython: language_level=3, boundscheck=True
import atexit
import contextlib
import functools
import math
import multiprocessing as mp
import numbers
from _decimal import Underflow, Overflow
import abc
from collections.abc import Iterable
from functools import wraps
from typing import Dict, Tuple, List, Optional, Union, Set

import cython
import numpy as np
import simplejson
from numpy import int16, uint64
from numpy.core import bool_

from Octeres.bitmask.bitmask_funcs import (
    process_np_chunk,
    update_internal_mask_and_bit_count_inner_int16,
)
from Octeres.bitmask.bitmask_funcs import (
    update_internal_mask_and_bit_count_divide_memory,
)
from Octeres.bitmask.bitmask_globals import __base_dtype__
from Octeres.bitmask.bitmask_globals import (
    __boxed_max_base_dtype_value__,
    __boxed_min_base_dtype_value__,
)
from Octeres.bitmask.bitmask_globals import __chunk_size__
from Octeres.bitmask.bitmask_globals import __zero_chunk__
from Octeres.serialization import DataDecoder
from Octeres.serialization import DataEncoder
from Octeres.util import memoize_fast
from Octeres.util import superprint

int_bits_forward = list(enumerate([(2**i) for i in range(63, -1, -1)]))
uint64_bits_forward = [(idx, np.uint64(value)) for idx, value in int_bits_forward]
int_bits_backward = list(enumerate([(2**i) for i in range(63, -1, -1)]))
int_bits_backward.reverse()
uint64_bits_backward = [(idx, np.uint64(value)) for idx, value in int_bits_backward]
# import ray
# ray.init()

"""
  FIXME / TODO
numpy arrays have size, this should equal length in an enhanced bitmask.
shape is 1 dimensional, so it should not be an iterable
bit_total should be the number of bits, or "size" or "length"
shared_memory should be replaced as it uses to many files.
parallel is slower than serial, figure out why
when comparing two different length arrays, __equal__ will return True/False, not an array of True/False
    This messes with (b0 == b1).all()

split the bitmask into a magnitude bitmask(int16), a selection bitmask(bool) and a serial bitmask(uint8)(check for overflow on serialization instead)

Three places where variables need to be defined for an EnhancedBitmask
    __update_internal_mask_and_bit_count__
    __array_finalize__
    __setstate__
    
    Then for pickle:
        __reduce__
        
When going to a np_arr = arr.view(np.ndarray) and arr is an EnhancedBitmask, use it temporarily but don't convert it back,
keep a reference to the original array(arr).  Converting back to an EnhancedBitmask wipes out the memory, currently.
WARNING: Never call sum on a bitmask like object!  It calls __getitem__ for every bit, use bitmask.bit_count
"""

bool_True = bool_(True)
bool_False = bool_(False)
_original_np_error_config_ = np.seterr(all="raise")


def check_value_overflow(value, debug=False):
    if debug:
        superprint("check_value_overflow called", jump_back_count=4)
    if value > __boxed_max_base_dtype_value__:
        raise Overflow(f"value({value}) must not be over {__boxed_max_base_dtype_value__}")


def check_value_underflow(value, debug=False):
    if debug:
        superprint("check_value_underflow called", jump_back_count=4)
    if value < __boxed_min_base_dtype_value__:
        raise Underflow(f"value({value}) must be greater than {__boxed_min_base_dtype_value__}.")


def check_array_overflow(arr, debug=False):
    if debug:
        superprint("check_value_overflow called", jump_back_count=4)
    if (arr > __boxed_max_base_dtype_value__).any():
        raise Overflow(f"arr({arr}) index must not be over {__boxed_max_base_dtype_value__}")


def check_array_underflow(arr, debug=False):
    if debug:
        superprint("check_value_overflow called", jump_back_count=4)
    if (arr < __boxed_min_base_dtype_value__).any():
        raise Underflow(f"arr({arr}) index must not be over {__boxed_min_base_dtype_value__}")


def check_eb_overunderflow(e0, debug=False):
    arr = e0.view(np.ndarray)
    check_array_overflow(arr, debug=debug)
    check_array_underflow(arr, debug=debug)


def _get_np_prepared_bitmask(dtype: np.generic, length: int, ones=False, debug=False):
    """produce a numpy bitmask of a size and dtype.  Since this operation can take some time
    we will keep a cache of them to be used."""
    if debug:
        superprint(f"_get_np_prepared_bitmask: {dtype=} {length=} {ones=}")
    try:
        # search for the bitmask
        _bitmask = _prepared_bitmasks[(dtype, length, ones)]
    except KeyError:
        if debug:
            superprint(f"_get_np_prepared_bitmask:Creating {dtype=} {length=} {ones=}")
        # don't need to copy this as it's the first one.
        if dtype == uint64:
            if not ones:
                bitmask = EnhancedBitmask64.zeros(length)
            else:
                bitmask = EnhancedBitmask64.ones(length)
        else:
            if not ones:
                bitmask = EnhancedBitmask.zeros(length)
            else:
                bitmask = EnhancedBitmask.ones(length)
        # set it
        _prepared_bitmasks[(dtype, length, ones)] = bitmask
    else:
        # get a copy, we don't want pointers to the cache's node
        # .copy() of the object must correctly copy internal objects and memory.
        if debug:
            superprint(f"_get_np_prepared_bitmask:Copying {dtype=} {length=} {ones=}")
        # FIXME evaluate the copy
        bitmask = _bitmask.copy()
    return bitmask


@memoize_fast
def build_cache():
    """used to produce a lookup dictionary for finding the hex value
    for an associated section of a mask
    0000 -> FFFF

    This has a memory weight to it and can be slow, thus it's preferred to
    memoize it.
    """
    base_dtype = __base_dtype__
    lookup_hexmask_to_bitmask = {}
    lookup_binary_string_to_hexmask = {}
    lookup_hexmask_to_binary_string = {}
    for x in range(0, 65536):
        binstr = bin(x)[2:]
        hexstr = hex(x)[2:]
        binstr = binstr.rjust(16, "0")
        hexstr = hexstr.rjust(4, "0")
        # assert len(hexstr) == 4
        mask = np.array([int(i) for i in list(binstr)], dtype=base_dtype)
        lookup_hexmask_to_bitmask[hexstr] = mask
        binary_str = mask.tobytes()
        lookup_binary_string_to_hexmask[binary_str] = hexstr
        lookup_hexmask_to_binary_string[hexstr] = binary_str
    for x in range(0, 256):
        binstr = bin(x)[2:]
        hexstr = hex(x)[2:]
        binstr = binstr.rjust(8, "0")
        hexstr = hexstr.rjust(2, "0")
        # assert len(hexstr) == 2
        mask = np.array([int(i) for i in list(binstr)], dtype=base_dtype)
        lookup_hexmask_to_bitmask[hexstr] = mask
        binary_str = mask.tobytes()
        lookup_binary_string_to_hexmask[binary_str] = hexstr
        lookup_hexmask_to_binary_string[hexstr] = binary_str
    for x in range(0, 16):
        binstr = bin(x)[2:]
        hexstr = hex(x)[2:]
        binstr = binstr.rjust(4, "0")
        hexstr = hexstr.rjust(1, "0")
        # assert len(hexstr) == 1
        mask = np.array([int(i) for i in list(binstr)], dtype=base_dtype)
        lookup_hexmask_to_bitmask[hexstr] = mask
        binary_str = mask.tobytes()
        lookup_binary_string_to_hexmask[binary_str] = hexstr
        lookup_hexmask_to_binary_string[hexstr] = binary_str
    return (
        lookup_hexmask_to_bitmask,
        lookup_binary_string_to_hexmask,
        lookup_hexmask_to_binary_string,
    )


def uint64_bitmask_any(uint64_bitmask):
    result = False
    for chunk in uint64_bitmask:
        if chunk > 0:
            # don't need to check all, if one is set.
            return True
    return result


def uint64_bitmask_all(uint64_bitmask, length):
    chunk_size = 64  # np.iinfo(np.uint64).bits
    result = False
    current_length = 0
    total_all = 0
    len_chunks = 0
    max_value_uint64 = 2**chunk_size - 1
    for chunk in uint64_bitmask:
        if chunk == max_value_uint64:
            total_all += 1
        elif current_length + chunk_size > length:
            head_count = length - current_length
            # we need to cut the tail as it's not valid
            value = int(chunk) >> (chunk_size - head_count)
            max_value = (2**head_count) - 1
            if max_value == value:
                total_all += 1
        current_length += chunk_size
        len_chunks += 1
    if total_all == len_chunks:
        result = True
    return result


class EnhancedBitmaskMemory:
    def __init__(self, size_bytes, buf: bytes = None):
        """buf takes in a size or buffer to allocate the memory."""
        if buf:
            self.buf = bytearray(buf)
        else:
            self.buf = bytearray(size_bytes)

    def __getstate__(self, *args, **kwargs):
        """this is called when the object is pickled via dumps.  This is paired with __setstate__."""
        state = {"buf": self.buf}
        return state

    def __setstate__(self, state):
        """this is called when un-pickled via loads.  This is paired with __reduce__."""
        for key, value in state.items():
            setattr(self, key, value)


def convert_bytes_to_EnhancedBitmaskMemory(data: bytes) -> EnhancedBitmaskMemory:
    """
    convert a set of bytes to a buffer of int16.
    b'00fa' -> b'\x00\x00\x00\x00\x00\x00\x00\x00' \
               b'\x00\x00\x00\x00\x00\x00\x00\x00' \
               b'\x01\x00\x01\x00\x01\x00\x01\x00' \
               b'\x01\x00\x00\x00\x01\x00\x00\x00' \
    Note the endian being little.
    TODO: this is probably really really slow.
    """
    ints = [int(c, 16) for c in data.decode("utf8")]
    buf = b"".join([b"".join([b"\x00\x00" if digit == "0" else b"\x01\x00" for digit in bin(c)[2:].rjust(4, "0")]) for c in ints])
    memory = EnhancedBitmaskMemory(len(buf), buf=buf)
    return memory


class EnhancedBitmaskContext:
    @staticmethod
    def __debug_print__(*msg, jump_back_count=4):
        """this is the debug print based on the flag set in EB"""
        if EnhancedBitmask.__debug_flag__ is True:
            superprint(msg, jump_back_count=jump_back_count)

    def __init__(self, processes=4):
        #
        # self.bitmask_chunks = 128  # size * 128 is the amount we grab in one shot
        # self.chunk_lookup: Dict[Tuple[int, int]] = {}  # outter_chunk index, inner_chunk_index
        self.__debug_print__("ebc __init__ called")
        self.processes = processes
        if EnhancedBitmask.__parallel__:
            self.pool = mp.Pool(processes)
        # self.smm = SharedMemoryManager()
        # self.smm.start()

    def set_processes(self, processes):
        self.processes = processes
        if EnhancedBitmask.__parallel__:
            self.pool.terminate()
            self.pool.join()
            self.pool = mp.Pool(processes)

    def allocate_memory(self, size_bytes):
        """
        create shared memory that can be used inside a numpy ndarray as buffer.
        shm has a .name which allows getting of it.
        """
        self.__debug_print__(f"ebc allocate_memory({size_bytes}) called")
        # memory = self.smm.SharedMemory(size_bytes)
        memory = EnhancedBitmaskMemory(size_bytes)
        self.__debug_print__(f"ebc allocate_memory({size_bytes}) returning {id(memory)}")
        return memory

    def allocate_memory_for_int16(self, bits: int, ones=False):
        """
        create shared memory that can be used inside a numpy ndarray as buffer.
        shm has a .name which allows getting of it.
        """
        size_bytes = math.ceil((bits / 64) * 2) * 64
        self.__debug_print__("ebc allocate_memory_for_int16(", bits, ") called")
        # memory = self.smm.SharedMemory(size_bytes)
        memory = EnhancedBitmaskMemory(size_bytes)
        if ones:
            chunk_step = 2
            # set every other bit, note the endian...
            memory.buf[::chunk_step] = (b"\x01" * bits) + (b"\x00" * (int(size_bytes / 2) - bits))
        return memory

    def allocate_memory_for_uint64(self, ints: int, ones=False):
        """
        create shared memory that can be used inside a numpy ndarray as buffer.
        shm has a .name which allows getting of it.
        """
        size_bytes = ints * 8
        self.__debug_print__("ebc allocate_memory_for_uint64(", ints, ") called")
        # memory = self.smm.SharedMemory(size_bytes)
        memory = EnhancedBitmaskMemory(size_bytes)
        if ones:
            chunk_step = 2
            # set every other bit, note the endian...
            memory.buf[::chunk_step] = (b"\xff" * ones) + (b"\xff" * (int(size_bytes / 2) - ones))
        return memory

    def allocate_memory_from_bytes(self, buf):
        if type(buf) == bytes:
            # bytes cannot be modified so it must be a bytearray
            # the id of a bytes will not be equal to the bytearray
            buf = bytearray(buf)
        size_bytes = len(buf)
        memory = EnhancedBitmaskMemory(size_bytes, buf=buf)
        return memory

    def create_EnhancedBitmask_from_EnhancedBitmaskMemory(self, bits: int, memory: EnhancedBitmaskMemory) -> "EnhancedBitmask":
        # should this be size or shape???
        arr = np.ndarray((bits,), dtype=__base_dtype__, buffer=memory.buf)
        bitmask = arr.view(EnhancedBitmask)
        eb._set_memory(bitmask, memory)
        bitmask.__update_internal_mask_and_bit_count__()
        return bitmask

    def create_EnhancedBitmask(self, bits):
        self.__debug_print__("ebc create_EnhancedBitmask called")
        """
        size_bytes for an int16
        30 -> 128, should be 64
        32 -> 64
        33 -> 128
            
        """
        memory = self.allocate_memory_for_int16(bits)
        bitmask = EnhancedBitmask(bits, dtype=__base_dtype__, memory=memory)
        bitmask.__update_internal_mask_and_bit_count__()
        return bitmask

    def shutdown(self):
        self.__debug_print__("ebc shutdown called")
        if EnhancedBitmask.__parallel__:
            self.pool.terminate()
            self.pool.join()
        # self.smm.shutdown()


def bool_is_subset(left_right):
    """Is b a subset of a?"""
    left, right = left_right
    return 1 if (right & (not left)) else 0


class EnhancedBitmask64(np.ndarray):
    """create a EnhancedBitmask64 that contains a memory buffer."""

    __debug_flag__ = False

    @staticmethod
    def __debug_print__(*msg, jump_back_count=4):
        if EnhancedBitmask64.__debug_flag__ is True:
            superprint(msg, jump_back_count=jump_back_count)

    def __new__(
        cls,
        shape,
        dtype=uint64,
        buffer=None,
        offset=0,
        strides=None,
        order=None,
        memory=None,
    ):
        EnhancedBitmask.__debug_print__("eb64 __new__ called")
        obj = super().__new__(
            cls,
            shape,
            dtype=dtype,
            buffer=memory.buf,
            offset=offset,
            strides=strides,
            order=order,
        )
        obj.memory = memory
        return obj

    def __array_finalize__(self, obj):
        EnhancedBitmask.__debug_print__("eb64 __array_finalize__ called")
        if obj is None:
            return
        self.memory = getattr(obj, "memory", None)

    def __reduce__(self, *args, **kwargs):
        """this is called when the object is pickled via dumps.  This is paired with __setstate__."""
        self.__debug_print__("eb64 __reduce__ called")
        # ----
        state = super().__reduce__()
        func, cls, tail = state
        tail = tail + (self.memory,)
        return func, cls, tail

    def __setstate__(self, state):
        """this is called when un-pickled via loads.  This is paired with __reduce__."""
        EnhancedBitmask.__debug_print__("eb64 __setstate__ ", state)
        memory = state[-1]
        self.memory = memory
        obj = super().__setstate__(state[0:-1])
        return obj

    @staticmethod
    def zeros(shape: Union[int, Tuple[int]], dtype=uint64, order="C"):
        size_bytes = shape * 8
        memory = ebc.allocate_memory(size_bytes)
        arr = np.ndarray(shape, dtype=dtype, order=order, buffer=memory.buf)
        obj = arr.view(EnhancedBitmask64)
        obj.memory = memory
        return obj

    @staticmethod
    def ones(shape: Union[int, Tuple[int]], dtype=uint64, order="C"):
        size_bytes = shape * 8
        memory = ebc.allocate_memory(size_bytes)
        # set every other bit
        memory.buf[:] = b"\xff" * size_bytes
        arr = np.ndarray(shape, dtype=dtype, order=order, buffer=memory.buf)
        obj = arr.view(EnhancedBitmask64)
        obj.memory = memory
        return obj

    def tobytes(self, order="C", direct=False):
        """
        Warning, this will give you direct access to the numpy array's memory in 1 of the 3 cases.
        """
        if hasattr(self, "memory") and self.memory:
            if direct:
                data = self.memory.buf
            else:
                data = self.memory.buf[:]
        else:
            data = super().tobytes(order)
        return data

    def copy(self, *args, **kwargs):
        self.__debug_print__("eb64 copy called")
        eb64_copy = super(EnhancedBitmask64, self).copy()
        # now we have to deal with .memory
        # deepcopy is slow.
        if hasattr(self, "memory") and self.memory:
            # eb64_copy.memory = deepcopy(eb64_copy.memory)
            eb64_copy.memory = EnhancedBitmaskMemory(None, self.memory.buf)
        else:
            eb64_copy.memory = None
        return eb64_copy

    def all(self, axis=None, out=None, keepdims=False) -> bool:
        self.__debug_print__("all called")
        obj = self.view(np.ndarray)
        return obj.all()

    def any(self, axis=None, out=None, keepdims=False) -> bool:
        self.__debug_print__("any called")
        obj = self.view(np.ndarray)
        return obj.any()

    # used for debugging only.
    # def __setattr__(self, key, value):
    #     self.__debug_print__(f'EB64 __setattr__ called {key=} {value=}')
    #     super().__setattr__(key, value)


class EnhancedBitmaskLike(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def to_json(self):
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def from_json(json_str) -> "EnhancedBitmaskLike":
        raise NotImplementedError()

    # @abc.abstractmethod
    # def to_dict(self) -> dict:
    #     raise NotImplementedError()

    @property
    @abc.abstractmethod
    def flags(self) -> object:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def length(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def bit_count(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def bit_idx_first(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def bit_idx_last(self) -> int:
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def array(bit_lst, **kwargs):
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def zeros(length: int) -> "EnhancedBitmaskLike":
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def ones(length: int) -> "EnhancedBitmaskLike":
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def zeros_and_set_bits(length: int, bit_lst: Union[List[int], Tuple[int, ...], Set[int]], **kwargs):
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def zeros_and_set_bit(length: int, bit_idx):
        raise NotImplementedError()

    # @staticmethod
    # @abc.abstractmethod
    # def zeros_and_set_ranges(length: int, bit_ranges: Tuple[Tuple[int, int]]):
    #     raise NotImplementedError()

    def all(self, **kwargs) -> bool:
        return bool_True if self.length == self.bit_count else bool_False

    def any(self, **kwargs) -> bool:
        return bool_True if self.bit_count > 0 else bool_False

    def none(self, **kwargs) -> bool:
        return bool_True if self.bit_count == 0 else bool_False

    @abc.abstractmethod
    def copy(self, *args, **kwargs):
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def add(x1, x2, **kwargs):
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def logical_xor(x, y):
        raise NotImplementedError()

    def __len__(self):
        try:
            length = self.length
        except (AttributeError, UnboundLocalError):
            length = -1
        return length

    @staticmethod
    @abc.abstractmethod
    def logical_not(x):
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def logical_xnor(x, y):
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def logical_or(x, y):
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def logical_and(x, y):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def logical_is_subset(cls, left, right):
        raise NotImplementedError()

    @abc.abstractmethod
    def intersection_any(self, other: "EnhancedBitmaskLike"):
        raise NotImplementedError()

    def intersection_all(self, other: "EnhancedBitmaskLike"):
        """find the intersection between self and other enhanced bitmasks.
        If it finds a full intersection it will true.
        if one is 0 1 1 1 0 0 0 0 and the other is
                  0 1 1 1 0 0 0 0, it will return true.
        if one is 1 1 1 1 0 0 0 0 and the other is
                  0 1 1 1 1 0 0 0, it will return false.
        if one is 1 1 1 1 0 0 0 0 and the other is
                  0 1 1 1 0 0 0 0, it will return false.
        """
        # type_other = type(other)
        if isinstance(other, EnhancedBitmaskLike) and self.bit_count != other.bit_count:
            result = False
        # if type_other == EnhancedBitmask \
        #         and self.bit_count != other.bit_count:
        #     result = False
        elif (
            isinstance(other, EnhancedBitmaskLike)
            and self.bit_count > 0
            and other.bit_count > 0
            and (self.bit_idx_last < other.bit_idx_first or self.bit_idx_first > other.bit_idx_last)
        ):
            result = False
        else:
            xnored = self.logical_xnor(self, other)
            if xnored.bit_count == xnored.length:
                result = True
            else:
                result = False
        return result

    @classmethod
    def join_bitmasks(cls, bitmask_lst: list, operation) -> "EnhancedBitmaskLike":
        raise NotImplementedError()

    @staticmethod
    def get_bitmask_from_idx_lst(bit_count, idx_lst) -> "EnhancedBitmaskLike":
        """create a bitmask from an index list"""
        raise NotImplementedError()

    @staticmethod
    def get_idx_lst_from_bitmask(bitmask, bit_idx_first: int = 0, bit_idx_last: int = None) -> list:
        """return the indexes set inside the bitmask"""
        raise NotImplementedError()


def magnitude_add(left: "EnhancedBitmask", right: "EnhancedBitmask"):
    base, applier = (left.copy(), right) if left.bit_count > right.bit_count else (right.copy(), left)
    for i in np.where(applier != 0):
        base[i] += applier[i]
    # EnhancedBitmask.get_idx_lst_from_bitmask(applier)
    # result = np.add(base, applier)
    return base


class EnhancedBitmask(np.ndarray, EnhancedBitmaskLike):
    """This is an enhanced version of a numpy ndarray.
    The base dtype is a np.int16, but it surrounds a np.int8 to prevent overflows.
    shape can only be an int for a 1D array

    Much use of the information on this page:
    https://numpy.org/doc/stable/user/basics.subclassing.html
    """

    __debug_flag__ = False  # used to turn on and off debug messages
    __flow_check__ = False  # will go away later and always check
    __chunk_size__ = __chunk_size__  # <<<<<<<--\  64 bit chunks
    __chunk_zeros__ = "0" * __chunk_size__  # from above
    __extra_attr_count__ = 2

    length: int = 0  # same as size
    bit_count: int = 0
    bit_idx_first = -1
    bit_idx_last = -1

    __parallel__ = False  # use multiprocessing
    __allowed_dtypes__ = [np.int16]  # do not change, np.int16 will be the only supported dtype
    __state_attrib_count__ = 5  # see __setstate__ and __reduce__

    __direct_equal__ = False  # this is to allow checking with == and return a bool.

    __version_class__ = "2.0"

    @classmethod
    def set_parallel(cls, value: bool):
        """turn on parallelization."""
        cls.__debug_print__("set_parallel ", value)
        if value:
            raise NotImplementedError("__parallel__ is not implemented")
        cls.__parallel__ = value

    @staticmethod
    def __debug_print__(*msg, jump_back_count=4):
        if EnhancedBitmask.__debug_flag__ is True:
            if not EnhancedBitmask.__parallel__:
                superprint(*msg, jump_back_count=jump_back_count)
            else:
                superprint(*msg, jump_back_count=jump_back_count, parallel=True)

    @classmethod
    def debugger(cls):
        class EnhancedBitmaskDebugger:
            def __init__(self):
                print()
                EnhancedBitmask.__debug_flag__ = False
                EnhancedBitmask64.__debug_flag__ = False

            def __enter__(self):
                EnhancedBitmask.__debug_flag__ = True
                EnhancedBitmask64.__debug_flag__ = True

            def __exit__(self, exc_type, exc_value, traceback):
                EnhancedBitmask.__debug_flag__ = False
                EnhancedBitmask64.__debug_flag__ = False

        obj = EnhancedBitmaskDebugger()
        return obj

    # def __init__(self, *args, **kwargs):
    #     self.__bool_array = None
    #     EnhancedBitmask.__debug_print__("__init__")
    #     np.ndarray.__init__(self, *args, **kwargs)

    def __repr_fast__(self):
        """can only be used if all the items in the mask are 0 or 1"""
        # TODO: use this.
        str_lst = []
        for i in range(self.length // 64):
            str_lst.append(f"{bin(self.bitmask[i])[2:]:0>64}")
        remainder = self.length % 64
        if remainder:
            bitmask_str = bin(self.bitmask[-1])[2:].zfill(64)[:remainder]
            if len(bitmask_str) < remainder:
                bitmask_str.zfill(remainder)
            str_lst.append(bitmask_str)
        result = "".join(str_lst)
        # assert len(str_lst) == self.length//64 + (1 if remainder else 0)
        result = f"EnhancedBitmask.array([{', '.join(result)}])"
        return result

    def __repr__(self):
        # FIXME: this doesn't work.
        # self[self > 1]
        # result = self.__repr_fast__()
        try:
            result = f"EnhancedBitmask.array([{str(', '.join(map(str, self.tolist())))}])"
        except TypeError:
            result = f"EnhancedBitmask.array([{str(int(self))}])"
        return result

    @staticmethod
    def _set_memory(obj, memory):
        EnhancedBitmask.__debug_print__("_set_memory", memory, jump_back_count=4)
        obj.memory = memory

    @staticmethod
    def _set_bitmask(obj, length, ones=False):
        pass

    def __new__(
        cls,
        shape,
        buffer=None,
        offset=0,
        strides=None,
        order=None,
        dtype=None,
        memory=None,
    ):
        EnhancedBitmask.__debug_print__("eb __new__ called")
        if buffer:
            raise NotImplemented("buffer not allowed")
        dtype = __base_dtype__
        assert isinstance(shape, int), "shape(" + str(shape) + ") must be an 1D array"
        if dtype == int16:
            if memory:
                obj = np.ndarray.__new__(
                    cls,
                    shape,
                    dtype=dtype,
                    buffer=memory.buf,
                    offset=offset,
                    strides=strides,
                    order=order,
                )
            else:
                memory = ebc.allocate_memory(shape * 2)
                obj = np.ndarray.__new__(
                    cls,
                    shape,
                    dtype=dtype,
                    buffer=memory.buf,
                    offset=offset,
                    strides=strides,
                    order=order,
                )
        else:
            obj = np.ndarray.__new__(
                cls,
                shape,
                dtype=dtype,
                buffer=buffer,
                offset=offset,
                strides=strides,
                order=order,
            )
        if dtype == int16:
            EnhancedBitmask._set_memory(obj, memory)
        # add the new attribute to the created instance
        # obj.bitmask = np.zeros(math.ceil(shape/__chunk_size__), dtype=uint64)
        chunk_count = math.ceil(shape / __chunk_size__)
        obj.bitmask = _get_np_prepared_bitmask(uint64, chunk_count, debug=cls.__debug_flag__)
        obj.bit_count = 0
        obj.length = 0
        obj.bit_idx_first = -1
        obj.bit_idx_last = -1
        obj.__bool_array = None
        return obj

    def get_chunk_bit_count_and_chunk_count(self, obj):
        self.__debug_print__(f"eb get_chunk_bit_count_and_chunk_count called {obj.bit_count=} {self.shape=}")
        bit_count = obj.bit_count
        if self.shape == ():
            int_self = int(self)
            if int_self > 0:
                bit_count = 1
            chunk_count = 1
        else:
            if obj.shape == ():
                chunk_count = 1
            else:
                chunk_count = math.ceil(obj.shape[0] / __chunk_size__)
        return bit_count, chunk_count

    # used for debugging only.
    # def __setattr__(self, key, value):
    #     self.__debug_print__(f'EB __setattr__ called {key=} {value=}')
    #     super().__setattr__(key, value)

    def __array_finalize__(self, obj):
        EnhancedBitmask.__debug_print__(f"eb __array_finalize__ called on {id(self)=}")
        if obj is None:
            EnhancedBitmask.__debug_print__(f"__array_finalize__ {obj=}")
        #    return
        if hasattr(obj, "bit_count"):
            EnhancedBitmask.__debug_print__(f"obj:{id(obj)=} bit_count:{obj.bit_count}")
            # we need to copy as in the case of a == operation, it's one of the objects.
            self.bitmask = obj.bitmask.copy()
            if self.shape == ():
                length = 1
            else:
                length = obj.length
            # FIXME: this is where memory gets set.  If mul is called, we don't want to link them.
            EnhancedBitmask._set_memory(self, None)
            self.__bool_array = None
            bit_count, chunk_count = self.get_chunk_bit_count_and_chunk_count(obj)
            bit_idx_first, bit_idx_last = self.__update_bit_idx_first_and_last__()
            self.bit_idx_first = bit_idx_first
            self.bit_idx_last = bit_idx_last
        else:
            """
            This could be the case of eb.zeros(10) and it's created the array and started
            to build the EB, but not done yet.
            """
            EnhancedBitmask.__debug_print__("NO bit_count")
            EnhancedBitmask._set_memory(self, None)
            if self.shape == ():
                length = 1
            else:
                length = self.shape[0]
            if self.shape == (0,) and hasattr(self, "bit_count") and self.bit_count == 0:
                EnhancedBitmask.__debug_print__("incomplete self and obj, hopefully __setstate__ is going to be called.")
                self.bitmask = None
                bit_count = None
                chunk_count = None
            else:
                bit_count, chunk_count = self.get_chunk_bit_count_and_chunk_count(self)
                try:
                    self.bitmask = _get_np_prepared_bitmask(
                        uint64, chunk_count, debug=self.__debug_flag__
                    )  # create a new bitmask.
                except:
                    superprint(f"{bit_count=} {self.shape=} {chunk_count=} {self=} {self.shape} {obj=}")
                    superprint(f"{self.to_json()}")
                    raise
            self.bit_idx_first = length
            self.bit_idx_last = -1
        self.length: int = length
        self.bit_count = bit_count
        self.__chunk_count__ = chunk_count
        self.__not_cache = None

    def __reduce__(self, *args, **kwargs):
        """this is called when the object is pickled via dumps.  This is paired with __setstate__."""
        self.__debug_print__("eb __reduce__ called")
        # ----
        state = super(EnhancedBitmask, self).__reduce__()
        func, cls, tail = state
        # add to the tail, notice self.bitmask is a np.ndarray and will be pickle later.
        # tail = tail + (self.bitmask, self.bit_count, self.length, self.memory, self.bit_idx_first, self.bit_idx_last)
        tail = tail + (
            self.bitmask,
            self.bit_count,
            self.length,
            self.bit_idx_first,
            self.bit_idx_last,
        )
        return func, cls, tail

    def __setstate__(self, state):
        """this is called when un-pickled via loads.  This is paired with __reduce__."""
        EnhancedBitmask.__debug_print__(f"eb __setstate__ ")
        # get the tail
        # bitmask, bit_count, length, memory, bit_idx_first, bit_idx_last = state[-self.__state_attrib_count__:]
        bitmask, bit_count, length, bit_idx_first, bit_idx_last = state[-self.__state_attrib_count__ :]
        self.bit_count = bit_count
        self.bitmask = bitmask
        self.length: int = length
        # self.memory = memory
        self.bit_idx_first: Optional[int] = bit_idx_first
        self.bit_idx_last: Optional[int] = bit_idx_last
        self.__bool_array = None
        self.__not_cache = None
        self.__chunk_count__ = math.ceil(self.length / self.__chunk_size__)
        # Call the parent's __setstate__ with the other tuple elements.
        obj = super(EnhancedBitmask, self).__setstate__(state[0 : -self.__state_attrib_count__])
        bit_idx_first, bit_idx_last = self.__update_bit_idx_first_and_last__(
            bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last
        )
        self.bit_idx_first = bit_idx_first
        self.bit_idx_last = bit_idx_last
        return obj

    # def __getstate__(self, *args, **kwargs):
    #     """not used, look at __reduce__"""
    #     EnhancedBitmask.__debug_print__('__getstate__')
    #     obj = np.ndarray.__getstate__(self, *args, **kwargs)
    #     return obj

    def __array_wrap__(self, out_arr):
        EnhancedBitmask.__debug_print__(f"eb __array_wrap__ {id(self)=} {id(out_arr)=}")
        if out_arr.shape == self.shape:
            obj = np.ndarray.__array_wrap__(self, out_arr)
            obj.__update_internal_mask_and_bit_count__()
        else:  # in the case of sum, it returns a uint64
            obj = np.asarray(out_arr)
        return obj

    def validate(self):
        EnhancedBitmask.__debug_print__(f"validate:{id(self)=} bit_count:{self.bit_count}")
        e0_bc0 = int(np.sum(self.to_bool_array()))
        e0_bc1 = self.bit_count
        e0_len = self.length
        # self.bitmask is the internal bitmask, not the usable one
        bitmask_lst = self.bitmask.tolist()
        e0_bc2 = sum([bin(bigint).lstrip("0b").count("1") for bigint in bitmask_lst])
        assert e0_bc0 == e0_bc1, "%s != %s" % (e0_bc0, e0_bc1)
        assert e0_bc1 == e0_bc2, "%s != %s" % (e0_bc1, e0_bc2)
        len_bitmask = len(list([_ for _ in self]))
        assert e0_len == len_bitmask, "%s != %s" % (e0_len, len_bitmask)
        # assert e0_len == len(bitmask_lst), "%s != %s" % (e0_len, len(bitmask_lst))
        self.__debug_print__(e0_bc0, e0_bc1, e0_bc2)
        chunk_count = math.ceil(self.shape[0] / self.__chunk_size__)
        assert chunk_count == self.__chunk_count__
        if hasattr(self, "memory"):
            if self.memory:
                assert type(self.memory.buf) != bytes
        idx_lst = self.get_idx_lst_from_bitmask(self)
        if idx_lst:
            assert min(idx_lst) == self.bit_idx_first
            assert max(idx_lst) == self.bit_idx_last
        else:
            assert self.bit_idx_first == self.length
            assert self.bit_idx_last == -1

    @staticmethod
    def logical_and(x, y):
        EnhancedBitmask.__debug_print__("logical_and")
        out_arr = _get_np_prepared_bitmask(__base_dtype__, len(x), debug=EnhancedBitmask.__debug_flag__)
        arr = np.logical_and(x, y, out=out_arr)
        obj = arr  # obj = arr.view(EnhancedBitmask)
        obj.__update_internal_mask_and_bit_count__()
        return obj

    @staticmethod
    def logical_or(x, y):
        EnhancedBitmask.__debug_print__("logical_or")
        out_arr = _get_np_prepared_bitmask(__base_dtype__, len(x), debug=EnhancedBitmask.__debug_flag__)
        arr = np.logical_or(x, y, out=out_arr)
        obj = arr  # obj = arr.view(EnhancedBitmask)
        obj.__update_internal_mask_and_bit_count__()
        return obj

    @staticmethod
    def logical_not(x):
        EnhancedBitmask.__debug_print__("logical_not")
        out_arr = _get_np_prepared_bitmask(__base_dtype__, len(x), debug=EnhancedBitmask.__debug_flag__)
        arr = np.logical_not(x, out=out_arr)
        obj = arr  # obj = arr.view(EnhancedBitmask)
        obj.__update_internal_mask_and_bit_count__()
        return obj

    @staticmethod
    def logical_xor(x, y):
        EnhancedBitmask.__debug_print__("logical_xor")
        out_arr = _get_np_prepared_bitmask(__base_dtype__, len(x), debug=EnhancedBitmask.__debug_flag__)
        arr = np.logical_xor(x, y, out=out_arr)
        obj = arr  # obj = arr.view(EnhancedBitmask)
        obj.__update_internal_mask_and_bit_count__()
        return obj

    # this is an attempt to short circuit the use case for logical_xnor

    @staticmethod
    def _logical_xnor(x, y):
        """this prepares the numpy xnor'ed arrays"""
        out_arr = _get_np_prepared_bitmask(__base_dtype__, len(x), debug=EnhancedBitmask.__debug_flag__)
        arr = np.logical_xor(x, y, out=out_arr)
        arr = np.logical_not(arr)
        return arr

    @staticmethod
    def logical_xnor(x, y):
        """logical xnor of two arrays"""
        EnhancedBitmask.__debug_print__("logical_xnor")
        arr = EnhancedBitmask._logical_xnor(x, y)
        obj = arr  # obj = arr.view(EnhancedBitmask)
        obj.__update_internal_mask_and_bit_count__()
        return obj

    @staticmethod
    def logical_truth(operator, truth, x, y):
        EnhancedBitmask.__debug_print__("logical_truth, ", operator, truth)
        truths = (
            "any",
            "all",
        )
        operator = operator.lstrip("_")
        operators = ("logical_xnor",)
        if truth not in truths:
            raise Exception("Truth %s is not valid, please use %s" % (truth, ",".join(truths)))
        if operator not in operators:
            raise Exception("Operator %s is not valid, please use %s" % (operator, ",".join(operators)))
        logical_method = getattr(EnhancedBitmask, operator)
        arr = logical_method(x, y)
        if truth == "all":
            result = arr.all()
        else:  # any fall through
            result = arr.any()
        return result

    @staticmethod
    def get_bitmask_from_idx_lst(bit_count, idx_lst) -> "EnhancedBitmask":
        """create a bitmask from an index list"""
        bitmask = EnhancedBitmask.zeros_and_set_bits(bit_count, idx_lst)
        return bitmask

    @staticmethod
    def get_idx_lst_from_bitmask(bitmask, bit_idx_first: int = 0, bit_idx_last: int = None) -> list:
        """return the indexes set inside the bitmask"""
        idx_lst = []
        idx_lst_append = idx_lst.append
        bit_idx_first = bitmask.bit_idx_first if bit_idx_first is None else bit_idx_first
        bit_idx_last = bitmask.bit_idx_last if bit_idx_last is None else bit_idx_last
        if bit_idx_first == bitmask.length:
            pass  # no bits set
        else:
            bit_delta = (bit_idx_last + 1) - bit_idx_first
            if bit_delta == bitmask.bit_count:
                idx_lst = list(range(bit_idx_first, bit_idx_last + 1))
            else:
                idx_lst = np.where(bitmask[bit_idx_first : bit_idx_last + 1] != 0)[0].tolist()
                # idx = bit_idx_first
                # for bit in bitmask[bit_idx_first: bit_idx_last + 1]:
                #     if bit > 0:
                #         idx_lst_append(idx)
                #     idx += 1
        # VERY SLOW
        # for idx, bit in enumerate(bitmask):
        #     if bit > 0:  # a bit could be > 1
        #         idx_lst_append(idx)
        return idx_lst

    @staticmethod
    def zeros(length: int) -> "EnhancedBitmask":
        EnhancedBitmask.__debug_print__(f"zeros, {length=}")
        size_bytes = math.ceil((length / 64) * 2) * 64
        memory = ebc.allocate_memory(size_bytes)
        arr = np.ndarray(length, dtype=__base_dtype__, order="C", buffer=memory.buf)
        # this will call __array_finalize__ ...
        obj = arr.view(EnhancedBitmask)
        # __array_finalize__ doesn't know about memory, but we do, set it.
        EnhancedBitmask._set_memory(obj, memory)
        obj.__update_internal_mask_and_bit_count__(bit_idx_first=obj.length, bit_idx_last=-1)
        return obj

    @staticmethod
    def ones(length: int) -> "EnhancedBitmask":
        EnhancedBitmask.__debug_print__(f"ones, {length=}")
        size_bytes = math.ceil((length / 64) * 2) * 64
        memory = ebc.allocate_memory(size_bytes)
        # set every other bit
        memory.buf[::2] = (b"\x01" * length) + (b"\x00" * (int(size_bytes / 2) - length))
        arr = np.ndarray(length, dtype=__base_dtype__, order="C", buffer=memory.buf)
        obj = arr.view(EnhancedBitmask)
        EnhancedBitmask._set_memory(obj, memory)
        obj.__update_internal_mask_and_bit_count__(bit_idx_first=0, bit_idx_last=obj.length - 1)
        return obj

    @staticmethod
    def array(p_object, dtype=__base_dtype__, copy=True, order=None, subok=False, ndmin=0):
        EnhancedBitmask.__debug_print__(f"array, {type(p_object)=}")
        # FIXME: add code to use memory.  This may be difficult.
        # FIXME: buffer up to multiples of 64 bits.  if you pass in 14 -> 28, tobytes gets angry..  note, left shifted bitmasks
        bit_idx_first = None
        bit_idx_last = None
        uimabc = True
        if dtype not in EnhancedBitmask.__allowed_dtypes__:
            raise NotImplementedError()
        if isinstance(p_object, np.ndarray):
            if p_object.dtype == __base_dtype__:
                arr = p_object
            elif p_object.dtype == bool:
                arr = p_object
            elif p_object.dtype != __base_dtype__:
                check_array_overflow(p_object, debug=EnhancedBitmask.__debug_flag__)
                check_array_underflow(p_object, debug=EnhancedBitmask.__debug_flag__)
                arr = p_object.astype(dtype=__base_dtype__)
            else:
                arr = p_object
        elif isinstance(p_object, list):
            for _ in p_object:
                check_value_overflow(_, debug=EnhancedBitmask.__debug_flag__)
                check_value_underflow(_, debug=EnhancedBitmask.__debug_flag__)
            length = len(p_object)
            arr = EnhancedBitmask.zeros(length)
            np_arr = arr.view(np.ndarray)  # we will do all the updates at the end, switch to ndarray for now.
            for idx, value in enumerate(p_object):
                # this will call __setitem__ a bunch of times.
                if bit_idx_first is None and value > 0:
                    bit_idx_first = idx
                if bit_idx_first is not None and value > 0:
                    bit_idx_last = idx
                np_arr[idx] = value
            # arr = np.array(p_object, dtype=dtype, copy=copy, order=order, subok=subok, ndmin=ndmin)
            # obj = arr
            uimabc = True
        else:
            arr = np.array(p_object, dtype=dtype, copy=copy, order=order, subok=subok, ndmin=ndmin)
        if type(arr) != EnhancedBitmask:
            obj = arr.view(EnhancedBitmask)
        else:
            obj = arr
        if uimabc or obj.shape == ():
            obj.__update_internal_mask_and_bit_count__(bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last)
        assert hasattr(obj, "bitmask")
        return obj

    def astype(self, dtype, order="K", casting="unsafe", subok=True, copy=True):
        if dtype != __base_dtype__:
            raise NotImplementedError()
        self.__debug_print__("eb astype called")
        obj = super().astype(dtype, order=order, casting=casting, subok=subok, copy=copy)
        return obj

    @staticmethod
    def empty(length):
        return EnhancedBitmask.zeros(length)

    @staticmethod
    def zeros_and_set_bit(length: int, bit_idx: int):
        return EnhancedBitmask.zeros_and_set_bits(length, [bit_idx])

    @staticmethod
    def zeros_and_set_bits(length: int, bit_lst, **kwargs):
        # size_bytes = shape * 2  # int16 is the number of bits * 2
        memory = ebc.allocate_memory_for_int16(length)
        arr = np.ndarray(length, dtype=__base_dtype__, order="C", buffer=memory.buf)
        if type(bit_lst) == list:
            arr[bit_lst] = 1
        else:
            for bit_idx in bit_lst:
                if bit_idx is not None:
                    arr[bit_idx] = 1
        obj = arr.view(EnhancedBitmask)
        EnhancedBitmask._set_memory(obj, memory)
        if bit_lst:
            bit_idx_first: int = min(bit_lst)
            bit_idx_last: int = max(bit_lst)
        else:
            bit_idx_first: int = length
            bit_idx_last: int = -1
        obj.__update_internal_mask_and_bit_count__(bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last)
        return obj

    def describe(self, deep=False):
        superprint(f" ---- {self.__class__} ---- ")
        superprint(f"     {str(self.bit_count)=: >16}    number of bits set.")
        superprint(f"        {str(self.length)=: >16}    length of the array. (may be renamed to bit_total)")
        superprint(f" {str(self.bit_idx_first)=: >16}    first index of set bit.")
        superprint(f"  {str(self.bit_idx_last)=: >16}    last idx of set bit.")
        superprint(f"           {str(id(self))=: >16}")
        superprint(f"   {str(id(self.bitmask))=: >16}")
        superprint(f"    {str(id(self.memory))=: >16}    {id(self.bitmask.memory)=}")
        if hasattr(self.memory, "buf"):
            superprint(f"    {id(self.memory.buf)=}   {id(self.bitmask.memory.buf)=}")
            superprint(f"   {len(self.memory.buf)=}   {len(self.bitmask.memory.buf)=}")
        else:
            superprint(f"   buf does not exist.")
        if deep:
            superprint(f"{self=}")
            superprint(f"{self.memory=}")
            superprint(f"{self.bitmask=} {bin(self.bitmask[0])=}")
            superprint(f"{self.bitmask.memory=}")

    @staticmethod
    def ones_and_unset_bits(bits: int, bit_lst):
        memory = ebc.allocate_memory_for_int16(bits, ones=True)
        arr = np.ndarray(bits, dtype=__base_dtype__, order="C", buffer=memory.buf)
        for bit_idx in bit_lst:
            arr[bit_idx] = 0
        obj = arr.view(EnhancedBitmask)
        EnhancedBitmask._set_memory(obj, memory)
        obj.__update_internal_mask_and_bit_count__()
        return obj

    @staticmethod
    def fromstring(string, dtype=__base_dtype__, count=-1, sep=""):
        arr = np.fromstring(string, dtype=dtype, count=count, sep=sep)
        obj = arr.view(EnhancedBitmask)
        obj.__update_internal_mask_and_bit_count__()
        assert hasattr(obj, "bitmask")
        return obj

    @staticmethod
    def fromhex(size: int, hexmask: bytes):
        """create an EnhancedBitmask from a hexmask."""
        memory = convert_bytes_to_EnhancedBitmaskMemory(hexmask)
        obj = ebc.create_EnhancedBitmask_from_EnhancedBitmaskMemory(size, memory)
        return obj

    @staticmethod
    def frombuffer(buffer, dtype=__base_dtype__, count=-1, offset=0):
        """this is not implemented because you require a shape or size which buffers don't have.
        Enhanced bitmasks buffer up memory to the next chunk size(64).  so if 30 bits are used on
        an int16 that would be 30*2 bytes of data + zeros."""
        raise NotImplementedError()
        # if dtype != __base_dtype__:
        #     raise NotImplementedError()
        # memory = ebc.allocate_memory_from_bytes(buffer)
        # obj = ebc.create_EnhancedBitmask_from_EnhancedBitmaskMemory(size???, memory)

        # arr = np.frombuffer(buffer, dtype=dtype, count=count, offset=offset)
        # obj = arr.view(EnhancedBitmask)
        # obj.__update_internal_mask_and_bit_count__()
        # assert hasattr(obj, 'bitmask')
        # return obj

    @staticmethod
    def concatenate(other_lst):
        EnhancedBitmask.__debug_print__("concatenate")
        length = sum([len(a) for a in other_lst])
        eb_array = EnhancedBitmask.zeros(length)
        np.concatenate(other_lst, out=eb_array)
        eb_array.__update_internal_mask_and_bit_count__()
        return eb_array

    @staticmethod
    def sign(x1: "EnhancedBitmask", *args, **kwargs) -> "EnhancedBitmask":
        """this has a slightly different use than np.sign as this is used to
        take a mask that has bit's set > 1 back down to 1"""
        EnhancedBitmask.__debug_print__("sign")
        arr = np.sign(x1, *args, **kwargs)
        if EnhancedBitmask.__flow_check__:
            truth: EnhancedBitmask = np.greater_equal(x1, 0)
            try:
                assert truth.all()
            except AssertionError:
                raise Underflow("Underflow (x1 >= 0).all() is False")
        obj = arr.view(EnhancedBitmask)
        obj.__update_internal_mask_and_bit_count__()
        return obj

    @staticmethod
    def from_json(json_str):
        arr = simplejson.loads(json_str, cls=DataDecoder)
        obj = arr.view(EnhancedBitmask)
        # obj.__update_internal_mask_and_bit_count__()
        return obj

    @staticmethod
    def kron(a, b):
        EnhancedBitmask.__debug_print__("eb kron called")
        arr = np.kron(a, b)
        obj = arr.view(EnhancedBitmask)
        obj.__update_internal_mask_and_bit_count__()
        return obj

    @staticmethod
    def subtract(x1, x2, **kwargs):
        arr = np.subtract(x1, x2, **kwargs)
        check_eb_overunderflow(arr, debug=EnhancedBitmask.__debug_flag__)
        obj = arr.view(EnhancedBitmask)
        return obj

    @staticmethod
    def add(x1, x2, **kwargs):
        arr = np.add(x1, x2, **kwargs)
        check_eb_overunderflow(arr, debug=EnhancedBitmask.__debug_flag__)
        obj = arr  # obj = arr.view(EnhancedBitmask)
        return obj

    def tobytes(self, order="C", direct=False):
        """
        Warning, this will give you direct access to the numpy array's memory in 1 of the 3 cases.
        """
        if hasattr(self, "memory") and self.memory:
            self.__debug_print__("eb tobytes called(memory)")
            if direct:
                data = self.memory.buf
            else:
                if self.dtype == np.int16:
                    # each array item is 2 bytes at int16
                    # buffers come in chunks, and are left indexes so the minimum chunk is 64 based on the internal uint464
                    data = self.memory.buf[: self.length * 2]
                else:
                    raise Exception(f"unable to convert to bytes, {self}")
        else:
            self.__debug_print__("eb tobytes called")
            data = super().tobytes(order)
        return data

    def to_json(self):
        self.__debug_print__("eb to_json called")
        json_str = simplejson.dumps(self, cls=DataEncoder)
        return json_str

    def copy(self, *args, **kwargs):
        self.__debug_print__(f"eb copy called")
        if self.dtype == __base_dtype__:
            # extract the existing data and create a new buffer of the correct size.
            # FIXME: EVALUATE OLD WAY, PRE-EnhancedBitmask64.  Recursively call copy on self.bitmask also?
            data = self.tobytes()
            size = self.size
            zeros = (math.ceil(len(data) / __chunk_size__) * __chunk_size__) - len(data)
            buf = bytes(data) + (zeros * b"\x00")
            # create the memory object
            memory = ebc.allocate_memory_from_bytes(buf)
            # use that memory to create the bitmask
            obj = ebc.create_EnhancedBitmask_from_EnhancedBitmaskMemory(size, memory)
        else:
            arr = np.copy(self)
            obj = arr.view(EnhancedBitmask)
            obj.__update_internal_mask_and_bit_count__()
        return obj

    def intersection(self, other):
        """find the intersection between self and other enhanced bitmasks.
        This uses the internal object.bitmask to calculate this much faster
        initially.  If it finds an intersection it will get the true intersection."""
        self.__debug_print__("eb intersection called")
        intersection = self.bitmask & other.bitmask
        if intersection.any():
            # we cannot just and the two arrays
            # value in one could be 1 and ther other could be 2
            # 1 & 2 = 0, so we must flatten them to bools first
            # then convert back to an eb array.
            a = self.to_bool_array()
            b = other.to_bool_array()
            real_intersection = eb.array(a & b)
            # real_intersection = self & other
        else:
            real_intersection = None
        return real_intersection

    def intersection_any(self, other: EnhancedBitmaskLike):
        """find the intersection between self and other enhanced bitmasks.
        If it finds an intersection it will true."""
        self.__debug_print__("eb intersection_any called")
        if (
            isinstance(other, EnhancedBitmaskLike)
            and self.bit_count > 0
            and other.bit_count > 0
            and (self.bit_idx_last < other.bit_idx_first or self.bit_idx_first > other.bit_idx_last)
        ):
            result = False
        else:
            # 4x slower
            # intersection = self & other
            # if intersection.bit_count == 0:
            #     result = False
            # else:
            #     result = True
            intersection = self.bitmask & other.bitmask
            if intersection.any():
                result = True
            else:
                result = False
        # superprint(f"intersection bit_count:{intersection.bit_count}")
        # result = intersection.any()
        return result

    def intersection_all(self, other):
        self.__debug_print__("eb intersection_all called")
        return super().intersection_all(other)

    def is_subset(self, other):
        """is self a subset of other."""
        self.__debug_print__("eb is_subset called")
        # result = (other & self == self).all()
        # assert result == self.logical_is_subset(other, self)
        # assert result == self.logical_is_subset2(other, self)
        result = self.logical_is_subset(other, self)
        return result

    @classmethod
    def logical_is_subset(cls, left, right):
        """
        Is right a subset of left?

        Inhib logic uses for determining subset
            is bitmask_right a subset of bitmask_left -> result
        a - b
        0   0   0
        0   1   0
        1   0   1
        1   1   0
        To get the reverse/subset, switch a and b, thus right then left.
        Inhib is the following:
        ¬(P⇒Q)
        Using Demorgan's logic
        ¬(¬a ∨ b)
        a ^ ¬b
        Then, if no bits are set, it's a subset, thus the .any()
        """
        cls.__debug_print__("eb logical_is_subset called")
        # slow because of |
        # result = (eb.logical_not(right) | left).all()
        # super slow because of -
        # result = (right - (right & left)).none()
        # slow because of & but faster than |
        # result = (right & eb.logical_not(left)).none()
        # slow because of the logical_not
        # result = eb.logical_and_none(right, eb.logical_not(left))
        result = eb.logical_and_none(right, left.not_and_cache())
        return result

    @classmethod
    def logical_and_none(cls, left, right):
        cls.__debug_print__("eb logical_and_none called")
        if left.bit_idx_last < right.bit_idx_first:
            result = True  # result is true if they won't and together
        elif left.bit_idx_first > right.bit_idx_last:
            result = True
        else:
            # result = (left & right).none()
            result = not (left.bitmask & right.bitmask).any()
        return result

    @classmethod
    def logical_is_subset2(cls, left, right):
        """
        Is right a subset of left?

        Inhib logic uses for determining subset
            is bitmask_right a subset of bitmask_left -> result
        a - b
        0   0   0
        0   1   0
        1   0   1
        1   1   0
        To get the reverse/subset, switch a and b, thus right then left.
        Inhib is the following:
        ¬(P⇒Q)
        Using Demorgan's logic
        ¬(¬a ∨ b)
        a ^ ¬b
        Then, if no bits are set, it's a subset, thus the .any()
        """
        cls.__debug_print__("eb logical_is_subset called")
        result = (right & eb.logical_not(left)).any()
        if not result:
            result = True
        else:
            result = False
        return result

    @classmethod
    def logical_is_subset_slow(cls, left, right):
        """
        Is right a subset of left?
        """
        cls.__debug_print__("eb logical_is_subset2 called")
        result = functools.reduce(lambda a, b: a or b, map(bool_is_subset, zip(left, right)))
        return not result

    def is_superset(self, other):
        """is self a subset of other."""
        self.__debug_print__("eb is_superset called")
        result = (self | other == self).all()
        return result

    def not_and_cache(self):
        """calculate and cache the not of myself for use."""
        if self.__not_cache is None:
            self.__not_cache = eb.logical_not(self)
        return self.__not_cache

    def to_bool_array(self):
        """provides a way to cache the bool of the array.
        returns a numpy array, not a Enhanced Bitmask"""
        if self.__bool_array is None:
            # since these are flattening the array to a bool, we can return a numpy array.
            self.__bool_array = np.array(self, dtype=bool)
        return self.__bool_array

    def sum(self, axis=None, dtype=None, out=None, keepdims=False, **kwargs):
        self.__debug_print__("eb sum called")
        if not dtype:
            dtype = __base_dtype__
        obj = super().sum(axis=axis, dtype=dtype, out=out, keepdims=keepdims, **kwargs)
        return obj

    def __getitem__(self, item):
        # self.__debug_print__('__getitem__:', item)
        obj = super().__getitem__(item)
        if not isinstance(self, EnhancedBitmask):
            return obj
        if isinstance(item, slice):
            obj.__update_internal_mask_and_bit_count__()
        return obj

    @staticmethod
    def shrink(obj, dest_size):
        """Given an enhanced bitmask, shrink it.  This group the bitmask
        according to the dest_size and the result will now be a series of
        1's, not a sum of the groupings"""
        result = None
        if dest_size < len(obj):
            # let make sure we do this nicely.
            assert len(obj) % dest_size == 0
            offset = int(len(obj) / dest_size)
            _result = eb.zeros(dest_size)
            # lets walk forward in chunks
            for idx in range(0, _result.length):
                chunk = obj[int(idx * offset) : int(idx * offset) + offset]
                _result[idx] = chunk.any()
            result = _result
        elif dest_size > len(obj):
            raise Exception("Please use kron to grow the array")
        else:  # equal
            result = obj.copy()
        return result

    # @staticmethod
    # def apply_along_axis(func1d, axis, arr, *args, **kwargs):
    #     arr = np.apply_along_axis(func1d, axis, arr, *args, **kwargs)
    #     obj = arr.view(EnhancedBitmask)
    #     return obj

    def __setitem__(self, idx, value):
        EnhancedBitmask.__debug_print__("__setitem__ ", idx, value)
        # very specific to the bitmask.
        # FIXME: we need to handle slice and mask going in
        # https://docs.python.org/2.3/whatsnew/section-slices.html

        if isinstance(value, numbers.Number):
            check_value_overflow(value, debug=self.__debug_flag__)
            check_value_underflow(value, debug=self.__debug_flag__)
        elif isinstance(value, List):
            list(map(lambda v: check_value_underflow(v, debug=self.__debug_flag__), value))
            list(map(lambda v: check_value_overflow(v, debug=self.__debug_flag__), value))
        elif isinstance(value, Iterable):
            check_eb_overunderflow(value, debug=self.__debug_flag__)
        else:
            check_value_overflow(value, debug=self.__debug_flag__)
            check_value_underflow(value, debug=self.__debug_flag__)
        # this may have allocated new memory...  What are we going to do!
        # np.ndarray.__setitem__(self, idx, value)
        super().__setitem__(idx, value)
        self.__update_internal_mask_and_bit_count__()
        return None

    def __sub__(self, other):
        EnhancedBitmask.__debug_print__(f"__sub__ called! ", other)
        result = np.ndarray.__sub__(self, other)
        return result

    def __add__(self, other):
        EnhancedBitmask.__debug_print__(f"__add__ called! ", other)
        # result = np.ndarray.__add__(self, other)
        result = magnitude_add(self, other)
        return result

    def __mul__(self, other):
        EnhancedBitmask.__debug_print__(f"__mul__ called!")
        result = np.ndarray.__mul__(self, other)
        return result

    def __divmod__(self, other):
        EnhancedBitmask.__debug_print__(f"__divmod__ called!", other)
        result = np.ndarray.__divmod__(self, other)
        return result

    def __invert__(self):
        EnhancedBitmask.__debug_print__("__invert__ called!")
        # FIXME: this may not do what you expect!!
        #   It should probably take the sign of self and invert that and convert back to the self.dtype
        result = np.ndarray.__invert__(self)
        return result

    def internal_bitmask_any(self):
        return uint64_bitmask_any(self.bitmask)

    def internal_bitmask_all(self):
        return uint64_bitmask_all(self.bitmask, self.length)

    def __and__(self, other):
        # arr = super(EnhancedBitmask, self).__and__(other)
        # obj = arr.view(EnhancedBitmask)
        # return obj
        self.__debug_print__(f"eb __and__ called")
        bit_idx_first = None
        bit_idx_last = None
        update = True
        if type(other) == EnhancedBitmask and self.bit_idx_last < other.bit_idx_first and self.bit_idx_first > other.bit_idx_last:
            # all zeros
            arr = _get_np_prepared_bitmask(self.dtype, self.length, debug=self.__debug_flag__)
            bit_idx_first = self.length
            bit_idx_last = -1
        else:
            try:
                # anded = self.bitmask & other.bitmask
                anded = np.bitwise_and(self.bitmask, other.bitmask)
            except AttributeError:  # a numpy array
                # arr = super(EnhancedBitmask, self).__and__(other)
                arr = np.bitwise_and(self, other)
            else:
                is_any = uint64_bitmask_any(anded)
                if is_any:
                    if uint64_bitmask_all(anded, self.length):
                        arr = _get_np_prepared_bitmask(
                            self.dtype,
                            self.length,
                            ones=True,
                            debug=self.__debug_flag__,
                        )
                    else:  # Some
                        # FIXME: make faster
                        # arr = np.bitwise_and(self, other)
                        arr = super(EnhancedBitmask, self).__and__(other)
                        update = False
                else:
                    # all zeros
                    arr = _get_np_prepared_bitmask(self.dtype, self.length, debug=self.__debug_flag__)
                    bit_idx_first = self.length
                    bit_idx_last = -1
        if type(arr) != EnhancedBitmask:
            obj = arr.view(EnhancedBitmask)
        else:
            obj = arr
        if update:
            obj.__update_internal_mask_and_bit_count__(bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last)
        return obj

    def view(self, dtype=None, *args, **kwargs):
        EnhancedBitmask.__debug_print__("view enter")
        obj = super().view(dtype=dtype, *args, **kwargs)
        EnhancedBitmask.__debug_print__("view exit")
        return obj

    def __or__(self, other):
        # arr = super(EnhancedBitmask, self).__or__(other)
        # obj = arr.view(EnhancedBitmask)
        # return obj
        self.__debug_print__("eb __or__ called")
        try:
            ored = self.bitmask | other.bitmask
        except AttributeError:  # a numpy array
            arr = super(EnhancedBitmask, self).__or__(other)
            obj = arr  # obj = arr.view(EnhancedBitmask)
        else:
            is_any = uint64_bitmask_any(ored)
            if is_any:
                if uint64_bitmask_all(ored, self.length):
                    obj = _get_np_prepared_bitmask(self.dtype, self.length, ones=True, debug=self.__debug_flag__)
                    # obj = arr.view(EnhancedBitmask)
                    # at this point the obj.bitmask is not set.
                else:
                    arr = super(EnhancedBitmask, self).__or__(other)
                    obj = arr  # obj = arr.view(EnhancedBitmask)
            else:
                # this is basically zeros
                obj = _get_np_prepared_bitmask(self.dtype, self.length, debug=self.__debug_flag__)
                # obj = arr.view(EnhancedBitmask)
        obj.__update_internal_mask_and_bit_count__()
        return obj

    def compareable(self, other):
        result = False
        if type(self) == type(other):
            if self.bit_count == other.bit_count:
                try:
                    if self.__version_class__ == other.__version_class__:
                        if self.__chunk_size__ == other.__chunk_size__:
                            result = True
                except AttributeError:
                    pass
        return result

    def __eq__(self, other):
        self.__debug_print__("eb __eq__ called")
        if self.__direct_equal__:
            compareable = self.compareable(other)
            if compareable:
                result = self.intersection_all(other)
            else:
                result = super(EnhancedBitmask, self).__eq__(other)
        else:
            result = super(EnhancedBitmask, self).__eq__(other)
        return result

    @classmethod
    def intersection_possible(cls, bitmaska, bitmaskb):
        if bitmaska.length != bitmaskb.length:
            result = False
        elif bitmaska.bit_idx_first > bitmaskb.bit_idx_last:
            result = False
        elif bitmaska.bit_idx_last < bitmaska.bit_idx_first:
            result = False
        else:
            result = True
        return result

    def equal_all(self, other) -> bool:
        """
        only works on EnhancedBitmasks
        """
        self.__debug_print__("eb equal_all called")
        if self.intersection_possible(self, other):
            equality = self.bitmask == other.bitmask
            if type(equality) == bool:
                result = equality
            else:
                result = equality.all()
        else:
            result = False
        return result

    def equal_any(self, other) -> bool:
        """
        only works on EnhancedBitmasks
        """
        self.__debug_print__("eb equal_any called")
        if self.intersection_possible(self, other):
            equality = self.bitmask == other.bitmask
            if type(equality) == bool:
                result = equality
            else:
                result = equality.any()
        else:
            result = False
        # result2 = (self.bitmask == other.bitmask).any()
        # assert result == result2
        return result

    def __update_bit_idx_first_and_last__(self, bit_idx_first=None, bit_idx_last=None) -> Tuple[int, int]:
        self.__debug_print__(f"eb __update_bit_idx_first_and_last__ called on {id(self)=} {bit_idx_first=} {bit_idx_last=}")
        int_lst = self.bitmask
        if not bit_idx_first:
            bit_idx_first = self.length
            depth = 0
            for value in int_lst:
                if value != 0:
                    for sub_idx, one_bit_set in uint64_bits_forward:
                        if value & one_bit_set:
                            bit_idx_first = sub_idx + depth
                            break
                    if bit_idx_first is not None:
                        break
                depth = depth + 64
        if not bit_idx_last:
            bit_idx_last = -1
            depth = (len(int_lst) - 1) * 64
            for i in range(-1, -len(int_lst) - 1, -1):
                value = int_lst[i]  # this will be negative, going backwards
                if value != 0:
                    for sub_idx, one_bit_set in uint64_bits_backward:
                        if value & one_bit_set:
                            bit_idx_last = sub_idx + depth
                            break
                    if bit_idx_last is not None:
                        break
                depth = depth - 64
        return bit_idx_first, bit_idx_last

    def __update_internal_mask_and_bit_count__(self, bit_idx_first: Optional[int] = None, bit_idx_last: Optional[int] = None):
        """
        This function updates the internal mask and bit_count
        """
        self.__debug_print__(f"eb __update_internal_mask_and_bit_count__ called on {id(self)=}")
        chunk_size: int = self.__chunk_size__
        chunk_count: int = self.__chunk_count__
        # FIXME: at this point during the mul, the object has memory, which is the same as the other object!!!
        bit_count, int_lst = update_internal_mask_and_bit_count(self, chunk_size, chunk_count)
        self.bit_count = bit_count
        if hasattr(self, "bitmask") and hasattr(self.bitmask, "memory"):
            memory = self.bitmask.memory
        else:
            memory = None
        self.bitmask = bitmask_create_uint64_from_int_lst(memory, int_lst)
        self.length = self.size
        self.__bool_array = None
        bit_idx_first, bit_idx_last = self.__update_bit_idx_first_and_last__(
            bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last
        )
        self.bit_idx_first = bit_idx_first
        self.bit_idx_last = bit_idx_last
        self.__not_cache = None

    @classmethod
    def join_bitmasks(cls, bitmask_lst: list, operation) -> EnhancedBitmaskLike:
        """
        given a list of bitmasks, join them
        """
        if len(bitmask_lst) == 0:
            raise Exception("Must provide a bitmask to join with.  The length is required.")
        elif len(bitmask_lst) == 1:
            result_bitmask = bitmask_lst[0].copy()
        else:
            length = 0
            # ordering them by bit_count
            if operation in [cls.logical_and, cls.logical_or]:
                bitmask_lst_orig = sorted(bitmask_lst, key=lambda obj: -obj.bit_count)
                first_orig_bitmask = bitmask_lst_orig[0]
                if operation == cls.logical_or and first_orig_bitmask.length == first_orig_bitmask.bit_count:
                    result_bitmask = first_orig_bitmask.copy()
                else:
                    for bitmask in bitmask_lst_orig:
                        if length == 0:
                            length = bitmask.length
                        if bitmask.length != length:
                            raise AssertionError("Cannot join bitmasks with different sizes")
                    if operation == cls.logical_or:
                        arr_lst = [bitmask.view(np.ndarray) for bitmask in bitmask_lst_orig]
                        idx_lst = []
                        for bitmask in arr_lst:
                            idx_array = np.where(bitmask == 1)
                            assert len(idx_array) == 1
                            idx_lst.extend(idx_array[0].tolist())
                        idx_lst = sorted(list(set(idx_lst)))
                        result_bitmask = cls.zeros_and_set_bits(length, idx_lst)
                    # elif operation == cls.logical_and:
                    else:
                        arr_lst = [bitmask.view(np.ndarray) for bitmask in bitmask_lst_orig]
                        # we want the least number of bits first as it will skip processing the next set.
                        arr_lst.reverse()
                        result_bitmask = arr_lst[0]
                        for idx, bit in enumerate(result_bitmask):
                            result_value = result_bitmask[idx]
                            if result_value == 1:
                                for bitmask in arr_lst[1:]:
                                    if bitmask[idx] != 1:
                                        result_bitmask[idx] = 0
                                        break
                        result_bitmask = cls.array(result_bitmask)
            else:
                arr_lst = [bitmask.view(np.ndarray) for bitmask in bitmask_lst]
                result_bitmask = cls.zeros(length)
                for bitmask in arr_lst:
                    result_bitmask = operation(result_bitmask, bitmask)
        return result_bitmask


eb = EnhancedBitmask


def bitmask_create_uint64_from_int_lst(memory, int_lst):
    if not memory:
        memory = ebc.allocate_memory_for_uint64(len(int_lst))
    bitmask = EnhancedBitmask64(len(int_lst), dtype=uint64, memory=memory)
    # build and/or update the bitmask memory.
    offset = 8
    for idx, value in enumerate(int_lst):
        # superprint(f"setting bytes, {offset*idx=}:{offset*(idx+1)} => {value}/{int.to_bytes(value, length=8, byteorder='little')}")
        memory.buf[offset * idx : offset * (idx + 1)] = int.to_bytes(value, length=8, byteorder="little")
    return bitmask


def update_internal_mask_and_bit_count(bitmask: EnhancedBitmask, chunk_size: cython.int, chunk_count: cython.int):
    """
    This generates the bit_count and int_lst for the internal uint64 bitmask.

    given a numpy array, walk through the array, in chunks and get the 64bit value from each
    chunk.  This _value will be accompanied by the _bit_count and _length.
    The _value is a 64bit unsigned int(numpy) that comes from 64 individual bits of the array
    The _bit_count is the total number of bits set
    The _length is the length of the array, which may not be 64, but processing the array needs
    to push the bits to the left side as the _value is indexed from left to right.
    """
    bit_count: cython.int = 0
    if hasattr(bitmask, "memory") and type(bitmask) == EnhancedBitmask:
        arr_memory = bitmask.memory
        arr = bitmask.view(np.ndarray)
    elif type(bitmask) == EnhancedBitmask:
        arr = bitmask.view(np.ndarray)
        arr_memory = None
    else:
        arr_memory = None
        arr = bitmask
    dtype = arr.dtype
    shape = arr.shape
    # we know how big int_lst is going to be, we could prealloc and set, then convert to a list on the way out.
    int_lst: list = []
    int_lst_append = int_lst.append
    if dtype == int16:
        chunk_step = 2
        big_idx: cython.int
        if shape != ():
            if arr_memory is not None:
                memory = arr_memory
            else:
                # FIXME: this is slow, requires alloc
                binfull = arr.tobytes()
                memory = ebc.allocate_memory_for_int16(arr.size)
                memory.buf[:] = binfull
            # if eb.__parallel__ is True:
            #    data = ebc.pool.starmap(update_internal_mask_and_bit_count_divide_memory, [(memory, big_idx, chunk_size, chunk_step) for big_idx in range(chunk_count)])
            # else:
            data = [
                update_internal_mask_and_bit_count_divide_memory(memory, big_idx, chunk_size, chunk_step)
                for big_idx in range(chunk_count)
            ]
            bit_count = 0
            for _value, _bit_count in data:
                bit_count = bit_count + _bit_count
                int_lst_append(_value)
            bit_count_old = functools.reduce(lambda l, r: l + r, [_bit_count for _value, _bit_count in data])
            int_lst_old = [_value for _value, _bit_count in data]
            assert bit_count == bit_count_old
            assert int_lst == int_lst_old
        else:
            # FIXME: don't know how to get here, sum???...
            binfull = arr.tobytes()
            _value, _bit_count = update_internal_mask_and_bit_count_inner_int16(binfull, chunk_size, chunk_step)
            bit_count += _bit_count
            int_lst_append(_value)
    else:
        chunk_step = 1
        big_idx: cython.int
        for big_idx in range(0, chunk_count):
            big_idx_mul_chunk_size: int = big_idx * chunk_size
            start = big_idx_mul_chunk_size
            end = big_idx_mul_chunk_size + chunk_size
            if arr.shape == ():
                chunk = arr
            else:
                chunk = arr[start:end]
            binchunk = chunk.tobytes()
            # the _value cannot be greater than an uint64, thus chunk_size should be 64
            if binchunk == __zero_chunk__:
                _value, _bit_count = 0, 0
            else:
                _value, _bit_count = process_np_chunk(binchunk, chunk_size, chunk_step)
            bit_count += _bit_count
            int_lst_append(_value)
    return bit_count, int_lst


@contextlib.contextmanager
def direct_compare(*args, **kwargs):
    try:
        eb.__direct_equal__ = True
        yield None
    finally:
        eb.__direct_equal__ = False


_prepared_bitmasks: Dict[Tuple[np.generic, int, bool], EnhancedBitmask] = dict()
ebc = EnhancedBitmaskContext(processes=16)
prepared_bitmasks = _prepared_bitmasks


def shutdown():
    ebc.shutdown()


atexit.register(shutdown)
