# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

# cython: language_level=3, boundscheck=True
import copy
from pprint import pformat

import numpy as np
from collections import deque
from dataclasses import dataclass
from typing import Dict, Protocol, Union
from typing import List, Optional

import networkx as nx

from Octeres.bitmask import EnhancedBitmask, EnhancedBitmaskLike
from Octeres.bitmask.bitmask_graph import node_difference
from Octeres.bitmask.bitmask_lite import LiteBitmask
from Octeres.util import superprint


@dataclass
class SearchStats:
    total_nodes: int = 0  # total nodes
    total_edges: int = 0  # total number of edges
    search_complete: bool = False

    def show_stats(self):
        superprint(" --- SearchStats ---")
        superprint(f"search_complete:{self.search_complete}")
        if self.search_complete:
            superprint(f"total_nodes:{self.total_nodes}")
        superprint(f"total_edges:{self.total_edges}")

    def increment_total_edges_searched(self, count):
        self.total_edges += count

    def increment_total_nodes_searched(self, count):
        self.total_nodes += count


class SearchFunc(Protocol):
    def __call__(
        self,
        parent_name: Optional[str],
        child_name: str,
        result: List[Union[str, dict]],
        debug: bool = False,
    ) -> bool: ...


def depth_first_search(
    root_name: str,
    graph: nx.DiGraph,
    stat: SearchStats,
    head_func: SearchFunc,
    tail_func=None,
    stop_when_found=False,
    debug=False,
) -> List[str]:
    if stop_when_found and tail_func:
        raise Exception("stop_when_found and tail_func are not supported together.")
    result: List[str] = []
    tail_queue = deque()
    if tail_func:
        tail_queue.append(root_name)
    visited = {root_name}
    root_dct = graph.nodes[root_name]
    # search stats vars
    stat.search_complete = False
    stop_when_found_trigger = False
    stat.increment_total_nodes_searched(1)
    func_result = head_func(None, root_name, result, debug=debug)
    if stop_when_found and func_result:
        stop_when_found_trigger = True
    if not stop_when_found_trigger:
        for node_name, child_name in nx.dfs_edges(graph, source=root_name):
            node_dct = graph.nodes[node_name]
            child_dct = graph.nodes[child_name]
            stat.increment_total_nodes_searched(1)
            stat.increment_total_edges_searched(1)
            func_result = head_func(node_name, child_name, result, debug=debug)
            if stop_when_found and func_result:
                stop_when_found_trigger = True
                break
            if node_name not in visited:
                visited.add(node_name)
                if tail_func:
                    tail_queue.append(node_name)

        if not stop_when_found_trigger:
            if tail_func:
                while tail_queue:
                    node_name = tail_queue.pop()
                    tail_func(node_name, result, debug=debug)
            stat.search_complete = True
    return result


# FIXME: SearchFunc doesn't work for head_func and tail_func, the signature is not the same.


def breadth_first_search(
    root_name: str,
    graph: nx.DiGraph,
    stat: SearchStats,
    head_func: SearchFunc,
    tail_func: Optional[SearchFunc] = None,
    stop_when_found=False,
    sort_neighbors=None,
    tail_retry_count=0,
    debug=False,
) -> List[str]:
    """
    Execute a breadth first search of the graph.

    Execute the head_func on entry to that node and pass results to the head_func.
    Execute the tail_func on return and pass results to the tail_func

    :param root_name: the root of the graph.
    :param graph: the graph
    :param stat: the SearchStats object
    :param head_func: search_func(self, parent_name: Optional[str], node_name: str, results: List[str], debug=False) -> bool:
    :param tail_func: tail_func(self, node_name: str, results: List[str], debug=False):
    :param stop_when_found:  stops when search finds something
    :param sort_neighbors:  directly sent into nx.bfs_edges, see details there
    :param tail_retry_count:  attempt count for each tail attempt to process
    :param debug: debug or not
    :return: results
    """
    if stop_when_found and tail_func:
        raise Exception("stop_when_found and tail_func are not supported together.")
    results: List[Union[str, dict]] = []
    tail_queue = deque()
    if tail_func:
        tail_queue.append(root_name)
    visited = {root_name}

    root_dct = graph.nodes[root_name]
    # search stats vars
    stat.search_complete = False
    stop_when_found_trigger = False
    stat.increment_total_nodes_searched(1)
    func_result = head_func(None, root_name, results, debug=debug)
    if stop_when_found and func_result:
        stop_when_found_trigger = True
    if not stop_when_found_trigger:
        for node_name, child_name in nx.bfs_edges(graph, source=root_name, sort_neighbors=sort_neighbors):
            # node_dct = graph.nodes[node_name]
            # child_dct = graph.nodes[child_name]
            stat.increment_total_nodes_searched(1)
            stat.increment_total_edges_searched(1)
            func_result = head_func(node_name, child_name, results, debug=debug)
            if stop_when_found and func_result:
                stop_when_found_trigger = True
                break
            # if node_name not in visited:
            #     visited.add(node_name)
            #     tail_queue.append(node_name)
            if child_name not in visited:
                visited.add(child_name)
                if tail_func:
                    tail_queue.append(child_name)

        tail_retry_lookup = {}
        if not stop_when_found_trigger:
            if tail_func:
                while tail_queue:
                    node_name = tail_queue.pop()
                    try:
                        tail_func(node_name, results, debug=debug)
                    except ErrorBitmaskSearchOutofOrder as err:
                        if debug:
                            print(f"processing ErrorBitmaskSearchOutofOrder:{str(err)}")
                        # We hit an error.  This could be because a bfs when going out is correct,
                        # but the return may go to a higher level due to 2 branches that go down
                        # different paths with different hops, like a -> b, a -> c, b -> c may return
                        node_count = tail_retry_lookup.get(node_name, 0)
                        # the reason for this catch
                        if node_count >= tail_retry_count:
                            raise
                        tail_queue.appendleft(node_name)
                        node_count += 1
                        tail_retry_lookup[node_name] = node_count
            stat.search_complete = True
    return results


class ErrorBitmaskSearchOutofOrder(Exception):
    pass


class BitmaskApplier:
    def __init__(
        self,
        root_name: str,
        graph: nx.DiGraph,
        bit_total,
        is_bitmask_indexable,
        bitmask_class: type[EnhancedBitmaskLike],
    ):
        """apply a bitmask to a graph.
        bit_total is a simple check that the graph is being indexed correctly.
        is_bitmask_indexable is usually simple via node != root -> true, but
        if you have branches that turn into leaves that you don't want to be leaves, provide
        a different is_bitmask_indexable function."""
        self.graph = graph
        self.root_name = root_name
        self.desired_bit_total = bit_total
        self.bit_total = -1
        self.current_index_leaf = 0
        self.is_bitmask_indexable = is_bitmask_indexable
        self.bitmask_class = bitmask_class
        # self.graph.in_edges(node_name)  # parents
        # self.graph.edges(node_name)  # children
        # self.graph.nodes[node_name]  # node
        # self.graph.nodes[node_name]['bitmask']  # node attr
        #                            ['idx']
        #                            ['obj_type']
        #                            ['is_schedulable']

    def search_func(
        self,
        parent_name: Optional[str],
        node_name: str,
        results: List[str],
        debug=False,
    ) -> bool:
        """if it's indexable, and there are in_edges and no out_edges, index it."""
        if self.is_bitmask_indexable(self.root_name, node_name):
            if self.graph.in_edges(node_name):
                if not self.graph.edges(node_name):
                    self.graph.nodes[node_name]["idx"] = self.current_index_leaf
                    self.current_index_leaf += 1
        if debug is True:
            node_dct = self.graph.nodes[node_name]
            superprint(f"{node_name=} {node_dct=} {self.current_index_leaf=}")
        return False

    def tail_func(self, node_name: str, results: List[str], debug=False):
        self.bit_total = self.current_index_leaf
        node_dct: dict = self.graph.nodes[node_name]
        if debug is True:
            superprint(f"tail_func(in): {node_name:{' '}>20} {node_dct} {self.current_index_leaf=}")
        # assert self.bit_total == self.desired_bit_total, f"{self.bit_total} != {self.desired_bit_total}"
        if "idx" in node_dct:  #  if I have been indexed, we start the bitmask from me.
            # I am the bottom
            nxnode_idx = node_dct["idx"]
            bitmask: EnhancedBitmaskLike = self.bitmask_class.zeros_and_set_bits(self.bit_total, [nxnode_idx])
            bitmask.flags.writeable = False
            node_dct["bitmask"] = bitmask
            # assert self.bit_total == len(bitmask)
        else:
            bitmask_lst = []
            # loop through my children and grab their bitmasks.
            edges_to_children = self.graph.edges(node_name)
            for u, child_name in edges_to_children:
                child_dct: dict = self.graph.nodes[child_name]  # {'obj_type': 'tensor', 'idx': 7}
                try:
                    bitmask_lst.append(child_dct["bitmask"])
                except KeyError:
                    raise ErrorBitmaskSearchOutofOrder(f"{u}->{child_name} unable to find bitmask in {pformat(child_dct)}")
            # if there is an idx_set==1 and no bitmask_lst
            if len(bitmask_lst) == 1:
                # I am not the bottom, but I have one child.  Grab their bitmask
                bitmask = bitmask_lst[0]
            elif len(bitmask_lst) > 1:
                # I have children, gather their bitmasks.
                bitmask = self.bitmask_class.join_bitmasks(bitmask_lst, self.bitmask_class.logical_or)
            else:
                # pruning tree
                self.graph.remove_node(node_name)
                bitmask = None
            if bitmask is not None:
                bitmask.flags.writeable = False
                node_dct["bitmask"] = bitmask
                # assert self.bit_total == node_dct["bitmask"].length, (
                #     f"bit_total:{self.bit_total} != " f"len(bitmask):{node_dct['bitmask'].length}"
                # )
        if debug is True:
            superprint(f"tail_func(out):{node_name:{' '}>20} {node_dct}")


def is_bitmask_indexable(root_name, node_name):
    """this is used to allow a node to possibly be indexed for a bitmask.
    It still must have no out_edges."""
    if root_name == node_name:
        result = 0
    else:
        result = 1
    return result


def apply_bitmask_to_nxgraph(
    root_name,
    bit_total,
    graph: nx.DiGraph,
    is_bitmask_indexable_func,
    bitmask_class=EnhancedBitmask,
    debug=False,
    sort_neighbors=None,
    tail_retry_count=0,
) -> nx.DiGraph:
    stat = SearchStats()
    graph = copy.deepcopy(graph)
    ba = BitmaskApplier(root_name, graph, bit_total, is_bitmask_indexable_func, bitmask_class)
    breadth_first_search(
        root_name,
        graph,
        stat,
        ba.search_func,
        tail_func=ba.tail_func,
        stop_when_found=False,
        sort_neighbors=sort_neighbors,
        tail_retry_count=tail_retry_count,
        debug=debug,
    )
    return graph


class GraphComparison:
    def __init__(self, graph_a: nx.DiGraph, graph_b: nx.DiGraph):
        self.graph_a = graph_a
        self.graph_b = graph_b
        self.required_comparisons = [
            "missing_nodes_a",
            "missing_nodes_b",
            "missing_edges_a",
            "missing_edges_b",
            "node_difference",
            "isomorphic",
            "bitmask",
        ]
        self.comparisons = []
        self.missing_nodes_a: List[str] = []
        self.missing_nodes_b: List[str] = []
        self.missing_edges_a: List[str] = []
        self.missing_edges_b: List[str] = []
        self.isomorphic = None
        self.node_difference_dct: Dict[str, list] = {}

    # def __repr__(self):
    #     return self.is_identical()

    def compare_nx_graph(self):
        graph_a, graph_b = self.graph_a, self.graph_b
        self.isomorphic = nx.is_isomorphic(graph_a, graph_b)
        self.comparisons.append("isomorphic")

    def compare_nodes(self, compare_edges=True):
        nxgraph_a, nxgraph_b = self.graph_a, self.graph_b
        missing_nodes_a: List[str] = []
        missing_nodes_b: List[str] = []
        node_difference_dct: Dict[str, List[str]] = {}

        node_name_a: str
        for node_name_a in nxgraph_a.nodes:
            node_a_dct = nxgraph_a.nodes[node_name_a]
            if node_name_a not in nxgraph_b:
                missing_nodes_b.append(node_name_a)
            else:
                # node_b_dct = nxgraph_b.nodes[node_name_a]
                differences = node_difference(
                    nxgraph_a,
                    node_name_a,
                    nxgraph_b,
                    node_name_a,
                    compare_edges=compare_edges,
                )
                if differences:
                    node_difference_dct[node_name_a] = differences

        node_name_b: str
        for node_name_b in nxgraph_b.nodes:
            # node_b_dct = nxgraph_b.nodes[node_name_b]
            if node_name_b not in nxgraph_a:
                missing_nodes_a.append(node_name_b)

        self.missing_nodes_a = missing_nodes_a
        self.missing_nodes_b = missing_nodes_b
        self.node_difference_dct = node_difference_dct
        self.comparisons.append("missing_nodes_a")
        self.comparisons.append("missing_nodes_b")
        self.comparisons.append("node_difference")
        self.comparisons.append("bitmask")

    def compare_edges(self):
        nxgraph_a, nxgraph_b = self.graph_a, self.graph_b
        missing_edge_a: List[str] = []
        missing_edge_b: List[str] = []

        for edge_name_a in nxgraph_a.edges:
            edge_a = nxgraph_a.edges[edge_name_a]
            if not nxgraph_b.has_edge(*edge_name_a):
                missing_edge_b.append(edge_a)

        for edge_name_b in nxgraph_b.edges:
            edge_b = nxgraph_b.edges[edge_name_b]
            if not nxgraph_a.has_edge(*edge_name_b):
                missing_edge_a.append(edge_b)

        self.missing_edges_a = missing_edge_a
        self.missing_edges_b = missing_edge_b
        self.comparisons.append("missing_edges_a")
        self.comparisons.append("missing_edges_b")

    def get_comparison_incomplete_lst(self):
        incomplete_lst = []
        for required_check in self.required_comparisons:
            if required_check not in self.comparisons:
                incomplete_lst.append(required_check)
        return incomplete_lst

    def is_identical(self):
        identical = False
        comparison_incomplete_lst = self.get_comparison_incomplete_lst()
        if len(comparison_incomplete_lst) == 0:
            if (
                len(self.missing_nodes_a) == 0
                and len(self.missing_nodes_b) == 0
                and len(self.missing_edges_a) == 0
                and len(self.missing_edges_b) == 0
                and self.isomorphic is True
            ):
                identical = True
        return identical

    def summary(self, full=False):
        identical = self.is_identical()
        comparison_incomplete_lst = self.get_comparison_incomplete_lst()
        if full:
            difference_summary = {
                "missing_nodes_a": self.missing_nodes_a,
                "missing_nodes_b": self.missing_nodes_b,
                "missing_edges_a": self.missing_edges_a,
                "missing_edges_b": self.missing_edges_b,
                "node_difference_dct": self.node_difference_dct,
                "isomorphic": self.isomorphic,
                "identical": identical,
                "comparison_incomplete_lst": comparison_incomplete_lst,
            }
        else:
            difference_summary = {
                "len(missing_nodes_a)": len(self.missing_nodes_a),
                "len(missing_nodes_b)": len(self.missing_nodes_b),
                "len(missing_edges_a)": len(self.missing_edges_a),
                "len(missing_edges_b)": len(self.missing_edges_b),
                "node_difference_dct": self.node_difference_dct,
                "isomorphic": self.isomorphic,
                "identical": identical,
                "comparison_incomplete_lst": comparison_incomplete_lst,
            }
        return difference_summary


def compare_graphs(graph_a: nx.DiGraph, graph_b: nx.DiGraph) -> GraphComparison:
    gc = GraphComparison(graph_a, graph_b)
    gc.compare_nodes()
    gc.compare_edges()
    gc.compare_nx_graph()
    comparison_incomplete_lst = gc.get_comparison_incomplete_lst()
    if len(comparison_incomplete_lst) > 0:
        raise Exception(f"incomplete compare")
    return gc


def bitmask_func_true(
    node: str,
    bitmask_left: EnhancedBitmask,
    bitmask_right: EnhancedBitmask,
    debug=False,
):
    result = True
    if debug:
        superprint(f"bitmask testing {node} with bitmask_func_true -> {result}")
    return result


def bitmask_func_intersection_any(
    node: str,
    bitmask_left: EnhancedBitmask,
    bitmask_right: EnhancedBitmask,
    debug=False,
) -> bool:
    result = bitmask_left.intersection_any(bitmask_right)
    if debug:
        superprint(f"bitmask testing {node} with bitmask_func_intersection_any -> {result}")
    return result


def bitmask_func_intersection_all(
    node: str,
    bitmask_left: EnhancedBitmask,
    bitmask_right: EnhancedBitmask,
    debug=False,
) -> bool:
    result = bitmask_left.intersection_all(bitmask_right)
    if debug:
        superprint(f"bitmask testing {node} with bitmask_func_intersection_all -> {result}")
    return result


def bitmask_func_subset(
    node: str,
    bitmask_left: EnhancedBitmask,
    bitmask_right: EnhancedBitmask,
    debug=False,
) -> bool:
    result = EnhancedBitmask.logical_is_subset(bitmask_left, bitmask_right)
    if debug:
        superprint(f"bitmask testing {node} with bitmask_func_subset -> {result}")
    return result


def breadth_first_bitmask_search(
    root_name: str,
    graph: nx.DiGraph,
    bitmask: Optional[EnhancedBitmask],
    bitmask_func,
    filter_func,
    debug=False,
    animated_graph=None,
) -> List[str]:
    """
    if bitmask is None or bitmask_func is None, it will go throught the entire tree.
    once we pass the bitmask/bitmask_func test, it runs the filter_func on all nodes allowed to search.
    """
    # FIXME eval: decide if we want to return the node_dct + node_name???
    # FIXME priority 1: this is not BFS.  Might need to create queue, empty it while creating a new queue.
    nodes_searched = 0
    if bitmask_func is None:
        bitmask_func = bitmask_func_true
    result: List[str] = []
    root_dct = graph.nodes[root_name]
    # search stats vars
    if bitmask is None or bitmask_func(root_name, bitmask, root_dct["bitmask"], debug=debug):
        search_queue = deque()
        visited = {root_name}
        func_result = filter_func(root_name, root_dct, debug=debug)
        if func_result:
            if animated_graph:
                animated_graph.add(root_name, color="green")
            result.append(root_name)
        else:
            if animated_graph:
                animated_graph.add(root_name, color="red")
        for edge in graph.edges(root_name):
            search_queue.appendleft(edge)
        # search_queue.append(graph.edges(root_name))
        while search_queue:
            # edges = search_queue.pop()
            # for edge in edges:
            edge = search_queue.pop()
            nodes_searched += 1
            node_name, child_name = edge
            if child_name not in visited:
                visited.add(child_name)
                child_dct = graph.nodes[child_name]
                child_bitmask = child_dct["bitmask"]
                # FIXME: create logical_subset to return bool and not do the allocs. !((a & b) == a).all()
                #   this could be a function to pass in.
                # check to make sure we go deeper and that the bitmasks intersect.
                if bitmask is None or bitmask_func(child_name, bitmask, child_bitmask, debug=debug):
                    if debug:
                        superprint(f"going deeper into {child_name} edges")
                    # found a subset, go deeper.  If only a few match, it won't go deeper.
                    for edge in graph.edges(child_name):
                        search_queue.appendleft(edge)
                    # search_queue.appendLeft(graph.edges(child_name))
                    # do we add this to the resultset
                    func_result = filter_func(child_name, child_dct, debug=debug)
                    if func_result:
                        if animated_graph:
                            animated_graph.add(child_name, color="green")
                        result.append(child_name)
                    else:
                        if animated_graph:
                            animated_graph.add(child_name, color="red")
    if debug:
        superprint(f"{len(graph.nodes)=} {nodes_searched=}")
    return result


def convert_bitmask_in_graph_to(graph: nx.DiGraph, bitmask_type):
    """bitmask_type -> list, None, np.ndarray"""
    graph = copy.deepcopy(graph)
    for node_name, node_dct in graph.nodes.items():
        if bitmask_type is None:
            del node_dct["bitmask"]
        elif bitmask_type == list:
            node_dct["bitmask"] = node_dct["bitmask"].tolist()
        elif bitmask_type == np.ndarray:
            node_dct["bitmask"] = np.array(node_dct["bitmask"])
        elif bitmask_type == LiteBitmask:
            # node_dct['bitmask'] might already be a LiteBitmask
            bit_idx_lst = EnhancedBitmask.get_idx_lst_from_bitmask(node_dct["bitmask"])
            node_dct["bitmask"] = LiteBitmask.zeros_and_set_bits(node_dct["bitmask"].length, bit_idx_lst)
        elif bitmask_type == EnhancedBitmask:
            pass
        else:
            raise Exception(f"invalid bitmask type {bitmask_type}")
    return graph
