# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import abc
import copy
import itertools
import math
from functools import reduce
from itertools import groupby
from typing import Union, Tuple, Generator, List, Set, OrderedDict
import heapq
import simplejson

from Octeres.bitmask import EnhancedBitmaskLike, EnhancedBitmask
from Octeres.serialization import DataEncoder, DataDecoder

"""
invariants:
    all BitIntervals must have ordered and merged intervals.
    Evaluate 
        LiteBitmask should be immutable.
        changing bit_idx_first and last to a property that's immutable.
    Do a pass later an drop underscores on functions not needed outside.
There is still more work that needs to be done in here.
    LiteBitmaskSlots can be directly used like a Row in spark because it can act like a dictionary.
"""


IntervalList = List[List[int]]


class LiteBitmaskSlotsLike(metaclass=abc.ABCMeta):
    """Creation of a 'trait' like class that allows LiteBitmask and LiteBitmaskSlots to interact."""

    @property
    @abc.abstractmethod
    def length(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def bit_count(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def intervals(self) -> list:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def bit_idx_first(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def bit_idx_last(self) -> int:
        raise NotImplementedError()

    def all(self, **kwargs) -> bool:
        return True if self.length == self.bit_count else False

    def any(self, **kwargs) -> bool:
        return True if self.bit_count > 0 else False

    def none(self, **kwargs) -> bool:
        return True if self.bit_count == 0 else False

    def __getitem__(self, item):
        raise NotImplementedError()


class LiteBitmaskSlots(dict, LiteBitmaskSlotsLike):
    """This is an alternate version of LiteBitmask that is heavily used in serialization and can go directly in and
    out of spark with minimal changes.  At first __slots__ was just used, but then __getitem__ was needed to allow
    the object to act like a dictionary, thus the dictionary superclass.  __slots__ is still useful as it provides
    the setters that LiteBitmaskSlotsLike requires."""

    __slots__ = ("length", "bit_count", "intervals", "bit_idx_first", "bit_idx_last")

    def __init__(self, length, bit_count, intervals, bit_idx_first, bit_idx_last):
        super().__init__(
            length=length, bit_count=bit_count, intervals=intervals, bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last
        )
        self.length = length
        self.bit_count = bit_count
        self.intervals = intervals
        self.bit_idx_first = bit_idx_first
        self.bit_idx_last = bit_idx_last

    def unpack(self):
        return self.length, self.bit_count, self.intervals, self.bit_idx_first, self.bit_idx_last

    def __getitem__(self, item):
        if type(item) == int:
            # FIXME: this is a copy of LiteBitmask int __getitem__
            if item >= self.length:
                raise IndexError
            # TODO: do a binary search or some neat source.
            result = False
            for start, end in self.intervals:
                if start <= item <= end:
                    result = True
                    break
        else:
            result = super().__getitem__(item)
        return result

    def __eq__(self, other: LiteBitmaskSlotsLike) -> bool:
        """does self equal other?"""
        try:
            result = (
                self.bit_count == other.bit_count
                and self.bit_idx_first == other.bit_idx_first
                and self.bit_idx_last == other.bit_idx_last
                and self.intervals == other.intervals
                and self.length == other.length
            )
        except AttributeError:
            result = (
                self["bit_count"] == other["bit_count"]
                and self["bit_idx_first"] == other["bit_idx_first"]
                and self["bit_idx_last"] == other["bit_idx_last"]
                and self["intervals"] == other["intervals"]
                and self["length"] == other["length"]
            )
        return result

    def __bool__(self) -> bool:
        """the truth of an object is not enough"""
        raise ValueError("Truth of an LiteBitmaskSlots is ambiguous, use .any() or .all()")


# def intersection_possible(a_bitmask: LiteBitmaskSlotsLike, b_bitmask: LiteBitmaskSlotsLike):
#     if a_bitmask is None or b_bitmask is None:
#         result = False
#     elif a_bitmask.bit_count == 0 or b_bitmask.bit_count == 0:
#         result = False
#     elif a_bitmask.bit_idx_first > b_bitmask.bit_idx_last:
#         result = False
#     elif a_bitmask.bit_idx_last < b_bitmask.bit_idx_first:
#         result = False
#     else:
#         result = True
#     return result


def intersection_possible(a_bitmask: LiteBitmaskSlotsLike, b_bitmask: LiteBitmaskSlotsLike):
    if a_bitmask is None or b_bitmask is None:
        result = False
    elif a_bitmask["bit_count"] == 0 or b_bitmask["bit_count"] == 0:
        result = False
    elif a_bitmask["bit_idx_first"] > b_bitmask["bit_idx_last"]:
        result = False
    elif a_bitmask["bit_idx_last"] < b_bitmask["bit_idx_first"]:
        result = False
    else:
        result = True
    return result


def intersection_any_v0(a_bitmask: LiteBitmaskSlotsLike, b_bitmask: LiteBitmaskSlotsLike):
    if not intersection_possible(a_bitmask, b_bitmask):
        result = False
    else:
        first_a = a_bitmask["bit_idx_first"]
        last_a = a_bitmask["bit_idx_last"]
        first_b = b_bitmask["bit_idx_first"]
        last_b = b_bitmask["bit_idx_last"]

        # if a_bitmask.bit_idx_first == b_bitmask.bit_idx_first or a_bitmask.bit_idx_first == b_bitmask.bit_idx_last:
        if first_a == first_b or first_a == last_b:
            result = True
        # elif a_bitmask.bit_idx_last == b_bitmask.bit_idx_first or a_bitmask.bit_idx_last == b_bitmask.bit_idx_last:
        elif last_a == first_b or last_a == last_b:
            result = True
        else:
            result = False
            bit_idx_first = first_a if first_a >= first_b else first_b
            bit_idx_last = last_a if last_a < last_b else last_b
            a_gen = get_interval_bits_generator(a_bitmask, bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last)
            b_gen = get_interval_bits_generator(b_bitmask, bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last)
            try:
                a_idx = next(a_gen)
                b_idx = next(b_gen)
            except StopIteration:
                # if either Stop, get out.
                a_idx = None
                b_idx = None
            while a_idx and b_idx:
                if a_idx == b_idx:
                    result = True
                    break
                elif a_idx > b_idx:
                    try:
                        b_idx = next(b_gen)
                    except StopIteration:
                        break
                else:  # elif b_idx > a_idx:
                    try:
                        a_idx = next(a_gen)
                    except StopIteration:
                        break

    return result


def intersection_any_v1(a_bitmask: LiteBitmaskSlotsLike, b_bitmask: LiteBitmaskSlotsLike):
    if not intersection_possible(a_bitmask, b_bitmask):
        result = False
    else:
        first_a = a_bitmask["bit_idx_first"]
        last_a = a_bitmask["bit_idx_last"]
        first_b = b_bitmask["bit_idx_first"]
        last_b = b_bitmask["bit_idx_last"]

        # result = (first_a == first_b) or ((last_b - first_a) <= 0)
        result = (first_a == first_b) or (first_a == last_b) or (last_a == first_b) or (last_a == last_b)

        if not result:
            bit_idx_first = max(first_a, first_b)
            bit_idx_last = min(last_a, last_b)
            a_gen = get_interval_bits_generator(a_bitmask, bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last)
            b_gen = get_interval_bits_generator(b_bitmask, bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last)
            try:
                a_idx = next(a_gen)
                b_idx = next(b_gen)
            except StopIteration:
                # if either Stop, get out.
                a_idx = None
                b_idx = None
            while a_idx and b_idx:
                if a_idx == b_idx:
                    result = True
                    break
                elif a_idx > b_idx:
                    try:
                        b_idx = next(b_gen)
                    except StopIteration:
                        break
                else:  # elif b_idx > a_idx:
                    try:
                        a_idx = next(a_gen)
                    except StopIteration:
                        break
    return result


def intersection_all(a_bitmask: LiteBitmaskSlotsLike, b_bitmask: LiteBitmaskSlotsLike):
    if intersection_possible(a_bitmask, b_bitmask):
        if a_bitmask["bit_count"] != b_bitmask["bit_count"]:
            result = False
        elif a_bitmask["bit_count"] == 0 or b_bitmask["bit_count"] == 0:
            result = False
        else:
            xnored = logical_xnor(a_bitmask, b_bitmask)
            if xnored["bit_count"] == xnored["length"]:
                result = True
            else:
                result = False
    else:
        result = False
    return result


def find_interval_properties(intervals: "IntervalList"):
    # intervals, and thus intervals. should be immutable, sorted and merged
    # assume sorted and no overlaps or consecutive intervals (i.e., [5,6],[7,8] should be [5,8])
    bit_count = 0
    bit_idx_first = -1
    bit_idx_last = -1
    try:
        for start, end in intervals:  # interval must be a length of 2
            assert end >= start
            bit_count += end - start + 1
    except ValueError as err:
        if "too many values to unpack" in str(err):
            raise IntervalError(f"interval must be a length of 2:{intervals}")
        elif "not enough values to unpack" in str(err):
            raise IntervalError(f"interval must be a length of 2:{intervals}")
        else:
            raise
    if bit_count:
        bit_idx_first = intervals[0][0]
        bit_idx_last = intervals[-1][1]
    return bit_count, bit_idx_first, bit_idx_last


def convert_intervals_list_to_LiteBitmask(length: int, intervals: "IntervalList", validate=False, merge=True) -> "LiteBitmask":
    """merge will create a new intervals list so no need to copy"""
    if merge:
        if intervals:
            intervals, _ = merge_intervals(intervals)
        else:  # create a new list as merge_intervals would
            intervals, _ = [], 0
    bit_count, bit_idx_first, bit_idx_last = find_interval_properties(intervals)
    bitmask = LiteBitmask(length, bit_count, intervals, bit_idx_first, bit_idx_last)
    if validate:
        validate_litebitmaskslotslike(bitmask)
    return bitmask


def convert_intervals_to_LiteBitmaskSlots(length: int, intervals: "IntervalList", validate=False, merge=True) -> LiteBitmaskSlots:
    """merge will create a new intervals list so no need to copy"""
    if merge:
        if intervals:
            intervals, _ = merge_intervals(intervals)
        else:  # create a new list as merge_intervals would
            intervals, _ = [], 0
    bit_count, bit_idx_first, bit_idx_last = find_interval_properties(intervals)
    bitmask_slots = LiteBitmaskSlots(length, bit_count, intervals, bit_idx_first, bit_idx_last)
    if validate:
        validate_litebitmaskslotslike(bitmask_slots)
    return bitmask_slots


def convert_bitmask_to_litebitmask(bitmask: EnhancedBitmask):
    length: int = bitmask.length
    obj = LiteBitmask.zeros_and_set_bits(length, bitmask.get_idx_lst_from_bitmask(bitmask))
    return obj


def get_interval_delta(bir):
    a = bir[0]
    b = bir[1]
    if a > b:
        raise IntervalError(f"negative interval: {bir}")
    return (b - a) + 1


def merge_intervals(intervals: IntervalList) -> Tuple[IntervalList, int]:
    merged_intervals = []
    merged_intervals_append = merged_intervals.append
    bit_count = 0
    try:
        start, end = intervals[0]
    except ValueError:
        raise IntervalError(f"Invalid interval structure, 2 items are required in each interval, found: {len(intervals[0])}")
    for i in range(1, len(intervals)):
        current_start, current_end = intervals[i]
        if current_start > end + 1:
            merged_intervals_append([start, end])
            bit_count += end - start + 1
            start, end = current_start, current_end
        else:
            end = end if end >= current_end else current_end
    merged_intervals_append([start, end])
    bit_count += end - start + 1
    return merged_intervals, bit_count


def validate_litebitmaskslotslike(obj: LiteBitmaskSlotsLike):
    intervals = obj.intervals
    assert type(obj.length) == int
    if obj.bit_count > 0:
        assert obj.bit_idx_first == intervals[0][0]
        assert obj.bit_idx_last == intervals[-1][1]
    else:
        assert obj.bit_idx_first == -1
        assert obj.bit_idx_last == -1
    len_intervals = len(intervals)
    bit_count = 0
    for i, alar in enumerate(intervals):
        al, ar = alar
        sub_bit_count = (ar - al) + 1
        if sub_bit_count < 1:
            raise IntervalError(f"negative interval: [{al}, {ar}]")
        bit_count += sub_bit_count
        if al > ar:
            raise IntervalError(f"negative interval: {alar}")
        if (i + 1) < len_intervals:
            blbr = intervals[i + 1]
            if alar == blbr:
                raise IntervalError(f"duplicate interval: {alar}")
            bl, br = blbr
            if (ar + 1) == bl:
                raise IntervalError(f"interval not merged: {ar}+1 == {bl} for {alar}->{blbr}")
            if ar > bl:
                raise IntervalError(f"interval out of order or not merged: {ar} > {bl} for {alar}->{blbr}")
    assert bit_count == obj.bit_count


def get_interval_bits_generator_v0(
    bitmask_slots: LiteBitmaskSlotsLike, bit_idx_first=0, bit_idx_last=None
) -> Generator[int, None, None]:
    """get the indexes that are set.
    use start if you are looking for indexes above a count.
    This is slow and there are probably other better ways to do this.  All calls should evaluate this.
    """
    intervals: List[List] = bitmask_slots["intervals"]
    for interval in intervals:
        for i in range(interval[0], interval[1] + 1):
            if i < bit_idx_first:
                continue
            if bit_idx_last is not None and (i > bit_idx_last):
                continue
            yield i


def get_interval_bits_generator_v1(
    bitmask_slots: LiteBitmaskSlotsLike, bit_idx_first=0, bit_idx_last=None
) -> Generator[int, None, None]:
    intervals = bitmask_slots["intervals"]
    result = (
        i
        for interval in intervals
        for i in range(interval[0], interval[1] + 1)
        if (i >= bit_idx_first and (bit_idx_last is None or i <= bit_idx_last))
    )
    return result


def get_interval_bits(bitmask_slots: LiteBitmaskSlotsLike, bit_idx_first=0, bit_idx_last=None) -> List:
    """get the indexes that are set.
    use start if you are looking for indexes above a count.
    This is slow and there are probably other better ways to do this.  All calls should evaluate this.
    """
    intervals: List[List] = bitmask_slots["intervals"]
    bit_lst = []
    bit_lst_append = bit_lst.append
    for interval in intervals:
        for i in range(interval[0], interval[1] + 1):
            if i < bit_idx_first:
                continue
            if bit_idx_last is not None and (i > bit_idx_last):
                continue
            bit_lst_append(i)
    return bit_lst


def get_interval_bits_set_v0(bitmask_slots: LiteBitmaskSlots) -> set:
    """get the indexes that are set."""
    # FIXME: priority 1: stop using this, it's slow
    return {idx for interval in bitmask_slots["intervals"] for idx in [i for i in range(interval[0], interval[1] + 1)]}


def get_interval_bits_set(bitmask_slots: LiteBitmaskSlots) -> Set[int]:
    bit_set: Set[int] = set()
    for interval in bitmask_slots["intervals"]:
        bit_set.update(range(interval[0], interval[1] + 1))
    return bit_set


def logical_subtract_v0(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """subtract the indexes of a list of intervals from another"""
    # FIXME: this needs work, it may be very slow.
    set_a = get_interval_bits_set(bitmask_slots_a)
    set_b = get_interval_bits_set(bitmask_slots_b)
    set_c = set_a - set_b
    length = bitmask_slots_a["length"]
    result = LiteBitmaskSlots(length, *index_list_to_intervals_and_properties(set_c))
    return result


def logical_subtract_v1(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Directly perform subtract function on the intervals"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    result_intervals = []
    result_intervals_append = result_intervals.append

    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    while i < a_intervals_len:
        a_start, a_end = a_intervals[i]
        while j < b_intervals_len:
            b_start, b_end = b_intervals[j]
            if b_end < a_start:
                j += 1
                continue
            if b_start > a_end:
                break
            if b_start <= a_start and b_end >= a_end:
                a_start = a_end + 1
                break
            elif b_start <= a_start:
                a_start = b_end + 1
            elif b_end >= a_end:
                result_intervals_append([a_start, b_start - 1])
                a_start = a_end + 1
                break
            else:
                result_intervals_append([a_start, b_start - 1])
                a_start = b_end + 1
            j += 1
        if a_start <= a_end:
            result_intervals_append([a_start, a_end])
        i += 1

    if not result_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        merged_intervals, bit_count = merge_intervals(result_intervals)
        # bit_count = sum(end - start + 1 for start, end in merged_intervals)
        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_subtract_v2(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Refine the logic during the interval walkthrough"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    result_intervals = []
    result_intervals_append = result_intervals.append

    i, j = 0, 0
    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)

    while i < a_intervals_len:
        a_start, a_end = a_intervals[i]

        # Skip b_intervals that end before the current a_start
        while j < b_intervals_len and b_intervals[j][1] < a_start:
            j += 1

        while j < b_intervals_len:
            b_start, b_end = b_intervals[j]
            if b_start > a_end:
                break

            # Handle overlapping intervals
            if b_start <= a_start and b_end >= a_end:
                a_start = a_end + 1
                break
            elif b_start <= a_start:
                a_start = b_end + 1
            elif b_end >= a_end:
                result_intervals_append([a_start, b_start - 1])
                a_start = a_end + 1
                break
            else:
                result_intervals_append([a_start, b_start - 1])
                a_start = b_end + 1
            j += 1

        if a_start <= a_end:
            result_intervals_append([a_start, a_end])

        i += 1

    if not result_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        merged_intervals, bit_count = merge_intervals(result_intervals)
        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_subtract_v3(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """An attempt to cut the merge interval process, directly generate the result during the walkthrough"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []
    final_intervals_append = final_intervals.append

    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    start, end = None, None

    while i < a_intervals_len:
        a_start, a_end = a_intervals[i]
        while j < b_intervals_len:
            b_start, b_end = b_intervals[j]
            if b_end < a_start:
                j += 1
                continue
            if b_start > a_end:
                break
            if b_start <= a_start and b_end >= a_end:
                a_start = a_end + 1
                break
            elif b_start <= a_start:
                a_start = b_end + 1
            elif b_end >= a_end:
                if a_start < b_start:
                    if start is None:
                        start, end = a_start, b_start - 1
                    else:
                        if a_start > end + 1:
                            final_intervals_append([start, end])
                            start, end = a_start, b_start - 1
                        else:
                            end = b_start - 1
                a_start = a_end + 1
                break
            else:
                if start is None:
                    start, end = a_start, b_start - 1
                else:
                    if a_start > end + 1:
                        final_intervals_append([start, end])
                        start, end = a_start, b_start - 1
                    else:
                        end = b_start - 1
                a_start = b_end + 1
            j += 1
        if a_start <= a_end:
            if start is None:
                start, end = a_start, a_end
            else:
                if a_start > end + 1:
                    final_intervals_append([start, end])
                    start, end = a_start, a_end
                else:
                    end = a_end
        i += 1

    if start is not None:
        final_intervals_append([start, end])

    if not final_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        bit_count = sum(end - start + 1 for start, end in final_intervals)
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_subtract_v4(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Anothe attempt to cut the merge interval process, refining the walkthrough logic"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []
    final_intervals_append = final_intervals.append

    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    start, end = None, None
    bit_count = 0

    while i < a_intervals_len:
        a_start, a_end = a_intervals[i]
        while j < b_intervals_len:
            b_start, b_end = b_intervals[j]
            if b_end < a_start:
                j += 1
                continue
            if b_start > a_end:
                break
            if b_start <= a_start and b_end >= a_end:
                a_start = a_end + 1
                break
            elif b_start <= a_start:
                a_start = b_end + 1
            elif b_end >= a_end:
                if a_start < b_start:
                    if start is None:
                        start, end = a_start, b_start - 1
                    else:
                        if a_start > end + 1:
                            final_intervals_append([start, end])
                            bit_count += end - start + 1
                            start, end = a_start, b_start - 1
                        else:
                            end = b_start - 1
                a_start = a_end + 1
                break
            else:
                if start is None:
                    start, end = a_start, b_start - 1
                else:
                    if a_start > end + 1:
                        final_intervals_append([start, end])
                        bit_count += end - start + 1
                        start, end = a_start, b_start - 1
                    else:
                        end = b_start - 1
                a_start = b_end + 1
            j += 1
        if a_start <= a_end:
            if start is None:
                start, end = a_start, a_end
            else:
                if a_start > end + 1:
                    final_intervals_append([start, end])
                    bit_count += end - start + 1
                    start, end = a_start, a_end
                else:
                    end = a_end
        i += 1

    if start is not None:
        final_intervals_append([start, end])
        bit_count += end - start + 1

    if not final_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def get_intervals_until_all_consumed(alst: IntervalList, blst: IntervalList) -> Generator[IntervalList, None, None]:
    aidx = 0
    bidx = 0
    while True:
        try:
            amin, amax = alst[aidx]
        except IndexError:
            amin, amax = None, None
        try:
            bmin, bmax = blst[bidx]
        except IndexError:
            bmin, bmax = None, None
        if amin is None and bmin is None:
            break
        if bmin is None or (amin is not None and amin <= bmin):
            r = [amin, amax]
            aidx += 1
        else:
            r = [bmin, bmax]
            bidx += 1
        yield r


def get_combined_intervals_until_overlap_impossible(alst: IntervalList, blst: IntervalList) -> Generator[IntervalList, None, None]:
    aidx = 0
    bidx = 0
    while True:
        try:
            amin, amax = alst[aidx]
        except IndexError:
            # amin = None if len(alst) else -1
            # amax = alst[-1][1] if len(alst) else -1
            amin, amax = None, alst[-1][1] if len(alst) else -1
        try:
            bmin, bmax = blst[bidx]
        except IndexError:
            bmin, bmax = None, blst[-1][1] if len(blst) else -1

        if (amin is None and bmin is None) or (amin is None and bmin > amax) or (bmin is None and amin > bmax):
            break

        if bmin is None or (amin is not None and amin <= bmin):
            r = [amin, amax]
            aidx += 1
        else:
            r = [bmin, bmax]
            bidx += 1
        yield r


def logical_or_v0(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """does a logical or of two bit intervals"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    interval_gen = get_intervals_until_all_consumed(a_intervals, b_intervals)
    try:
        prev_start, prev_end = next(interval_gen)
        intervals = [[prev_start, prev_end]]
        bit_count = prev_end - prev_start + 1
    except StopIteration:
        # both of the BitIntervals lists were empty
        bit_count = 0
        intervals = []
        bit_idx_first = -1
        bit_idx_last = -1
    else:
        for next_start, next_end in interval_gen:
            if next_start <= prev_end + 1:
                new_end = max(prev_end, next_end)
                intervals[-1] = [prev_start, new_end]
                bit_count += new_end - prev_end
                prev_end = new_end
            else:
                intervals.append([next_start, next_end])
                bit_count += next_end - next_start + 1
                prev_start, prev_end = next_start, next_end
        bit_idx_first = intervals[0][0]
        bit_idx_last = intervals[-1][1]
    bitmask_slots = LiteBitmaskSlots(length, bit_count, intervals, bit_idx_first, bit_idx_last)
    return bitmask_slots


def logical_or_v4(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Fixed the over merge problem occurred in v1 and v2"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []
    final_intervals_append = final_intervals.append
    i, j = 0, 0
    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    start, end = None, None

    def append_interval(start, end):
        if start is not None:
            if final_intervals:
                last_start, last_end = final_intervals[-1]
                if start <= last_end + 1:
                    final_intervals[-1][1] = max(last_end, end)
                else:
                    final_intervals_append([start, end])
            else:
                final_intervals_append([start, end])

    while i < a_intervals_len and j < b_intervals_len:
        a_start, a_end = a_intervals[i]
        b_start, b_end = b_intervals[j]

        if a_end < b_start:
            append_interval(start, end)
            start, end = a_start, a_end
            i += 1
        elif b_end < a_start:
            append_interval(start, end)
            start, end = b_start, b_end
            j += 1
        else:
            append_interval(start, end)
            start = min(a_start, b_start)
            end = max(a_end, b_end)
            if a_end > b_end:
                j += 1
            elif a_end < b_end:
                i += 1
            else:
                i += 1
                j += 1

    while i < a_intervals_len:
        a_start, a_end = a_intervals[i]
        append_interval(start, end)
        start, end = a_start, a_end
        i += 1

    while j < b_intervals_len:
        b_start, b_end = b_intervals[j]
        append_interval(start, end)
        start, end = b_start, b_end
        j += 1

    append_interval(start, end)

    if not final_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        bit_count = sum(end - start + 1 for start, end in final_intervals)
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_or_v5(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Optimized version of intervals_or with minimized condition checks"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []
    final_intervals_append = final_intervals.append
    i, j = 0, 0
    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    bit_count = 0

    def merge_or_append_interval_v5(start, end, bit_count):
        if final_intervals and final_intervals[-1][1] >= start - 1:
            previous_end = final_intervals[-1][1]
            final_intervals[-1][1] = max(final_intervals[-1][1], end)
            bit_count += final_intervals[-1][1] - previous_end
        else:
            final_intervals_append([start, end])
            bit_count += end - start + 1
        return bit_count

    while i < a_intervals_len and j < b_intervals_len:
        if a_intervals[i][0] <= b_intervals[j][0]:
            start, end = a_intervals[i]
            i += 1
        else:
            start, end = b_intervals[j]
            j += 1

        while i < a_intervals_len and a_intervals[i][0] <= end + 1:
            end = max(end, a_intervals[i][1])
            i += 1
        while j < b_intervals_len and b_intervals[j][0] <= end + 1:
            end = max(end, b_intervals[j][1])
            j += 1

        bit_count = merge_or_append_interval_v5(start, end, bit_count)

    while i < a_intervals_len:
        bit_count = merge_or_append_interval_v5(a_intervals[i][0], a_intervals[i][1], bit_count)
        i += 1

    while j < b_intervals_len:
        bit_count = merge_or_append_interval_v5(b_intervals[j][0], b_intervals[j][1], bit_count)
        j += 1

    if not final_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_or_v6(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Optimized version of intervals_or_v4 with cut the redundant overlap and margin check"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []
    final_intervals_append = final_intervals.append
    i, j = 0, 0
    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)

    while i < a_intervals_len and j < b_intervals_len:
        if a_intervals[i][1] < b_intervals[j][0]:
            if final_intervals and final_intervals[-1][1] >= a_intervals[i][0] - 1:
                final_intervals[-1][1] = max(final_intervals[-1][1], a_intervals[i][1])
            else:
                final_intervals_append(a_intervals[i])
            i += 1
        elif b_intervals[j][1] < a_intervals[i][0]:
            if final_intervals and final_intervals[-1][1] >= b_intervals[j][0] - 1:
                final_intervals[-1][1] = max(final_intervals[-1][1], b_intervals[j][1])
            else:
                final_intervals_append(b_intervals[j])
            j += 1
        else:
            start = min(a_intervals[i][0], b_intervals[j][0])
            end = max(a_intervals[i][1], b_intervals[j][1])
            while i < a_intervals_len and a_intervals[i][0] <= end:
                end = max(end, a_intervals[i][1])
                i += 1
            while j < b_intervals_len and b_intervals[j][0] <= end:
                end = max(end, b_intervals[j][1])
                j += 1
            if final_intervals and final_intervals[-1][1] >= start - 1:
                final_intervals[-1][1] = max(final_intervals[-1][1], end)
            else:
                final_intervals_append([start, end])

    while i < a_intervals_len:
        if final_intervals and final_intervals[-1][1] >= a_intervals[i][0] - 1:
            final_intervals[-1][1] = max(final_intervals[-1][1], a_intervals[i][1])
        else:
            final_intervals_append(a_intervals[i])
        i += 1

    while j < b_intervals_len:
        if final_intervals and final_intervals[-1][1] >= b_intervals[j][0] - 1:
            final_intervals[-1][1] = max(final_intervals[-1][1], b_intervals[j][1])
        else:
            final_intervals_append(b_intervals[j])
        j += 1

    if not final_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        bit_count = sum(end - start + 1 for start, end in final_intervals)
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def _merge_or_append_interval_v7(start, end, final_intervals, final_intervals_append, bit_count):
    """Deal with the interval merging boundary for v7 and calculate the bit_count"""
    if final_intervals and final_intervals[-1][1] >= start - 1:
        previous_end = final_intervals[-1][1]
        # VERY BAD     vvvvvvv v
        final_intervals[-1][1] = max(previous_end, end)
        bit_count += final_intervals[-1][1] - previous_end
    else:
        final_intervals_append([start, end])
        bit_count += end - start + 1
    return bit_count


def _merge_or_append_interval_v8(start, end, final_intervals, final_intervals_append, bit_count):
    """Deal with the interval merging boundary for v7 and calculate the bit_count"""
    if final_intervals:
        previous_end = final_intervals[-1][1]
        if previous_end >= start - 1:
            new_end = previous_end if previous_end >= end else end
            final_intervals[-1][1] = new_end
            bit_count += new_end - previous_end
        else:
            final_intervals_append([start, end])
            bit_count += end - start + 1
    else:
        final_intervals_append([start, end])
        bit_count += end - start + 1
    return bit_count


def logical_or_v7(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Try to take merging function outside the 'or' function to see if it could be faster"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []
    final_intervals_append = final_intervals.append
    i, j = 0, 0
    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    bit_count = 0

    while i < a_intervals_len and j < b_intervals_len:
        if a_intervals[i][0] <= b_intervals[j][0]:
            start, end = a_intervals[i]
            i += 1
        else:
            start, end = b_intervals[j]
            j += 1

        while i < a_intervals_len and a_intervals[i][0] <= end + 1:
            end = max(end, a_intervals[i][1])
            i += 1
        while j < b_intervals_len and b_intervals[j][0] <= end + 1:
            end = max(end, b_intervals[j][1])
            j += 1

        bit_count = _merge_or_append_interval(start, end, final_intervals, final_intervals_append, bit_count)

    while i < a_intervals_len:
        bit_count = _merge_or_append_interval(
            a_intervals[i][0], a_intervals[i][1], final_intervals, final_intervals_append, bit_count
        )
        i += 1

    while j < b_intervals_len:
        bit_count = _merge_or_append_interval(
            b_intervals[j][0], b_intervals[j][1], final_intervals, final_intervals_append, bit_count
        )
        j += 1

    if not final_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_or_v8(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Try to take merging function outside the 'or' function to see if it could be faster"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []
    final_intervals_append = final_intervals.append
    i, j = 0, 0
    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    bit_count = 0

    # while i < a_intervals_len and j < b_intervals_len:
    while i < a_intervals_len or j < b_intervals_len:
        if i < a_intervals_len and j >= b_intervals_len:
            start, end = a_intervals[i]
            i += 1
        elif j < b_intervals_len and i >= a_intervals_len:
            start, end = b_intervals[j]
            j += 1
        else:
            start_a, end_a = a_intervals[i]
            start_b, end_b = b_intervals[j]
            if start_a <= start_b:
                start, end = start_a, end_a
                i += 1
            else:
                start, end = start_b, end_b
                j += 1

            while i < a_intervals_len and a_intervals[i][0] <= end + 1:
                if end < a_intervals[i][1]:
                    end = a_intervals[i][1]
                i += 1
            while j < b_intervals_len and b_intervals[j][0] <= end + 1:
                if end < b_intervals[j][1]:
                    end = b_intervals[j][1]
                j += 1
        # bit_count = _merge_or_append_interval(start, end, final_intervals, final_intervals_append, bit_count)
        if final_intervals:
            previous_end = final_intervals[-1][1]
            if previous_end >= start - 1:
                new_end = previous_end if previous_end >= end else end
                final_intervals[-1][1] = new_end
                bit_count += new_end - previous_end
            else:
                final_intervals_append([start, end])
                bit_count += end - start + 1
        else:
            final_intervals_append([start, end])
            bit_count += end - start + 1

    if not final_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def get_combined_intervals_until_all_consumed_v1(bitmask_slots_lst: List[LiteBitmaskSlots]):
    """Combine up to 10,000 intervals using OR operation with space allocation only once"""
    final_intervals = []
    final_intervals_append = final_intervals.append
    bit_count = 0
    min_heap = []

    for bitmask_slots in bitmask_slots_lst:
        intervals = bitmask_slots["intervals"]
        if intervals:
            heapq.heappush(min_heap, (intervals[0][0], intervals[0][1], intervals, 1))

    while min_heap:
        start, end, current_intervals, idx = heapq.heappop(min_heap)

        while min_heap and min_heap[0][0] <= end + 1:
            next_start, next_end, next_intervals, next_idx = heapq.heappop(min_heap)
            end = max(end, next_end)
            if next_idx < len(next_intervals):
                heapq.heappush(min_heap, (next_intervals[next_idx][0], next_intervals[next_idx][1], next_intervals, next_idx + 1))

        bit_count = _merge_or_append_interval(start, end, final_intervals, final_intervals_append, bit_count)

        if idx < len(current_intervals):
            heapq.heappush(min_heap, (current_intervals[idx][0], current_intervals[idx][1], current_intervals, idx + 1))

    if not final_intervals:
        result = liteBitmaskSlotsZero
    else:
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        length = bitmask_slots_lst[0]["length"]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def get_combined_intervals_until_all_consumed_v2(bitmask_slots_lst: List[LiteBitmaskSlots]):
    """Do some define tricks on the v1"""
    final_intervals = []
    final_intervals_append = final_intervals.append
    bit_count = 0
    min_heap = []
    heapq_heappush = heapq.heappush
    heapq_heappop = heapq.heappop
    for bitmask_slots in bitmask_slots_lst:
        intervals = bitmask_slots["intervals"]
        if intervals:
            interval = intervals[0]
            heapq_heappush(min_heap, (interval[0], interval[1], intervals, 1))

    while min_heap:
        start, end, current_intervals, idx = heapq_heappop(min_heap)

        while min_heap and min_heap[0][0] <= end + 1:
            next_start, next_end, next_intervals, next_idx = heapq_heappop(min_heap)
            end = max(end, next_end)
            if next_idx < len(next_intervals):
                heapq_heappush(min_heap, (next_intervals[next_idx][0], next_intervals[next_idx][1], next_intervals, next_idx + 1))

        bit_count = _merge_or_append_interval(start, end, final_intervals, final_intervals_append, bit_count)

        if idx < len(current_intervals):
            heapq_heappush(min_heap, (current_intervals[idx][0], current_intervals[idx][1], current_intervals, idx + 1))

    if not final_intervals:
        result = liteBitmaskSlotsZero
    else:
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        length = bitmask_slots_lst[0]["length"]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def get_combined_intervals_until_all_consumed_v3(bitmask_slots_lst: List[LiteBitmaskSlots]):
    """Do some define tricks on the v1"""
    final_intervals = []
    final_intervals_append = final_intervals.append
    bit_count = 0
    min_heap = []
    heapq_heappush = heapq.heappush
    heapq_heappop = heapq.heappop
    for bitmask_slots in bitmask_slots_lst:
        intervals = bitmask_slots["intervals"]
        if intervals:
            interval = intervals[0]
            heapq_heappush(min_heap, (interval[0], interval[1], intervals, 1))

    while min_heap:
        start, end, current_intervals, idx = heapq_heappop(min_heap)

        while min_heap and min_heap[0][0] <= end + 1:
            next_start, next_end, next_intervals, next_idx = heapq_heappop(min_heap)
            end = max(end, next_end)
            if next_idx < len(next_intervals):
                heapq_heappush(min_heap, (next_intervals[next_idx][0], next_intervals[next_idx][1], next_intervals, next_idx + 1))

        if final_intervals:
            previous_end = final_intervals[-1][1]
            if previous_end >= start - 1:
                new_end = previous_end if previous_end >= end else end
                final_intervals[-1][1] = new_end
                bit_count += new_end - previous_end
            else:
                final_intervals_append([start, end])
                bit_count += end - start + 1
        else:
            final_intervals_append([start, end])
            bit_count += end - start + 1

        if idx < len(current_intervals):
            heapq_heappush(min_heap, (current_intervals[idx][0], current_intervals[idx][1], current_intervals, idx + 1))

    if not final_intervals:
        result = liteBitmaskSlotsZero
    else:
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        length = bitmask_slots_lst[0]["length"]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_not(bitmask_slots: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """negate the intervals"""
    length = bitmask_slots["length"]
    if bitmask_slots.all():
        new_bitmask_slots = get_bitmask_slots_zero(length)
    elif bitmask_slots.none():
        new_bitmask_slots = get_bitmask_slots_ones(length)
    else:
        bitmask_intervals = bitmask_slots["intervals"]
        if bitmask_intervals and bitmask_intervals[-1][1] >= length:
            raise IndexError("last interval exceeds specified length")
        intervals = []
        interval_append = intervals.append
        bit_count = 0
        not_start = 0
        for bi_start, bi_end in bitmask_intervals:
            if bi_start >= length:
                break
            not_end = bi_start - 1
            if not_end >= 0:
                interval_append([not_start, not_end])
            bit_count += not_end - not_start + 1
            not_start = bi_end + 1
        if not_start < length:
            interval_append([not_start, length - 1])
            bit_count += length - not_start
        bit_idx_first = intervals[0][0]
        bit_idx_last = intervals[-1][1]
        new_bitmask_slots = LiteBitmaskSlots(length, bit_count, intervals, bit_idx_first, bit_idx_last)
    return new_bitmask_slots


def logical_xor_set_v0(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """negate the intervals"""
    # FIXME: possibly use (a-b)|(b-a), but probably better to use iteration of one list (b?) and comparison with an iteration of
    # the other list (a?).  performance comparisons needed for the size of the two intervals.
    set_a = get_interval_bits_set(bitmask_slots_a)
    set_b = get_interval_bits_set(bitmask_slots_b)
    set_c = set_a.symmetric_difference(set_b)
    result = LiteBitmaskSlots(bitmask_slots_a.length, *index_list_to_intervals_and_properties_v1(set_c))
    return result


def logical_xor_set_v1(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Test a new version index_list_to_bitintervals"""
    set_a = get_interval_bits_set(bitmask_slots_a)
    set_b = get_interval_bits_set(bitmask_slots_b)
    set_c = set_a.symmetric_difference(set_b)
    result = LiteBitmaskSlots(bitmask_slots_a.length, *index_list_to_intervals_and_properties_v1(set_c))
    return result


def logical_xor_set_v2(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Test another version of index_list_to_bitintervals"""
    set_a = get_interval_bits_set(bitmask_slots_a)
    set_b = get_interval_bits_set(bitmask_slots_b)
    set_c = set_a.symmetric_difference(set_b)
    result = LiteBitmaskSlots(bitmask_slots_a.length, *index_list_to_intervals_and_properties_v1(set_c))
    return result


def logical_xor_v0(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Directly do xor on the intervals to avoid data type change"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    result_intervals = []
    i, j = 0, 0

    while i < len(a_intervals) and j < len(b_intervals):
        a_start, a_end = a_intervals[i]
        b_start, b_end = b_intervals[j]

        if a_end < b_start:
            result_intervals.append([a_start, a_end])
            i += 1
        elif b_end < a_start:
            result_intervals.append([b_start, b_end])
            j += 1
        else:
            if a_start < b_start:
                result_intervals.append([a_start, b_start - 1])
            if a_end > b_end:
                result_intervals.append([b_end + 1, a_end])
                j += 1
            elif a_end < b_end:
                result_intervals.append([a_end + 1, b_end])
                i += 1
            else:
                i += 1
                j += 1

    while i < len(a_intervals):
        result_intervals.append(a_intervals[i])
        i += 1

    while j < len(b_intervals):
        result_intervals.append(b_intervals[j])
        j += 1

    merged_intervals = []
    for start, end in result_intervals:
        if not merged_intervals or merged_intervals[-1][1] < start - 1:
            merged_intervals.append([start, end])
        else:
            merged_intervals[-1][1] = max(merged_intervals[-1][1], end)

    bit_count = sum(interval[1] - interval[0] + 1 for interval in merged_intervals)
    bit_idx_first = merged_intervals[0][0] if merged_intervals else -1
    bit_idx_last = merged_intervals[-1][1] if merged_intervals else -1
    result = LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_xor_v4(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Using copy to avoid the address problem when updating the value on original intervals"""
    a_intervals = bitmask_slots_a["intervals"].copy()
    b_intervals = bitmask_slots_b["intervals"].copy()
    length = bitmask_slots_a["length"]
    result_intervals = []
    result_intervals_append = result_intervals.append

    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    while i < a_intervals_len and j < b_intervals_len:
        a_start, a_end = a_intervals[i]
        b_start, b_end = b_intervals[j]

        if a_end < b_start:
            result_intervals_append([a_start, a_end])
            i += 1
        elif b_end < a_start:
            result_intervals_append([b_start, b_end])
            j += 1
        else:
            if a_start < b_start:
                result_intervals_append([a_start, b_start - 1])
                a_start = b_start
            if b_start < a_start:
                result_intervals_append([b_start, a_start - 1])

            if a_end > b_end:
                a_intervals[i] = [b_end + 1, a_end]
                j += 1
            elif a_end < b_end:
                b_intervals[j] = [a_end + 1, b_end]
                i += 1
            else:
                i += 1
                j += 1

    while i < a_intervals_len:
        start, end = a_intervals[i]
        result_intervals_append([start, end])
        i += 1

    while j < b_intervals_len:
        start, end = b_intervals[j]
        result_intervals_append([start, end])
        j += 1

    if not result_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        merged_intervals = []
        merged_intervals_append = merged_intervals.append

        start, end = result_intervals[0]
        result_intervals_len = len(result_intervals)
        for i in range(1, result_intervals_len):
            current_start, current_end = result_intervals[i]
            if current_start > end + 1:
                merged_intervals_append([start, end])
                start, end = current_start, current_end
            else:
                end = max(end, current_end)
        merged_intervals_append([start, end])

        bit_count = sum(end - start + 1 for start, end in merged_intervals)
        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_xor_v5(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Using temp intervals to avoid using copy for fast xor function"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    result_intervals = []
    result_intervals_append = result_intervals.append

    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    a_temp_interval = None
    b_temp_interval = None

    while i < a_intervals_len and j < b_intervals_len:
        if a_temp_interval is None:
            a_start, a_end = a_intervals[i]
        else:
            a_start, a_end = a_temp_interval

        if b_temp_interval is None:
            b_start, b_end = b_intervals[j]
        else:
            b_start, b_end = b_temp_interval

        if a_end < b_start:
            result_intervals_append([a_start, a_end])
            a_temp_interval = None
            i += 1
        elif b_end < a_start:
            result_intervals_append([b_start, b_end])
            b_temp_interval = None
            j += 1
        else:
            if a_start < b_start:
                result_intervals_append([a_start, b_start - 1])
                a_start = b_start
            if b_start < a_start:
                result_intervals_append([b_start, a_start - 1])
                b_start = a_start

            if a_end > b_end:
                a_temp_interval = [b_end + 1, a_end]
                b_temp_interval = None
                j += 1
            elif a_end < b_end:
                b_temp_interval = [a_end + 1, b_end]
                a_temp_interval = None
                i += 1
            else:
                a_temp_interval = None
                b_temp_interval = None
                i += 1
                j += 1

    while i < a_intervals_len:
        if a_temp_interval is None:
            start, end = a_intervals[i]
        else:
            start, end = a_temp_interval
            a_temp_interval = None
        result_intervals_append([start, end])
        i += 1

    while j < b_intervals_len:
        if b_temp_interval is None:
            start, end = b_intervals[j]
        else:
            start, end = b_temp_interval
            b_temp_interval = None
        result_intervals_append([start, end])
        j += 1

    if not result_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        merged_intervals = []
        merged_intervals_append = merged_intervals.append
        bit_count = 0
        start, end = result_intervals[0]
        result_intervals_len = len(result_intervals)
        for i in range(1, result_intervals_len):
            current_start, current_end = result_intervals[i]
            if current_start > end + 1:
                merged_intervals_append([start, end])
                bit_count += end - start + 1
                start, end = current_start, current_end
            else:
                end = max(end, current_end)

        merged_intervals_append([start, end])
        bit_count += end - start + 1
        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_xor_v6(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """The usage temp intervals could abandon the temp array of merged intervals"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []
    final_intervals_append = final_intervals.append

    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    a_temp_interval = None
    b_temp_interval = None
    start, end = None, None
    bit_count = 0

    while i < a_intervals_len and j < b_intervals_len:
        if a_temp_interval is None:
            a_start, a_end = a_intervals[i]
        else:
            a_start, a_end = a_temp_interval

        if b_temp_interval is None:
            b_start, b_end = b_intervals[j]
        else:
            b_start, b_end = b_temp_interval

        if a_end < b_start:
            if start is None:
                start, end = a_start, a_end
            else:
                if a_start > end + 1:
                    final_intervals_append([start, end])
                    bit_count += end - start + 1
                    start, end = a_start, a_end
                else:
                    end = a_end
            a_temp_interval = None
            i += 1
        elif b_end < a_start:
            if start is None:
                start, end = b_start, b_end
            else:
                if b_start > end + 1:
                    final_intervals_append([start, end])
                    bit_count += end - start + 1
                    start, end = b_start, b_end
                else:
                    end = b_end
            b_temp_interval = None
            j += 1
        else:
            if a_start < b_start:
                if start is None:
                    start, end = a_start, b_start - 1
                else:
                    if a_start > end + 1:
                        final_intervals_append([start, end])
                        bit_count += end - start + 1
                        start, end = a_start, b_start - 1
                    else:
                        end = b_start - 1
                a_start = b_start
            if b_start < a_start:
                if start is None:
                    start, end = b_start, a_start - 1
                else:
                    if b_start > end + 1:
                        final_intervals_append([start, end])
                        bit_count += end - start + 1
                        start, end = b_start, a_start - 1
                    else:
                        end = a_start - 1
                b_start = a_start

            if a_end > b_end:
                a_temp_interval = [b_end + 1, a_end]
                b_temp_interval = None
                j += 1
            elif a_end < b_end:
                b_temp_interval = [a_end + 1, b_end]
                a_temp_interval = None
                i += 1
            else:
                a_temp_interval = None
                b_temp_interval = None
                i += 1
                j += 1

    while i < a_intervals_len:
        if a_temp_interval is None:
            start_interval, end_interval = a_intervals[i]
        else:
            start_interval, end_interval = a_temp_interval
            a_temp_interval = None
        if start is None:
            start, end = start_interval, end_interval
        else:
            if start_interval > end + 1:
                final_intervals_append([start, end])
                bit_count += end - start + 1
                start, end = start_interval, end_interval
            else:
                end = end_interval
        i += 1

    while j < b_intervals_len:
        if b_temp_interval is None:
            start_interval, end_interval = b_intervals[j]
        else:
            start_interval, end_interval = b_temp_interval
            b_temp_interval = None
        if start is None:
            start, end = start_interval, end_interval
        else:
            if start_interval > end + 1:
                final_intervals_append([start, end])
                bit_count += end - start + 1
                start, end = start_interval, end_interval
            else:
                end = end_interval
        j += 1

    if start is not None:
        final_intervals_append([start, end])
        bit_count += end - start + 1

    if not final_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def handle_non_overlapping(start, end, bit_count, interval_start, interval_end, final_intervals_append):
    """Appending the non_overlapping intervals into the result, and calculate the bit_count"""
    if start is None:
        start, end = interval_start, interval_end
    else:
        if interval_start > end + 1:
            final_intervals_append([start, end])
            bit_count += end - start + 1
            start, end = interval_start, interval_end
        else:
            end = interval_end
    return start, end, bit_count


def process_remaining_intervals(start, end, bit_count, intervals, idx, temp_interval, final_intervals_append):
    """Deal with the remaining intervals not processed during the first walkthrough"""
    while idx < len(intervals):
        if temp_interval is None:
            start_interval, end_interval = intervals[idx]
        else:
            start_interval, end_interval = temp_interval
            temp_interval = None
        start, end, bit_count = handle_non_overlapping(start, end, bit_count, start_interval, end_interval, final_intervals_append)
        idx += 1
    return start, end, bit_count


def logical_xor_v7(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Take out similar conditions and boundary check as a function, might be slower"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []
    final_intervals_append = final_intervals.append

    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    a_temp_interval = None
    b_temp_interval = None
    start, end = None, None
    bit_count = 0

    while i < a_intervals_len and j < b_intervals_len:
        if a_temp_interval is None:
            a_start, a_end = a_intervals[i]
        else:
            a_start, a_end = a_temp_interval

        if b_temp_interval is None:
            b_start, b_end = b_intervals[j]
        else:
            b_start, b_end = b_temp_interval

        if a_end < b_start:
            start, end, bit_count = handle_non_overlapping(start, end, bit_count, a_start, a_end, final_intervals_append)
            a_temp_interval = None
            i += 1
        elif b_end < a_start:
            start, end, bit_count = handle_non_overlapping(start, end, bit_count, b_start, b_end, final_intervals_append)
            b_temp_interval = None
            j += 1
        else:
            if a_start < b_start:
                start, end, bit_count = handle_non_overlapping(start, end, bit_count, a_start, b_start - 1, final_intervals_append)
                a_start = b_start
            if b_start < a_start:
                start, end, bit_count = handle_non_overlapping(start, end, bit_count, b_start, a_start - 1, final_intervals_append)
                b_start = a_start

            if a_end > b_end:
                a_temp_interval = [b_end + 1, a_end]
                b_temp_interval = None
                j += 1
            elif a_end < b_end:
                b_temp_interval = [a_end + 1, b_end]
                a_temp_interval = None
                i += 1
            else:
                a_temp_interval = None
                b_temp_interval = None
                i += 1
                j += 1

    start, end, bit_count = process_remaining_intervals(
        start, end, bit_count, a_intervals, i, a_temp_interval, final_intervals_append
    )
    start, end, bit_count = process_remaining_intervals(
        start, end, bit_count, b_intervals, j, b_temp_interval, final_intervals_append
    )

    if start is not None:
        final_intervals_append([start, end])
        bit_count += end - start + 1

    if not final_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        bit_idx_first = final_intervals[0][0]
        bit_idx_last = final_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_and_v0(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """does an "and" of two intervals"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    ni = get_combined_intervals_until_overlap_impossible(a_intervals, b_intervals)
    intervals = []
    bit_count = 0
    try:
        prev_int = next(ni)
        next_int: list[int]
        for next_int in ni:
            max_low: int = max(prev_int[0], next_int[0])
            min_high: int = min(prev_int[1], next_int[1])
            max_high: int = max(prev_int[1], next_int[1])
            # print(f"{prev_int=}, {next_int=}, {max_low=}, {min_high=}, {max_high=}")
            if max_low <= min_high:
                bit_count += (min_high - max_low) + 1  # intervals need the +1
                intervals.append([max_low, min_high])
            if max_high > min_high:
                prev_int = [min_high + 1, max_high]
            else:
                prev_int = next(ni)
    except StopIteration:
        pass
    try:
        bit_idx_first = intervals[0][0]
        bit_idx_last = intervals[-1][1]
    except IndexError:
        bit_idx_first = -1
        bit_idx_last = -1
    return LiteBitmaskSlots(length, bit_count, intervals, bit_idx_first, bit_idx_last)


def logical_and_v1(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Use a better function to check the overlap"""
    # Check for full overlap
    if (
        bitmask_slots_a["bit_count"] == bitmask_slots_b["bit_count"]
        and bitmask_slots_a["bit_idx_first"] == bitmask_slots_b["bit_idx_first"]
        and bitmask_slots_a["bit_idx_last"] == bitmask_slots_b["bit_idx_last"]
        and len(bitmask_slots_a["intervals"]) == len(bitmask_slots_b["intervals"])
    ):
        result = LiteBitmaskSlots(
            bitmask_slots_a["length"],
            bitmask_slots_a["bit_count"],
            bitmask_slots_a["intervals"],
            bitmask_slots_a["bit_idx_first"],
            bitmask_slots_a["bit_idx_last"],
        )
    else:
        a_intervals = bitmask_slots_a["intervals"]
        b_intervals = bitmask_slots_b["intervals"]
        length = bitmask_slots_a["length"]
        ni = get_combined_intervals_until_overlap_impossible(a_intervals, b_intervals)
        intervals = []
        intervals_append = intervals.append
        bit_count = 0
        try:
            prev_int = next(ni)
            next_int: list[int]
            for next_int in ni:
                max_low: int = max(prev_int[0], next_int[0])
                min_high: int = min(prev_int[1], next_int[1])
                max_high: int = max(prev_int[1], next_int[1])
                if max_low <= min_high:
                    bit_count += (min_high - max_low) + 1
                    intervals_append([max_low, min_high])
                if max_high > min_high:
                    prev_int = [min_high + 1, max_high]
                else:
                    prev_int = next(ni)
        except StopIteration:
            pass
        try:
            bit_idx_first = intervals[0][0]
            bit_idx_last = intervals[-1][1]
        except IndexError:
            bit_idx_first = -1
            bit_idx_last = -1
        result = LiteBitmaskSlots(length, bit_count, intervals, bit_idx_first, bit_idx_last)
    return result


def logical_and_v2(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Use a better function to check the overlap"""
    # Check for full overlap
    if (
        bitmask_slots_a["bit_count"] == bitmask_slots_b["bit_count"]
        and bitmask_slots_a["bit_idx_first"] == bitmask_slots_b["bit_idx_first"]
        and bitmask_slots_a["bit_idx_last"] == bitmask_slots_b["bit_idx_last"]
        and len(bitmask_slots_a["intervals"]) == len(bitmask_slots_b["intervals"])
    ):
        result = LiteBitmaskSlots(
            bitmask_slots_a["length"],
            bitmask_slots_a["bit_count"],
            bitmask_slots_a["intervals"],
            bitmask_slots_a["bit_idx_first"],
            bitmask_slots_a["bit_idx_last"],
        )
    else:
        a_intervals = bitmask_slots_a["intervals"]
        b_intervals = bitmask_slots_b["intervals"]
        length = bitmask_slots_a["length"]
        ni = get_combined_intervals_until_overlap_impossible(a_intervals, b_intervals)
        intervals = []
        intervals_append = intervals.append
        bit_count = 0
        try:
            prev_int = next(ni)
            next_int: list[int]
            for next_int in ni:
                prev_int0, prev_int1 = prev_int
                next_int0, next_int1 = next_int
                # max_low: int = max(prev_int[0], next_int[0])
                # min_high: int = min(prev_int[1], next_int[1])
                # max_high: int = max(prev_int[1], next_int[1])
                # max_low: int = prev_int[0] if prev_int[0] >= next_int[0] else next_int[0]
                # min_high: int = prev_int[1] if prev_int[1] < next_int[1] else next_int[1]
                # max_high: int = prev_int[1] if prev_int[1] >= next_int[1] else next_int[1]
                max_low: int = prev_int0 if prev_int0 >= next_int0 else next_int0
                min_high: int = prev_int1 if prev_int1 < next_int1 else next_int1
                max_high: int = prev_int1 if prev_int1 >= next_int1 else next_int1
                if max_low <= min_high:
                    bit_count += (min_high - max_low) + 1
                    intervals_append([max_low, min_high])
                if max_high > min_high:
                    # prev_int = [min_high + 1, max_high]
                    prev_int = (min_high + 1, max_high)
                else:
                    prev_int = next(ni)
        except StopIteration:
            pass
        try:
            bit_idx_first = intervals[0][0]
            bit_idx_last = intervals[-1][1]
        except IndexError:
            bit_idx_first = -1
            bit_idx_last = -1
        result = LiteBitmaskSlots(length, bit_count, intervals, bit_idx_first, bit_idx_last)
    return result


def logical_and_v3(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Use a better function to check the overlap"""
    # Check for full overlap
    if (
        bitmask_slots_a["bit_count"] == bitmask_slots_b["bit_count"]
        and bitmask_slots_a["bit_idx_first"] == bitmask_slots_b["bit_idx_first"]
        and bitmask_slots_a["bit_idx_last"] == bitmask_slots_b["bit_idx_last"]
        and len(bitmask_slots_a["intervals"]) == len(bitmask_slots_b["intervals"])
    ):
        result = LiteBitmaskSlots(
            bitmask_slots_a["length"],
            bitmask_slots_a["bit_count"],
            bitmask_slots_a["intervals"],
            bitmask_slots_a["bit_idx_first"],
            bitmask_slots_a["bit_idx_last"],
        )
    else:
        a_intervals = bitmask_slots_a["intervals"]
        b_intervals = bitmask_slots_b["intervals"]
        length = bitmask_slots_a["length"]
        alst = a_intervals
        blst = b_intervals
        alst_len = len(a_intervals)
        blst_len = len(b_intervals)
        intervals = []
        intervals_append = intervals.append
        bit_count = 0

        first = True
        aidx = 0
        bidx = 0
        while True:
            try:
                amin, amax = alst[aidx]
            except IndexError:
                # amin = None if len(alst) else -1
                # amax = alst[-1][1] if len(alst) else -1
                amin, amax = None, alst[-1][1] if alst_len else -1
            try:
                bmin, bmax = blst[bidx]
            except IndexError:
                bmin, bmax = None, blst[-1][1] if blst_len else -1

            if (amin is None and bmin is None) or (amin is None and bmin > amax) or (bmin is None and amin > bmax):
                break

            if bmin is None or (amin is not None and amin <= bmin):
                r = (amin, amax)
                aidx += 1
            else:
                r = (bmin, bmax)
                bidx += 1
            if first:
                prev_int = r
                first = False
            else:
                next_int = r
                prev_int0, prev_int1 = prev_int
                next_int0, next_int1 = next_int
                max_low: int = prev_int0 if prev_int0 >= next_int0 else next_int0
                min_high: int = prev_int1 if prev_int1 < next_int1 else next_int1
                max_high: int = prev_int1 if prev_int1 >= next_int1 else next_int1
                if max_low <= min_high:
                    bit_count += (min_high - max_low) + 1
                    intervals_append([max_low, min_high])
                if max_high > min_high:
                    # prev_int = [min_high + 1, max_high]
                    prev_int = (min_high + 1, max_high)
                else:
                    prev_int = r
        try:
            bit_idx_first = intervals[0][0]
            bit_idx_last = intervals[-1][1]
        except IndexError:
            bit_idx_first = -1
            bit_idx_last = -1
        result = LiteBitmaskSlots(length, bit_count, intervals, bit_idx_first, bit_idx_last)
    return result


def logical_subtract_v5(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Perform logical_subtract directly on bitmasks"""
    a_intervals = copy.copy(bitmask_slots_a["intervals"])
    b_intervals = copy.copy(bitmask_slots_b["intervals"])
    length = bitmask_slots_a["length"]
    result_intervals = []
    result_intervals_append = result_intervals.append

    i, j = 0, 0

    while i < len(a_intervals) and j < len(b_intervals):
        a_start, a_end = a_intervals[i]
        b_start, b_end = b_intervals[j]

        if a_end < b_start:
            result_intervals_append([a_start, a_end])
            i += 1
        elif b_end < a_start:
            j += 1
        else:
            if a_start < b_start:
                result_intervals_append([a_start, b_start - 1])
            if a_end > b_end:
                a_intervals[i] = [b_end + 1, a_end]
                j += 1
            elif a_end < b_end:
                i += 1
            else:
                i += 1
                j += 1

    while i < len(a_intervals):
        result_intervals_append(a_intervals[i])
        i += 1

    # Merge overlapping intervals
    if not result_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        merged_intervals = []
        merged_intervals_append = merged_intervals.append
        bit_count = 0

        current_start, current_end = result_intervals[0]
        for start, end in result_intervals[1:]:
            if start > current_end + 1:
                merged_intervals_append([current_start, current_end])
                bit_count += current_end - current_start + 1
                current_start, current_end = start, end
            else:
                current_end = max(current_end, end)

        merged_intervals_append([current_start, current_end])
        bit_count += current_end - current_start + 1

        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_xnor_v0(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Call the intervals_xor and logical_not to implement the xor function"""
    bitmask_slots = logical_xor(bitmask_slots_a, bitmask_slots_b)
    bitmask_slots = logical_not(bitmask_slots)
    return bitmask_slots


def logical_xnor_v1(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Try to perform xor function in two walkthrough"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    result_intervals = []
    result_intervals_append = result_intervals.append

    i, j = 0, 0
    last_position = -1  # Track the last position we processed

    while i < len(a_intervals) and j < len(b_intervals):
        a_start, a_end = a_intervals[i]
        b_start, b_end = b_intervals[j]

        if a_end < b_start:
            # Append intervals where both are 0s (i.e., gaps)
            if last_position + 1 < a_start:
                result_intervals_append([last_position + 1, a_start - 1])
            last_position = a_end
            i += 1
        elif b_end < a_start:
            if last_position + 1 < b_start:
                result_intervals_append([last_position + 1, b_start - 1])
            last_position = b_end
            j += 1
        else:
            start = max(a_start, b_start)
            end = min(a_end, b_end)
            result_intervals_append([start, end])
            if a_end > b_end:
                last_position = b_end
                j += 1
            elif b_end > a_end:
                last_position = a_end
                i += 1
            else:
                last_position = a_end
                i += 1
                j += 1

    # Append any remaining intervals where both are 0s
    while i < len(a_intervals):
        a_start, a_end = a_intervals[i]
        if last_position + 1 < a_start:
            result_intervals_append([last_position + 1, a_start - 1])
        last_position = a_end
        i += 1

    while j < len(b_intervals):
        b_start, b_end = b_intervals[j]
        if last_position + 1 < b_start:
            result_intervals_append([last_position + 1, b_start - 1])
        last_position = b_end
        j += 1

    last_bit = length - 1
    # last_bit = max(a_bitmask.length, b_bitmask.length) - 1
    if last_position < last_bit:
        result_intervals_append([last_position + 1, last_bit])

    # Merge overlapping intervals
    if not result_intervals:
        result = LiteBitmaskSlots(length, 0, [], -1, -1)
    else:
        merged_intervals = []
        merged_intervals_append = merged_intervals.append
        bit_count = 0

        current_start, current_end = result_intervals[0]
        for start, end in result_intervals[1:]:
            if start > current_end + 1:
                merged_intervals_append([current_start, current_end])
                bit_count += current_end - current_start + 1
                current_start, current_end = start, end
            else:
                current_end = max(current_end, end)

        merged_intervals_append([current_start, current_end])
        bit_count += current_end - current_start + 1

        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
        result = LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)
    return result


def logical_xnor_v2(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Refining the logic of the start and end of each sub_intervals"""
    a_intervals = copy.copy(bitmask_slots_a["intervals"])
    b_intervals = copy.copy(bitmask_slots_b["intervals"])
    length = bitmask_slots_a["length"]

    result_intervals = []
    result_intervals_append = result_intervals.append
    i, j = 0, 0
    while i < len(a_intervals) and j < len(b_intervals):
        a_start, a_end = a_intervals[i]
        b_start, b_end = b_intervals[j]

        if a_end < b_start:
            result_intervals_append([a_start, a_end])
            i += 1
        elif b_end < a_start:
            result_intervals_append([b_start, b_end])
            j += 1
        else:
            start = min(a_start, b_start)
            end = max(a_end, b_end)
            result_intervals_append([start, end])
            if a_end > b_end:
                a_intervals[i] = [b_end + 1, a_end]
                j += 1
            elif b_end > a_end:
                b_intervals[j] = [a_end + 1, b_end]
                i += 1
            else:
                i += 1
                j += 1

    while i < len(a_intervals):
        result_intervals_append(a_intervals[i])
        i += 1

    while j < len(b_intervals):
        result_intervals_append(b_intervals[j])
        j += 1

    last_bit = length - 1
    if result_intervals:
        last_result_bit = result_intervals[-1][1]
        if last_result_bit < last_bit:
            result_intervals_append([last_result_bit + 1, last_bit])
    else:
        result_intervals_append([0, last_bit])

    # Merge overlapping intervals and handle gaps
    if not result_intervals:
        bitmask_slots = get_bitmask_slots_zero(length)
    else:
        merged_intervals = []
        merged_intervals_append = merged_intervals.append

        start, end = result_intervals[0]
        result_intervals_len = len(result_intervals)
        for i in range(1, result_intervals_len):
            current_start, current_end = result_intervals[i]
            if current_start > end + 1:
                # Include the gap
                merged_intervals_append([start, end])
                merged_intervals_append([end + 1, current_start - 1])
                start, end = current_start, current_end
            else:
                end = max(end, current_end)
        merged_intervals_append([start, end])

        bit_count = sum(end - start + 1 for start, end in merged_intervals)
        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
        bitmask_slots = LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)
    return bitmask_slots


def logical_xnor_v3(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Another effort dealing with the overlap in both intervals"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    final_intervals = []

    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    final_intervals_append = final_intervals.append
    i, j = 0, 0

    last_end = -1

    while i < a_intervals_len and j < b_intervals_len:
        a_start, a_end = a_intervals[i]
        b_start, b_end = b_intervals[j]

        if a_end < b_start:
            if last_end < a_start - 1:
                final_intervals_append([last_end + 1, a_start - 1])
            last_end = a_end
            i += 1
        elif b_end < a_start:
            if last_end < b_start - 1:
                final_intervals_append([last_end + 1, b_start - 1])
            last_end = b_end
            j += 1
        else:
            if last_end < min(a_start, b_start) - 1:
                final_intervals_append([last_end + 1, min(a_start, b_start) - 1])

            overlap_start = max(a_start, b_start)
            overlap_end = min(a_end, b_end)
            final_intervals_append([overlap_start, overlap_end])
            last_end = overlap_end

            if a_end > b_end:
                j += 1
            elif a_end < b_end:
                i += 1
            else:
                i += 1
                j += 1

    while i < a_intervals_len:
        a_start, a_end = a_intervals[i]
        if last_end < a_start - 1:
            final_intervals_append([last_end + 1, a_start - 1])
        last_end = a_end
        i += 1

    while j < b_intervals_len:
        b_start, b_end = b_intervals[j]
        if last_end < b_start - 1:
            final_intervals_append([last_end + 1, b_start - 1])
        last_end = b_end
        j += 1

    if last_end < length - 1:
        final_intervals_append([last_end + 1, length - 1])

    if not final_intervals and not a_intervals and not b_intervals:
        final_intervals = [[0, length - 1]]

    merged_intervals = []
    for interval in final_intervals:
        if not merged_intervals or merged_intervals[-1][1] < interval[0] - 1:
            merged_intervals.append(interval)
        else:
            merged_intervals[-1][1] = max(merged_intervals[-1][1], interval[1])

    if not merged_intervals:
        bit_count = 0
        bit_idx_first = -1
        bit_idx_last = -1
    else:
        bit_count = sum(end - start + 1 for start, end in merged_intervals)
        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
    return LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)


def logical_xnor_v4(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Usage of add_intervals to both speed up the append process and avoid merge process"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    merged_intervals = []
    merged_intervals_append = merged_intervals.append
    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    last_end = -1

    def add_interval(start, end):
        if not merged_intervals or merged_intervals[-1][1] < start - 1:
            merged_intervals_append([start, end])
        else:
            merged_intervals[-1][1] = max(merged_intervals[-1][1], end)

    while i < a_intervals_len or j < b_intervals_len:
        if i < a_intervals_len:
            a_start, a_end = a_intervals[i]
        else:
            a_start, a_end = length, length

        if j < b_intervals_len:
            b_start, b_end = b_intervals[j]
        else:
            b_start, b_end = length, length

        if a_end < b_start:
            if last_end < a_start - 1:
                add_interval(last_end + 1, a_start - 1)
            last_end = a_end
            i += 1
        elif b_end < a_start:
            if last_end < b_start - 1:
                add_interval(last_end + 1, b_start - 1)
            last_end = b_end
            j += 1
        else:
            if last_end < min(a_start, b_start) - 1:
                add_interval(last_end + 1, min(a_start, b_start) - 1)

            overlap_start = max(a_start, b_start)
            overlap_end = min(a_end, b_end)
            add_interval(overlap_start, overlap_end)
            last_end = overlap_end

            if a_end > b_end:
                j += 1
            elif a_end < b_end:
                i += 1
            else:
                i += 1
                j += 1

    if last_end < length - 1:
        add_interval(last_end + 1, length - 1)

    if not merged_intervals:
        bit_count = 0
        bit_idx_first = -1
        bit_idx_last = -1
    else:
        bit_count = sum(end - start + 1 for start, end in merged_intervals)
        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
    return LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)


def logical_xnor_v5(bitmask_slots_a: LiteBitmaskSlots, bitmask_slots_b: LiteBitmaskSlots) -> LiteBitmaskSlots:
    """Avoid the merge process, and refine the logic in appending in merge intervals by updating the boundary"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]
    merged_intervals = []
    merged_intervals_append = merged_intervals.append
    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    last_end = -1

    while i < a_intervals_len or j < b_intervals_len:
        if i < a_intervals_len:
            a_start, a_end = a_intervals[i]
        else:
            a_start, a_end = length, length

        if j < b_intervals_len:
            b_start, b_end = b_intervals[j]
        else:
            b_start, b_end = length, length
        # Append the non-overlapping intervals into the result
        if a_end < b_start:
            if last_end < a_start - 1:
                if not merged_intervals or merged_intervals[-1][1] < last_end:
                    merged_intervals_append([last_end + 1, a_start - 1])
                else:
                    merged_intervals[-1][1] = a_start - 1
            last_end = a_end
            i += 1
        elif b_end < a_start:
            if last_end < b_start - 1:
                if not merged_intervals or merged_intervals[-1][1] < last_end:
                    merged_intervals_append([last_end + 1, b_start - 1])
                else:
                    merged_intervals[-1][1] = b_start - 1
            last_end = b_end
            j += 1
        else:
            if last_end < min(a_start, b_start) - 1:
                if not merged_intervals or merged_intervals[-1][1] < last_end:
                    merged_intervals_append([last_end + 1, min(a_start, b_start) - 1])
                else:
                    merged_intervals[-1][1] = min(a_start, b_start) - 1

            overlap_start = max(a_start, b_start)
            overlap_end = min(a_end, b_end)
            if not merged_intervals or merged_intervals[-1][1] < overlap_start - 1:
                merged_intervals_append([overlap_start, overlap_end])
            else:
                merged_intervals[-1][1] = overlap_end
            last_end = overlap_end

            if a_end > b_end:
                j += 1
            elif a_end < b_end:
                i += 1
            else:
                i += 1
                j += 1

    if last_end < length - 1:
        if not merged_intervals or merged_intervals[-1][1] < last_end:
            merged_intervals_append([last_end + 1, length - 1])
        else:
            merged_intervals[-1][1] = length - 1

    if not merged_intervals:
        bit_count = 0
        bit_idx_first = -1
        bit_idx_last = -1
    else:
        bit_count = sum(end - start + 1 for start, end in merged_intervals)
        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
    return LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)


def logical_xnor_v6_append_interval(merged_intervals, merge_intervals_append, last_end, new_start, new_end, bit_count):
    """A function with merge interval conditions and sum of bit_count to avoid redundant code, but it is slower"""
    if not merged_intervals or merged_intervals[-1][1] < last_end:
        merge_intervals_append([new_start, new_end])
        bit_count += new_end - new_start + 1
    else:
        bit_count += new_end - merged_intervals[-1][1]
        merged_intervals[-1][1] = new_end
    return bit_count


def logical_xnor_v6(bitmask_slots_a: LiteBitmaskSlotsLike, bitmask_slots_b: LiteBitmaskSlotsLike) -> LiteBitmaskSlots:
    """Avoid the merge process, and refine the logic in appending in merge intervals by updating the boundary"""
    a_intervals = bitmask_slots_a["intervals"]
    b_intervals = bitmask_slots_b["intervals"]
    length = bitmask_slots_a["length"]

    merged_intervals = []
    merged_intervals_append = merged_intervals.append
    a_intervals_len = len(a_intervals)
    b_intervals_len = len(b_intervals)
    i, j = 0, 0

    last_end = -1
    bit_count = 0

    while i < a_intervals_len or j < b_intervals_len:
        if i < a_intervals_len:
            a_start, a_end = a_intervals[i]
        else:
            a_start, a_end = length, length

        if j < b_intervals_len:
            b_start, b_end = b_intervals[j]
        else:
            b_start, b_end = length, length
        # Append the non-overlapping intervals into the result
        if a_end < b_start:
            if last_end < a_start - 1:
                bit_count = logical_xnor_v6_append_interval(
                    merged_intervals, merged_intervals_append, last_end, last_end + 1, a_start - 1, bit_count
                )
            last_end = a_end
            i += 1
        elif b_end < a_start:
            if last_end < b_start - 1:
                bit_count = logical_xnor_v6_append_interval(
                    merged_intervals, merged_intervals_append, last_end, last_end + 1, b_start - 1, bit_count
                )
            last_end = b_end
            j += 1
        else:
            if last_end < min(a_start, b_start) - 1:
                bit_count = logical_xnor_v6_append_interval(
                    merged_intervals, merged_intervals_append, last_end, last_end + 1, min(a_start, b_start) - 1, bit_count
                )

            overlap_start = max(a_start, b_start)
            overlap_end = min(a_end, b_end)
            bit_count = logical_xnor_v6_append_interval(
                merged_intervals, merged_intervals_append, overlap_start - 1, overlap_start, overlap_end, bit_count
            )
            last_end = overlap_end
            if a_end > b_end:
                j += 1
            elif a_end < b_end:
                i += 1
            else:
                i += 1
                j += 1

    if last_end < length - 1:
        bit_count = logical_xnor_v6_append_interval(
            merged_intervals, merged_intervals_append, last_end, last_end + 1, length - 1, bit_count
        )

    if not merged_intervals:
        bit_idx_first = -1
        bit_idx_last = -1
    else:
        bit_idx_first = merged_intervals[0][0]
        bit_idx_last = merged_intervals[-1][1]
    bitmask_slots = LiteBitmaskSlots(length, bit_count, merged_intervals, bit_idx_first, bit_idx_last)
    return bitmask_slots


class IntervalError(Exception):
    pass


def index_list_to_intervals_and_properties_v0(
    idx_lst: Union[Set[int], List[int]], limit=None, sort=True
) -> Tuple[int, IntervalList, int, int]:
    """converting list to intervals"""
    bit_count = 0
    intervals = []
    bit_idx_first = -1
    bit_idx_last = -1
    if idx_lst:
        if sort:
            if type(idx_lst) == set:
                idx_lst = sorted(idx_lst)
            else:
                idx_lst = sorted(set(idx_lst))
        bit_count = len(idx_lst)
        bit_idx_first = idx_lst[0]
        bit_idx_last = idx_lst[-1]
        count = 0

        # VERY SLOW.
        for key, group in groupby(enumerate(idx_lst), lambda ix: ix[0] - ix[1]):
            interval = [item[1] for item in group]
            intervals.append([interval[0], interval[-1]])
            count += 1
            if limit and limit >= count:
                raise Exception(f"Over limit:{limit}")
    return bit_count, intervals, bit_idx_first, bit_idx_last


def index_list_to_intervals_and_properties_v1(
    idx_lst: Union[Set[int], List[int]], limit=None, sort=True
) -> Tuple[int, list, int, int]:
    """Improved function by avoid using groupby"""
    bit_count = 0
    intervals = []
    bit_idx_first = -1
    bit_idx_last = -1
    if idx_lst:
        if sort:
            if type(idx_lst) == set:
                idx_lst = sorted(idx_lst)
            else:
                idx_lst = sorted(set(idx_lst))

        bit_count = len(idx_lst)
        bit_idx_first = idx_lst[0]
        bit_idx_last = idx_lst[-1]
        count = 0
        start = idx_lst[0]
        prev = idx_lst[0]
        len_idx_lst = len(idx_lst)
        intervals_append = intervals.append
        for i in range(1, len_idx_lst):
            current = idx_lst[i]
            if current != prev + 1:
                intervals_append([start, prev])
                start = current
                count += 1
                if limit and count >= limit:
                    raise Exception(f"Over limit: {limit}")
            prev = current
        intervals_append([start, prev])
    return bit_count, intervals, bit_idx_first, bit_idx_last


def index_list_to_intervals_and_properties_v2(
    idx_lst: Union[Set[int], List[int]], limit=None, sort=True
) -> Tuple[int, list, int, int]:
    """Refine the logic with early return and sort check"""
    bit_count = 0
    intervals = []
    bit_idx_first = -1
    bit_idx_last = -1
    if idx_lst:
        if sort:
            idx_lst = sorted(idx_lst) if isinstance(idx_lst, set) else sorted(set(idx_lst))
        else:
            idx_lst = sorted(set(idx_lst))

        intervals = []
        start = prev = idx_lst[0]
        count = 0
        len_idx_lst = len(idx_lst)
        intervals_append = intervals.append
        for i in range(1, len_idx_lst):
            current = idx_lst[i]
            if current != prev + 1:
                intervals_append([start, prev])
                start = current
                count += 1
                if limit and count >= limit:
                    raise Exception(f"Over limit: {limit}")
            prev = current

        intervals_append([start, prev])
        bit_count = len_idx_lst
        bit_idx_first = idx_lst[0]
        bit_idx_last = idx_lst[-1]
    return bit_count, intervals, bit_idx_first, bit_idx_last


def join_bitmasks_v0(bitmask_lst: list, operation) -> "LiteBitmask":
    """
    given a list of bitmasks, join them
    """
    if len(bitmask_lst) == 0:
        raise Exception("Must provide a bitmask to join with.  The length is required.")
    elif len(bitmask_lst) == 1:
        result_bitmask = bitmask_lst[0].copy()
    else:
        length = 0
        for bitmask in bitmask_lst:
            if length == 0:
                length = bitmask.length
            if bitmask.length != length:
                raise AssertionError("Cannot join bitmasks with different sizes")
        if operation in [LiteBitmask.logical_or, "or", logical_or]:
            # ordering them by bit_count
            bitmask_lst_orig = sorted(bitmask_lst, key=lambda obj: -obj.bit_count)
            first_orig_bitmask = bitmask_lst_orig[0]
            if first_orig_bitmask.length == first_orig_bitmask.bit_count:
                result_bitmask = first_orig_bitmask.copy()
            else:
                result_bitmask = reduce(logical_or, bitmask_lst)
        else:
            result_bitmask = reduce(operation, bitmask_lst)
    return result_bitmask


def join_bitmasks_v1(bitmask_lst: list, operation) -> "LiteBitmask":
    """
    Join bitmasks using heapq to improve the performance
    """
    if len(bitmask_lst) == 0:
        raise Exception("Must provide a bitmask to join with. The length is required.")
    elif len(bitmask_lst) == 1:
        result_bitmask = bitmask_lst[0].copy()
    else:
        if operation in [LiteBitmask.logical_or, "or", logical_or]:
            length = 0
            final_intervals = []
            final_intervals_append = final_intervals.append
            bit_count = 0
            min_heap = []
            heapq_heappush = heapq.heappush
            heapq_heappop = heapq.heappop

            for bitmask in bitmask_lst:
                if length == 0:
                    length = bitmask.length
                if bitmask.length != length:
                    raise AssertionError("Cannot join bitmasks with different sizes")
                if bitmask.intervals:
                    interval = bitmask.intervals[0]
                    heapq_heappush(min_heap, (interval[0], interval[1], bitmask.intervals, 1))

            while min_heap:
                start, end, current_intervals, idx = heapq_heappop(min_heap)

                while min_heap and min_heap[0][0] <= end + 1:
                    next_start, next_end, next_intervals, next_idx = heapq_heappop(min_heap)
                    end = max(end, next_end)
                    if next_idx < len(next_intervals):
                        heapq_heappush(
                            min_heap, (next_intervals[next_idx][0], next_intervals[next_idx][1], next_intervals, next_idx + 1)
                        )

                bit_count = _merge_or_append_interval(start, end, final_intervals, final_intervals_append, bit_count)

                if idx < len(current_intervals):
                    heapq_heappush(min_heap, (current_intervals[idx][0], current_intervals[idx][1], current_intervals, idx + 1))

            if final_intervals:
                bit_idx_first = final_intervals[0][0]
                bit_idx_last = final_intervals[-1][1]
                result_bitmask = LiteBitmask(length, bit_count, final_intervals, bit_idx_first, bit_idx_last, no_copy=True)
            else:
                result_bitmask = liteBitmaskZero
        else:
            result_bitmask = reduce(operation, bitmask_lst)

    return result_bitmask


def join_bitmasks_v2(bitmask_lst: list, operation) -> "LiteBitmask":
    """
    Try to avoid calling external functions to improve the performance
    """
    if len(bitmask_lst) == 0:
        raise Exception("Must provide a bitmask to join with.  The length is required.")
    elif len(bitmask_lst) == 1:
        result_bitmask = bitmask_lst[0].copy()
    else:
        if operation in [LiteBitmask.logical_or, "or", logical_or]:
            """fast bitmask combine method with logical or using heapq"""
            length = 0
            final_intervals = []
            bit_count = 0
            min_heap = []
            heapq_heappush = heapq.heappush
            heapq_heappop = heapq.heappop

            for bitmask in bitmask_lst:
                if length == 0:
                    length = bitmask.length
                if bitmask.length != length:
                    raise AssertionError("Cannot join bitmasks with different sizes")
                if bitmask.intervals:
                    interval = bitmask.intervals[0]
                    heapq_heappush(min_heap, (interval[0], interval[1], bitmask.intervals, 1))

            current_start, current_end = None, None
            while min_heap:
                start, end, current_intervals, idx = heapq_heappop(min_heap)

                if current_start is None:
                    current_start, current_end = start, end
                elif start > current_end + 1:
                    if final_intervals and final_intervals[-1][1] >= current_start - 1:
                        previous_end = final_intervals[-1][1]
                        final_intervals[-1][1] = max(final_intervals[-1][1], current_end)
                        bit_count += final_intervals[-1][1] - previous_end
                    else:
                        final_intervals.append([current_start, current_end])
                        bit_count += current_end - current_start + 1
                    current_start, current_end = start, end
                else:
                    current_end = max(current_end, end)

                while min_heap and min_heap[0][0] <= current_end + 1:
                    next_start, next_end, next_intervals, next_idx = heapq_heappop(min_heap)
                    current_end = max(current_end, next_end)
                    if next_idx < len(next_intervals):
                        heapq_heappush(
                            min_heap, (next_intervals[next_idx][0], next_intervals[next_idx][1], next_intervals, next_idx + 1)
                        )

                if idx < len(current_intervals):
                    heapq_heappush(min_heap, (current_intervals[idx][0], current_intervals[idx][1], current_intervals, idx + 1))

            if current_start is not None:
                if final_intervals and final_intervals[-1][1] >= current_start - 1:
                    previous_end = final_intervals[-1][1]
                    final_intervals[-1][1] = max(final_intervals[-1][1], current_end)
                    bit_count += final_intervals[-1][1] - previous_end
                else:
                    final_intervals.append([current_start, current_end])
                    bit_count += current_end - current_start + 1

            if final_intervals:
                bit_idx_first = final_intervals[0][0]
                bit_idx_last = final_intervals[-1][1]
                result_bitmask = LiteBitmask(length, bit_count, final_intervals, bit_idx_first, bit_idx_last, no_copy=True)
            else:
                result_bitmask = liteBitmaskZero
        else:
            result_bitmask = reduce(operation, bitmask_lst)

    return result_bitmask


def join_bitmasks_v3(bitmask_lst: list, operation) -> "LiteBitmask":
    """
    Directly operate on the heapq without saving the order of intervals.
    """
    if len(bitmask_lst) == 0:
        raise Exception("Must provide a bitmask to join with. The length is required.")
    elif len(bitmask_lst) == 1:
        result_bitmask = bitmask_lst[0].copy()
    else:
        if operation in [LiteBitmask.logical_or, "or", logical_or]:
            """Fast bitmask combine method with logical OR using heapq"""
            length = bitmask_lst[0].length

            for bitmask in bitmask_lst:
                if bitmask.length != length:
                    raise AssertionError("Cannot join bitmasks with different sizes")

            final_intervals = []
            bit_count = 0
            min_heap = []
            heapq_heappush = heapq.heappush
            heapq_heappop = heapq.heappop

            for bitmask in bitmask_lst:
                for interval in bitmask.intervals:
                    heapq_heappush(min_heap, interval)

            if not min_heap:
                pass
            else:
                current_start, current_end = heapq_heappop(min_heap)

                while min_heap:
                    start, end = heapq_heappop(min_heap)
                    if start > current_end + 1:
                        bit_count = _merge_or_append_interval(
                            current_start, current_end, final_intervals, final_intervals.append, bit_count
                        )
                        current_start, current_end = start, end
                    else:
                        current_end = max(current_end, end)

                bit_count = _merge_or_append_interval(
                    current_start, current_end, final_intervals, final_intervals.append, bit_count
                )

            if final_intervals:
                bit_idx_first = final_intervals[0][0]
                bit_idx_last = final_intervals[-1][1]
                result_bitmask = LiteBitmask(length, bit_count, final_intervals, bit_idx_first, bit_idx_last, no_copy=True)
            else:
                result_bitmask = liteBitmaskZero
        else:
            result_bitmask = reduce(operation, bitmask_lst)

    return result_bitmask


def join_bitmasks_v4(bitmask_lst: list, operation) -> "LiteBitmask":
    """
    Try to use different methods based on the average density of bitmasks for better performance.
    """
    if len(bitmask_lst) == 0:
        raise Exception("Must provide a bitmask to join with. The length is required.")
    elif len(bitmask_lst) == 1:
        result = bitmask_lst[0].copy()
    else:
        length = bitmask_lst[0].length
        for bitmask in bitmask_lst:
            if bitmask.length != length:
                raise AssertionError("Cannot join bitmasks with different sizes")

        if operation not in [LiteBitmask.logical_or, "or", logical_or]:
            result = reduce(operation, bitmask_lst)
        else:
            total_intervals = sum(len(bitmask.intervals) for bitmask in bitmask_lst)
            avg_intervals_per_bitmask = total_intervals / len(bitmask_lst)

            if avg_intervals_per_bitmask < 0.1 * length:
                final_intervals = []
                bit_count = 0
                min_heap = []
                heapq_heappush = heapq.heappush
                heapq_heappop = heapq.heappop

                for bitmask in bitmask_lst:
                    for interval in bitmask.intervals:
                        heapq_heappush(min_heap, interval)

                if not min_heap:
                    pass
                else:
                    current_start, current_end = heapq_heappop(min_heap)

                    while min_heap:
                        start, end = heapq_heappop(min_heap)

                        if start > current_end + 1:
                            bit_count = _merge_or_append_interval(
                                current_start, current_end, final_intervals, final_intervals.append, bit_count
                            )
                            current_start, current_end = start, end
                        else:
                            current_end = max(current_end, end)

                    bit_count = _merge_or_append_interval(
                        current_start, current_end, final_intervals, final_intervals.append, bit_count
                    )

            else:
                final_intervals = []
                final_intervals_append = final_intervals.append
                bit_count = 0
                min_heap = []
                heapq_heappush = heapq.heappush
                heapq_heappop = heapq.heappop

                for bitmask in bitmask_lst:
                    if bitmask.intervals:
                        interval = bitmask.intervals[0]
                        heapq_heappush(min_heap, (interval[0], interval[1], bitmask.intervals, 1))

                while min_heap:
                    start, end, current_intervals, idx = heapq_heappop(min_heap)

                    while min_heap and min_heap[0][0] <= end + 1:
                        next_start, next_end, next_intervals, next_idx = heapq_heappop(min_heap)
                        end = max(end, next_end)
                        if next_idx < len(next_intervals):
                            heapq_heappush(
                                min_heap, (next_intervals[next_idx][0], next_intervals[next_idx][1], next_intervals, next_idx + 1)
                            )

                    bit_count = _merge_or_append_interval(start, end, final_intervals, final_intervals_append, bit_count)

                    if idx < len(current_intervals):
                        heapq_heappush(
                            min_heap, (current_intervals[idx][0], current_intervals[idx][1], current_intervals, idx + 1)
                        )

            if final_intervals:
                bit_idx_first = final_intervals[0][0]
                bit_idx_last = final_intervals[-1][1]
                result = LiteBitmask(length, bit_count, final_intervals, bit_idx_first, bit_idx_last, no_copy=True)
            else:
                result = liteBitmaskZero
    return result


def join_bitmasks_v5(LiteBitmask, bitmask_lst: list, operation) -> "LiteBitmask":
    """
    Apply some early loop termination logic to improve the performance.
    DOES NOT WORK, note the <= , continue
    """
    if len(bitmask_lst) == 0:
        raise Exception("Must provide a bitmask to join with. The length is required.")
    elif len(bitmask_lst) == 1:
        result_bitmask = bitmask_lst[0].copy()
    else:
        if operation in [LiteBitmask.logical_or, "or", logical_or]:
            length = 0
            final_intervals = []
            final_intervals_append = final_intervals.append
            bit_count = 0
            min_heap = []
            heapq_heappush = heapq.heappush
            heapq_heappop = heapq.heappop

            for bitmask in bitmask_lst:
                if length == 0:
                    length = bitmask.length
                if bitmask.length != length:
                    raise AssertionError("Cannot join bitmasks with different sizes")
                if bitmask.intervals:
                    interval = bitmask.intervals[0]
                    heapq_heappush(min_heap, (interval[0], interval[1], bitmask.intervals, 1))

            current_start, current_end = None, None
            while min_heap:
                start, end, current_intervals, idx = heapq_heappop(min_heap)

                if current_start is None:
                    current_start, current_end = start, end
                elif start > current_end + 1:
                    bit_count = _merge_or_append_interval(
                        current_start, current_end, final_intervals, final_intervals_append, bit_count
                    )
                    current_start, current_end = start, end
                else:
                    if end <= current_end:
                        continue
                    current_end = max(current_end, end)

                while min_heap and min_heap[0][0] <= current_end + 1:
                    next_start, next_end, next_intervals, next_idx = heapq_heappop(min_heap)
                    if next_end <= current_end:
                        continue
                    current_end = max(current_end, next_end)
                    if next_idx < len(next_intervals):
                        heapq_heappush(
                            min_heap, (next_intervals[next_idx][0], next_intervals[next_idx][1], next_intervals, next_idx + 1)
                        )

                if idx < len(current_intervals):
                    heapq_heappush(min_heap, (current_intervals[idx][0], current_intervals[idx][1], current_intervals, idx + 1))

            if current_start is not None:
                bit_count = _merge_or_append_interval(
                    current_start, current_end, final_intervals, final_intervals_append, bit_count
                )

            if final_intervals:
                bit_idx_first = final_intervals[0][0]
                bit_idx_last = final_intervals[-1][1]
                result_bitmask = LiteBitmask(length, bit_count, final_intervals, bit_idx_first, bit_idx_last, no_copy=True)
            else:
                result_bitmask = liteBitmaskZero
        else:
            result_bitmask = reduce(operation, bitmask_lst)
    return result_bitmask


def join_bitmasks_v6(bitmask_lst: List[LiteBitmaskSlotsLike], operation, length=None) -> LiteBitmaskSlots:
    """
    Skipping completely overlapping intervals for better performance.
    """
    if len(bitmask_lst) == 0:
        if length is not None:
            result_bitmask = get_bitmask_slots_zero(length)
        else:
            raise Exception("Must provide a bitmask to join with. The length is required.")
    elif len(bitmask_lst) == 1:
        result_bitmask = copy.deepcopy(bitmask_lst[0])
    else:
        if operation in [LiteBitmask.logical_or, "or", logical_or]:
            length = 0
            final_intervals = []
            final_intervals_append = final_intervals.append
            bit_count = 0
            min_heap = []
            heapq_heappush = heapq.heappush
            heapq_heappop = heapq.heappop
            for bitmask in bitmask_lst:
                if length == 0:
                    length = bitmask["length"]
                if bitmask["length"] != length:
                    raise AssertionError("Cannot join bitmasks with different sizes")
                if bitmask["intervals"]:
                    interval = bitmask["intervals"][0]
                    heapq_heappush(min_heap, (interval[0], interval[1], bitmask["intervals"], 1))

            current_start, current_end = None, None
            while min_heap:
                start, end, current_intervals, idx = heapq_heappop(min_heap)
                if current_start is None:
                    current_start, current_end = start, end
                elif start > current_end + 1:
                    bit_count = _merge_or_append_interval(
                        current_start, current_end, final_intervals, final_intervals_append, bit_count
                    )
                    current_start, current_end = start, end
                else:
                    if end < current_end:
                        continue
                    # current_end = max(current_end, end)
                    current_end = current_end if current_end >= end else end

                while min_heap and min_heap[0][0] <= current_end + 1:
                    next_start, next_end, next_intervals, next_idx = heapq_heappop(min_heap)
                    if next_end < current_end:
                        continue
                    # current_end = max(current_end, next_end)
                    current_end = current_end if current_end >= next_end else next_end
                    if next_idx < len(next_intervals):
                        heapq_heappush(
                            min_heap, (next_intervals[next_idx][0], next_intervals[next_idx][1], next_intervals, next_idx + 1)
                        )

                if idx < len(current_intervals):
                    heapq_heappush(min_heap, (current_intervals[idx][0], current_intervals[idx][1], current_intervals, idx + 1))
            if current_start is not None:
                bit_count = _merge_or_append_interval(
                    current_start, current_end, final_intervals, final_intervals_append, bit_count
                )

            if final_intervals:
                bit_idx_first = final_intervals[0][0]
                bit_idx_last = final_intervals[-1][1]
                result_bitmask = LiteBitmaskSlots(length, bit_count, final_intervals, bit_idx_first, bit_idx_last)
            else:
                result_bitmask = get_bitmask_slots_zero(length)
        else:
            result_bitmask = reduce(operation, bitmask_lst)
    return result_bitmask


def remap_bitmask(bitmask: "LiteBitmask", remap: "LiteBitmaskSlotsLike") -> "LiteBitmask":
    length = remap.bit_count
    # create a map
    if type(remap) == LiteBitmask:
        remap_dct = {pidx: nidx for nidx, pidx in enumerate(LiteBitmask.get_idx_lst_from_bitmask(remap))}
    else:
        remap_dct = {pidx: nidx for nidx, pidx in enumerate(get_interval_bits(remap))}
    # removed unwanted bits
    bitmask_masked = logical_and(bitmask, remap)  # returns slots
    new_idx_lst = [remap_dct[pidx] for pidx in get_interval_bits(bitmask_masked)]
    return LiteBitmask.zeros_and_set_bits(length, new_idx_lst)


class BitmaskRemapScaler:
    def __init__(self, remap_mask: LiteBitmaskSlotsLike, scale=1, debug=False):
        """given a remapping mask zoom into that part of a given mask using self.remap_bitmask.
        This also supports scaling on remap masks that have a single linear tuple.
        Ex: If a bit were to represent an interface and there were 4 of them, you could set scale to 4 to scale it down
        to just nodes.  If any of the 4 given interfaces were set, it will set that node or basically all 4 of them."""
        self.remap_mask = remap_mask
        self.length = remap_mask.bit_count  # the number of bits set in the remap mask.
        self.scale = scale
        self.remap_dct = {
            pidx: math.floor(nidx / scale)
            for nidx, pidx in enumerate(get_interval_bits(remap_mask))
            if pidx >= remap_mask.bit_idx_first and pidx <= remap_mask.bit_idx_last
        }
        # self.remap_dct = {pidx: math.floor(nidx / 8) for nidx, pidx in enumerate(get_interval_bits(remap_mask))}
        self.debug = debug

    def remap_bitmask(self, bitmask: "LiteBitmask") -> "LiteBitmask":
        if len(self.remap_mask.intervals) > 1:
            if self.debug:
                print("BitmaskRemapScaler: using slow method due to split remap mask. ")
            if self.scale != 1:
                raise Exception("scaling is not supported.")
            remapped_bitmask = remap_bitmask(bitmask, self.remap_mask)
        else:
            first = self.remap_mask.bit_idx_first
            last = self.remap_mask.bit_idx_last
            if bitmask.bit_idx_first >= first and bitmask.bit_idx_last <= last:
                intervals = [[self.remap_dct[x] for x in intr] for intr in bitmask.intervals]
            else:
                # we need to tear apart the tuples.
                intervals = []
                for a, b in bitmask.intervals:
                    if b < first:
                        #              f l
                        #   a        b
                        continue
                    elif a > last:
                        #   f l
                        #        a  b
                        continue
                    elif a < first <= b:
                        #      f l
                        #     f l
                        #   a  b
                        a = first
                    elif a < last <= b:
                        #      f l
                        #     f l
                        #      a b
                        b = last
                    intervals.append([self.remap_dct[a], self.remap_dct[b]])
            if len(intervals) == 0:
                remapped_bitmask = get_bitmask_zero(self.length)
            else:
                remapped_bitmask = LiteBitmask.zeros_and_set_intervals(self.length, intervals)
        return remapped_bitmask


def repr_LiteBitmask_zasb(bitmask: "LiteBitmask"):
    bits = (str(i) for i in get_interval_bits(bitmask))
    result = f"LiteBitmask.zeros_and_set_bits({bitmask.length}, ({', '.join(bits)}))"
    return result


def repr_LiteBitmask_zasi(bitmask: "LiteBitmask"):
    result = f"LiteBitmask.zeros_and_set_intervals({bitmask.length}, {str(bitmask.intervals)})"
    return result


class BitmaskFlags:
    """created for compatibility for LiteBitmask with EnhancedBitmask"""

    # writeable = False
    keys = ("writeable",)

    @property
    def writeable(self):
        return False

    @writeable.setter
    def writeable(self, value):
        assert value == False

    @staticmethod
    def from_dict(dct):
        bf = BitmaskFlags()
        for key, value in dct.items():
            if key in BitmaskFlags.keys:
                setattr(bf, key, value)
            else:
                raise Exception(f"invalid key {key} for BitmaskFlags")
        return bf

    def to_dict(self):
        return dict(writeable=self.writeable)


readonlyBitmaskFlags = BitmaskFlags()


class LiteBitmask(EnhancedBitmaskLike, LiteBitmaskSlotsLike):
    """This is LiteBitmask which was created after EnhancedBitmask to solve a speed problem with processing bitmasks.
    EnhancedBitmaskLike was created to allow some compatibility.  LiteBitmaskSlotsLike was created to allow LiteBitmask
    and LiteBitmaskSlots to be able to interact.  There is a bit of a hack inside __getitem__ that allows LiteBitmask
    to act like a dictionary and allow accessing the length by obj.length or obj['length'].

    Some of the problems solved:
        Serialization       Used to save the bitmask in memory(spark) or in a datalake/database
        Intersection Any    Used in a search to find a row that may be associated with that bitmask
        Subtraction         Used to remove a part of a bitmask and search with the new mask
        OR                  Used to join bitmasks
        XNOR                Used for intersection all
        Join Bitmasks       Used to join many bitmasks
    Remaining Problems:
        LiteBitmaskSlots and LiteBitmask are quite the same, but different due to EnhancedBitmaskLike vs dict
        Possibly remove all the logical_* classmethods, __and__, __or__, etc, and use the normal functions but
            this would break compatibility with EnhancedBitmaskLike.
    """

    length: int = 0  # same as size
    bit_count: int = 0
    bit_idx_first = -1
    bit_idx_last = -1
    intervals = []
    flags = None
    repr_func = repr_LiteBitmask_zasi

    def __init__(self, length, bit_count, intervals, bit_idx_first, bit_idx_last, no_copy=True):
        self.length = length
        self.bit_count = bit_count
        if no_copy:
            self.intervals = intervals
        else:
            self.intervals = copy.deepcopy(intervals)
        self.bit_idx_first = bit_idx_first
        self.bit_idx_last = bit_idx_last
        self.flags = readonlyBitmaskFlags
        if self.bit_idx_last > (self.length - 1):
            raise Exception(f"Bit set out of range: {self.bit_idx_last} length:{self.length}")

    def __setstate__(self, state):
        """used to recreate the object after pickle"""
        self.length = state["length"]
        self.bit_count = state["bit_count"]
        self.bit_idx_first = state["bit_idx_first"]
        self.bit_idx_last = state["bit_idx_last"]
        self.intervals = state["intervals"]
        self.flags = readonlyBitmaskFlags

    def __getstate__(self):
        """used to dump the state of an object for pickling"""
        return self.to_dict()

    def intersection_any(self, other: EnhancedBitmaskLike):
        return intersection_any(self, other)

    def __repr__(self):
        # https://github.com/pandas-dev/pandas/issues/17695
        # from pandas.io.formats import printing as pd_printing
        # pd_printing.is_sequence = lambda obj: False
        return LiteBitmask.repr_func(self)

    def __getitem__(self, item):
        # fixme: this doesn't handle negatives in the slices.
        if item in ["length", "bit_count", "intervals", "bit_idx_first", "bit_idx_last"]:
            # this happens when a bitmask is treated like a dictionary.
            result = getattr(self, item)
        elif not isinstance(item, slice):
            if item >= self.length:
                raise IndexError
            # TODO: do a binary search or some neat source.
            result = False
            for start, end in self.intervals:
                if start <= item <= end:
                    result = True
                    break
        else:
            lst = []
            c_idx = 0
            # if item.start >= self.length or item.stop >= self.length:
            #     pass  # if the array is
            start = item.start
            stop = item.stop
            # if you ask beyond the end of the array, isolate the range to the stop.
            if stop >= self.length:
                stop = self.length
            step = item.step
            rangeish = [i for i in [start, stop, step] if i is not None]
            for idx in range(*rangeish):  # Nlog(M)
                found = False
                for i, (start, end) in enumerate(self.intervals[c_idx:]):
                    if start <= idx <= end:
                        found = True
                        c_idx = c_idx + i
                        break
                lst.append(found)
            result = lst
        return result

    def __or__(self, other) -> "LiteBitmask":
        return LiteBitmask.logical_or(self, other)

    def __and__(self, other) -> "LiteBitmask":
        return LiteBitmask.logical_and(self, other)

    def __sub__(self, other):
        bitmask_slots = logical_subtract(self, other)
        bitmask = LiteBitmask.from_slots(bitmask_slots)
        return bitmask

    def __eq__(self, o: "LiteBitmask") -> "LiteBitmask":
        """this returns an array of equality.
        for full equality use:
        ex: (bitmask_left == bitmask_right).all()"""
        if type(o) == LiteBitmask:
            lst = map(lambda a: a[0] == a[1], zip(self.tolist(), o.tolist()))
            bits = [idx for idx, value in enumerate(lst) if value]
            result = LiteBitmask.zeros_and_set_bits(self.length, bits)
        else:
            result = super().__eq__(o)
        return result

    def copy(self, *args, **kwargs):
        return LiteBitmask.zeros_and_set_intervals(self.length, self.intervals)

    def validate(self):
        validate_litebitmaskslotslike(self)

    @staticmethod
    def get_bitmask_from_idx_lst(bit_count, idx_lst) -> "LiteBitmask":
        """create a bitmask from an index list"""
        return LiteBitmask.zeros_and_set_bits(bit_count, idx_lst)

    @staticmethod
    def get_idx_lst_from_bitmask(bitmask, bit_idx_first: int = 0, bit_idx_last: int = None) -> list:
        """return the indexes set inside the bitmask"""
        if isinstance(bitmask, LiteBitmask):
            bit_lst = [_ for _ in get_interval_bits(bitmask.to_slots(), bit_idx_first=bit_idx_first)]
        else:
            bit_lst = EnhancedBitmask.get_idx_lst_from_bitmask(bitmask, bit_idx_first=bit_idx_first, bit_idx_last=bit_idx_last)
        return bit_lst

    def tolist(self):
        """genereate the full bitmask as a list."""
        bit_idx_gen = LiteBitmask.get_idx_lst_from_bitmask(self)
        lst: list[bool] = [False] * self.length
        for idx in bit_idx_gen:
            lst[idx] = True
        return lst

    @staticmethod
    def array(lst, **kwargs) -> "LiteBitmask":
        bits = [idx for idx, value in enumerate(lst) if value]
        bitmask = LiteBitmask.zeros_and_set_bits(len(lst), bits)
        return bitmask

    @staticmethod
    def zeros(length: int) -> "LiteBitmask":
        return get_bitmask_zero(length).copy()

    @staticmethod
    def ones(length: int) -> "LiteBitmask":
        return LiteBitmask(length, length, [[0, length - 1]], 0, length - 1)

    @staticmethod
    def zeros_and_set_bits(length: int, bit_lst: Union[List[int], Tuple[int, ...], Set[int]], **kwargs):
        return LiteBitmask(length, *index_list_to_intervals_and_properties(bit_lst))

    @staticmethod
    def zeros_and_set_ranges(length: int, bit_ranges: list[Union[list[int, int], list[int, int, int]]]):
        """given a tuple of range slices, create a bitmask.
        range slices are a->b excluding b"""
        bit_indexes = sorted(set(i for bit_range in bit_ranges for i in range(*bit_range)))
        return LiteBitmask(length, *index_list_to_intervals_and_properties(bit_indexes))

    @staticmethod
    def zeros_and_set_intervals(length: int, intervals: IntervalList, validate=False, merge=True) -> "LiteBitmask":
        """given a tuple of intervals, create a bitmask.
        intervals are a->b including b
        do_sort_merge will sort the intervals and remove duplicates"""
        return convert_intervals_list_to_LiteBitmask(length, intervals, validate=validate, merge=merge)

    @staticmethod
    def zeros_and_set_bit(length: int, bit_idx):
        return LiteBitmask(length, 1, [[bit_idx, bit_idx]], bit_idx, bit_idx, no_copy=True)

    @staticmethod
    def from_dict(dct):
        bitmask = LiteBitmask(dct["length"], dct["bit_count"], dct["intervals"], dct["bit_idx_first"], dct["bit_idx_last"])
        bitmask.flags = readonlyBitmaskFlags
        return bitmask

    def to_dict(self):
        dct = {
            "length": self.length,
            "bit_count": self.bit_count,
            "bit_idx_first": self.bit_idx_first,
            "bit_idx_last": self.bit_idx_last,
            "intervals": self.intervals,
        }
        return dct

    @classmethod
    def from_spark(cls, dct, no_copy=False) -> "LiteBitmask":
        bit_count = dct["bit_count"]
        intervals = dct["intervals"]
        length = dct["length"]
        bit_idx_first = dct["bit_idx_first"]
        bit_idx_last = dct["bit_idx_last"]
        bitmask = LiteBitmask(length, bit_count, intervals, bit_idx_first, bit_idx_last, no_copy=no_copy)
        return bitmask

    def to_slots(self) -> LiteBitmaskSlots:
        return LiteBitmaskSlots(self.length, self.bit_count, self.intervals, self.bit_idx_first, self.bit_idx_last)

    @classmethod
    def from_slots(cls, obj: LiteBitmaskSlots):
        return LiteBitmask(obj.length, obj.bit_count, obj.intervals, obj.bit_idx_first, obj.bit_idx_last)

    def to_json(self):
        json_str = simplejson.dumps(self, cls=DataEncoder)
        return json_str

    @staticmethod
    def from_json(json_str):
        obj = simplejson.loads(json_str, cls=DataDecoder)
        return obj

    def __len__(self):
        return self.length

    def __add__(self, other):
        bitmask_slots = logical_or(self.to_slots(), other.to_slots())
        return LiteBitmask.from_slots(bitmask_slots)

    def __xor__(self, other):
        return LiteBitmask.logical_xor(self, other)

    @staticmethod
    def add(x1, x2, **kwargs):
        return LiteBitmask.logical_or(x1, x2)

    @staticmethod
    def logical_xor(x: "LiteBitmask", y: "LiteBitmask") -> "LiteBitmask":
        bitmask_slots = logical_xor(x.to_slots(), y.to_slots())
        return LiteBitmask.from_slots(bitmask_slots)

    @staticmethod
    def logical_not(x: "LiteBitmask") -> "LiteBitmask":
        bitmask_slots = logical_not(x.to_slots())
        return LiteBitmask.from_slots(bitmask_slots)

    @staticmethod
    def logical_xnor(x: "LiteBitmask", y: "LiteBitmask") -> "LiteBitmask":
        bitmask_slots = logical_xnor(x.to_slots(), y.to_slots())
        return LiteBitmask.from_slots(bitmask_slots)

    def intersection_all(self, other):
        return intersection_all(self, other)

    @staticmethod
    def logical_or(x: "LiteBitmask", y: "LiteBitmask") -> "LiteBitmask":
        bitmask_slots = logical_or(x.to_slots(), y.to_slots())
        return LiteBitmask.from_slots(bitmask_slots)

    @staticmethod
    def logical_subtract(x: "LiteBitmask", y: "LiteBitmask") -> "LiteBitmask":
        bitmask_slots = logical_subtract(x.to_slots(), y.to_slots())
        return LiteBitmask.from_slots(bitmask_slots)

    @staticmethod
    def logical_and(x: "LiteBitmask", y: "LiteBitmask") -> "LiteBitmask":
        if intersection_possible(x, y):
            bitmask_slots = logical_and(x.to_slots(), y.to_slots())
            bitmask = LiteBitmask.from_slots(bitmask_slots)
        else:
            bitmask = LiteBitmask.zeros(x.length)
        return bitmask

    def intersection(self, other):
        if intersection_possible(self, other):
            real_intersection = LiteBitmask.logical_and(self, other)
        else:
            real_intersection = None
        return real_intersection

    @classmethod
    def logical_is_subset(cls, left, right):
        # FIXME: see EnhancedBitmask for a much faster version.
        result = cls.logical_and(right, cls.logical_not(left))
        return result

    @classmethod
    def join_bitmasks(cls, bitmask_lst: list, operation) -> "LiteBitmask":
        """joining a list of bitmasks"""
        bitmask_slots = join_bitmasks(bitmask_lst, operation)
        return cls.from_slots(bitmask_slots)

    @staticmethod
    def sign(x1) -> "EnhancedBitmaskLike":
        return x1.copy()


DataEncoder.register(LiteBitmask, LiteBitmask.to_dict)
DataDecoder.register("LiteBitmask", LiteBitmask.from_dict)


def series_intersection_any(ser, search_bitmask: Union[EnhancedBitmaskLike, str]):
    """function to apply to a pandas series of bitmasks to search for an intersection."""
    return [bitmask.intersection_any(search_bitmask) if bitmask != "" else False for bitmask in ser]


"""Best versions so far"""
logical_subtract = logical_subtract_v2  # logical_subtract_v5
logical_xor = logical_xor_v7
logical_and = logical_and_v3
logical_xnor = logical_xnor_v6
logical_or = logical_or_v8
join_bitmasks = join_bitmasks_v6
intersection_any = intersection_any_v1
get_interval_bits_generator = get_interval_bits_generator_v1
_merge_or_append_interval = _merge_or_append_interval_v8
index_list_to_intervals_and_properties = index_list_to_intervals_and_properties_v2
get_combined_intervals_until_all_consumed = get_combined_intervals_until_all_consumed_v3

liteBitmaskZero = LiteBitmask(0, 0, [], -1, -1)
liteBitmaskSlotsZero = LiteBitmaskSlots(0, 0, [], -1, -1)


lite_bitmask_zero_cache = {}


def get_bitmask_zero(length) -> LiteBitmask:
    """cache the zero value bitmasks for later use.
    Warning, if you provide length as a float, it will string it when doing the dictionary hash lookup.
    When remapping bitmasks and dividing, this can cause problems when creating masks.
    Ex: 10.0 -> 10"""
    try:
        bitmask = lite_bitmask_zero_cache[length]
    except KeyError:
        bitmask = LiteBitmask(length, 0, [], -1, -1)
        lite_bitmask_zero_cache[length] = bitmask
    return bitmask


lite_bitmask_slots_zero_cache = {}


def get_bitmask_slots_zero(length) -> LiteBitmaskSlots:
    try:
        bitmask_slots = lite_bitmask_slots_zero_cache[length]
    except KeyError:
        bitmask_slots = LiteBitmaskSlots(length, 0, [], -1, -1)
        lite_bitmask_slots_zero_cache[length] = bitmask_slots
    return bitmask_slots


lite_bitmask_ones_cache = {}


def get_bitmask_slots_ones(length) -> LiteBitmaskSlots:
    try:
        bitmask_slots = lite_bitmask_ones_cache[length]
    except KeyError:
        bitmask_slots = LiteBitmaskSlots(length, length, [0, length - 1], 0, length - 1)
        lite_bitmask_ones_cache[length] = bitmask_slots
    return bitmask_slots


"""
An example of using arrays to contain different function names
intervals_subtract_funcs = [
    ('intervals_subtract_v0', intervals_subtract_v0),
    ('intervals_subtract_v1', intervals_subtract_v1),
    ('intervals_subtract_v2', intervals_subtract_v2),
]

intervals_subtract_funcs =['intervals_subtract_v0', 'intervals_subtract_v1', 'intervals_subtract_v2']
"""
