#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import datetime
from Octeres.forthwith import FORMAT_DATE_FULL


def binary_search_file(log, fo, datetime_length, datetime_format, ts, pos_start, pos_end, max_delta=1, max_depth=32, depth=0, direction="-",
                       buffer_back=0, buffer_forward=0):
    """binary search the file for the ts, recursively.
    pos_start is generally 0 to start with
    pos_end is generally the size of the file.
    buffer_back, jumps backwards that amount if max_depth is reached.
    buffer_forward, jumps forward that amount if max_depth is reached.
    only one of buffer_back or buffer_forward can be used a time.
    todo: if wrapped in a class, the kwarg count might be better and less items on the stack.
    """
    delta = pos_end - pos_start
    log.debug("%-16s %3s %16s    Δ:%16s" % (pos_start, direction, pos_end, delta))
    if delta <= max_delta:
        pos = fo.tell()
        log.debug(f"{delta=} found for {pos_end=} - {pos_start=}, {pos}")
    elif depth > max_depth:
        line = fo.readline()
        pos = fo.tell()
        log.debug(f"max_depth found for {depth=} > {max_depth=}, {pos}")
    else:
        pos_middle = pos_start + int((pos_end - pos_start) / 2)
        fo.seek(pos_middle)
        # this will most likely be a partial line
        line = fo.readline()
        # this should be the full line
        search_count = 128
        line_ts = None
        bad_lines = []
        while search_count > 0:
            line = fo.readline()
            line_ts_raw = line[:datetime_length]
            line_ts = datetime.datetime.strptime(line_ts_raw, datetime_format)
            try:
                line_ts = datetime.datetime.strptime(line_ts_raw, datetime_format)
                # line_ts = dp(line_ts_raw)
            except ValueError:
                # keep searching
                search_count = search_count - 1
                bad_lines.append((fo.tell(), line.strip()))
            else:
                # we found it
                search_count = 0
        if line_ts is None:
            log.debug("*** We tried to find a date, but could not ***")
            for line_pos, line in bad_lines:
                log.debug("LP#%s:%s" % (line_pos, line))
            raise Exception("No dates found in the file.")
        at_max_depth = (depth + 1) > max_depth
        if at_max_depth:
            log.debug(f"at max depth, using buffers if they exist.")
            if line_ts == ts:
                log.debug(f"Found the exact ts:{ts.strftime(FORMAT_DATE_FULL)}")
                pos = pos_middle
            elif buffer_back:
                log.debug(f"buffering backwards")
                fo.seek(pos_start - buffer_back)
                pos = binary_search_file(
                    log,
                    fo,
                    datetime_length,
                    datetime_format,
                    ts,
                    pos_start - buffer_back,
                    pos_middle,
                    max_delta=max_delta,
                    max_depth=max_depth,
                    depth=depth + 1,
                    direction="<-",
                    buffer_back=buffer_back,
                    buffer_forward=buffer_forward,
                    )
            else:  # lean_forward
                log.debug(f"buffering forwards")
                fo.seek(pos_end + buffer_forward)
                pos = binary_search_file(
                    log,
                    fo,
                    datetime_length,
                    datetime_format,
                    ts,
                    pos_middle,
                    pos_end + buffer_forward,
                    max_delta=max_delta,
                    max_depth=max_depth,
                    depth=depth + 1,
                    direction="->",
                    buffer_back=buffer_back,
                    buffer_forward=buffer_forward,
                    )
        else:
            if line_ts < ts:
                # search forward
                pos = binary_search_file(
                    log,
                    fo,
                    datetime_length,
                    datetime_format,
                    ts,
                    pos_middle,
                    pos_end,
                    max_delta=max_delta,
                    max_depth=max_depth,
                    depth=depth + 1,
                    direction="->",
                    buffer_back=buffer_back,
                    buffer_forward=buffer_forward,
                )
            elif line_ts > ts:
                pos = binary_search_file(
                    log,
                    fo,
                    datetime_length,
                    datetime_format,
                    ts,
                    pos_start,
                    pos_middle,
                    max_delta=max_delta,
                    max_depth=max_depth,
                    depth=depth + 1,
                    direction="<-",
                    buffer_back=buffer_back,
                    buffer_forward=buffer_forward,
                )
            else:
                log.debug(f"Found the exact ts:{ts.strftime(FORMAT_DATE_FULL)}")
                pos = pos_middle
    return pos


def read_cut(file_name, pos_start, pos_end):
    """read from start to end"""
    lines = []
    with open(file_name, "r") as fo:
        fo.seek(pos_start)
        line = fo.readline()
        pos = fo.tell()  # might miss first line.
        while line and pos < pos_end:
            lines.append(line)
            line = fo.readline()
            pos = fo.tell()
    return lines


def read_cut_generator(file_name, pos_start, pos_end, read_header=False):
    """read from start to end"""
    with open(file_name, "r") as fo:
        if read_header:
            header = fo.readline()
            yield header
        fo.seek(pos_start)
        line = fo.readline()
        pos = pos_start
        while line and pos < pos_end:  # convert position to integer for comparison
            yield line.strip()
            line = fo.readline()
            pos = fo.tell()
        if line:  # grab the last line.  Might be past the range.
            yield line.strip()
