#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import gzip
import logging
import math
import sys

import numpy as np
import networkx as nx
from pathlib import Path
from pprint import pformat
import click
import pickle

from Octeres.bitmask import EnhancedBitmask
from Octeres.bitmask.bitmask_graph import show_nxgraph
from Octeres.SL.modules.ltm import get_Super_Location_to_Mask
from Octeres.bitmask.bitmask_graph_search import convert_bitmask_in_graph_to
from Octeres.bitmask.bitmask_lite import LiteBitmask
from Octeres.forthwith import (
    FORMAT_DATE_DAY,
    FORMAT_DATE_SECOND_T,
    FORMAT_DATE_SECOND_T_NOCOLON,
)
from Octeres.graph import is_leaf, get_leaves


class Settings:
    def __init__(self) -> None:
        self.graph = None
        self.nodes = None
        self.roots = None
        self.filt = {}


pass_settings = click.make_pass_decorator(Settings, ensure=True)


@click.group()
@pass_settings
def cli(settings: Settings):
    pass


def cb_filt(ctx, param, value) -> dict:
    bad_args = []
    result = {}
    for filt in value:
        try:
            col, value = filt.split("=", 1)
        except ValueError:
            bad_args.append(filt)
        else:
            result[col] = value
    if bad_args:
        raise click.BadParameter(f"filter must be key=value.  Bad filts:{','.join(bad_args)}")
    return result


@cli.group()
@click.option(
    "-g",
    "--graph",
    type=click.Path(exists=True, dir_okay=False, path_type=Path, resolve_path=True),
    help="graph pickle",
    required=True,
)
@click.option("--filt", "-f", help="--filt queue=workq", multiple=True, callback=cb_filt)
@pass_settings
def show(settings: Settings, graph: Path, filt: dict):
    try:
        nxgraph = pickle.loads(graph.read_bytes())
    except:
        with gzip.open(graph, "rb") as f:
            nxgraph = pickle.loads(f.read())
    roots = [n for n, d in nxgraph.in_degree() if d == 0]
    nodes = [n for n, d in nxgraph.in_degree() if d != 0]
    settings.graph = nxgraph
    settings.nodes = nodes
    settings.roots = roots
    settings.filt = filt


def get_ltm_graph(machine_name, range_start, bitmask_class, debug=False):
    sltm = get_Super_Location_to_Mask(filt={"name": machine_name, "pit": range_start}, bitmask_class=bitmask_class)
    ltm_obj = sltm.ltm_object_factory(machine_name, range_start, bitmask_class=bitmask_class, debug=debug)
    rs = ltm_obj.range_start
    re = ltm_obj.range_end
    return ltm_obj.graph, rs, re


@cli.command()
@click.argument("machine-name", type=click.STRING, required=True)
@click.option(
    "-s",
    "--range-start",
    type=click.DateTime(formats=[FORMAT_DATE_DAY, FORMAT_DATE_SECOND_T]),
    required=True,
)
@click.option(
    "-b",
    "--include-bitmask",
    is_flag=True,
    show_default=True,
    default=False,
    help="include bitmask",
)
@click.option(
    "-e",
    "--include-ebitmask",
    is_flag=True,
    show_default=True,
    default=False,
    help="include EnhancedBitmask",
)
@click.option(
    "-l",
    "--include-lbitmask",
    is_flag=True,
    show_default=True,
    default=False,
    help="include LiteBitmask",
)
@click.option("--gz", is_flag=True, show_default=True, default=False, help="gzip graph")
@click.option("--debug", default=False, is_flag=True)
def export(machine_name, range_start, include_bitmask, include_ebitmask, include_lbitmask, gz, debug):
    if len([i for i in [include_lbitmask, include_bitmask, include_ebitmask] if i]) > 1:
        print(f"select only one bitmask type.")
        sys.exit(1)
    # todo, if we want a LiteBitmask, creating the ltm graph might be much quicker.
    if include_bitmask:
        bitmask_class = np.ndarray
    elif include_ebitmask:
        bitmask_class = EnhancedBitmask
    elif include_lbitmask:
        bitmask_class = LiteBitmask
    else:
        bitmask_class = None
    graph, rs, re = get_ltm_graph(machine_name, range_start, bitmask_class, debug=debug)
    # graph = convert_bitmask_in_graph_to(graph, bitmask_type)
    filename = f"graph_{machine_name}_{rs.strftime(FORMAT_DATE_SECOND_T_NOCOLON)}.pickle"
    if gz:
        filename = f"{filename}.gz"
        with gzip.open(filename, "wb") as fd:
            pickle.dump(graph, fd)
            # fd.write(pickle.dumps(graph))
    else:
        with open(filename, "wb") as fd:
            pickle.dump(graph, fd)
            # fd.write(pickle.dumps(graph))
    print(f"wrote {filename}")


@show.command()
@click.argument("node", type=click.STRING, required=True)
@pass_settings
def node_show(settings: Settings, node: str):
    print(f"Node:{node}")
    print(pformat(settings.graph.nodes[node]))
    print(pformat(list(settings.graph.successors(node))))


@show.command()
@pass_settings
def graph_show(settings: Settings):
    graph = settings.graph
    if settings.filt:
        remove_lst = []
        for node in graph.nodes:
            result = False
            node_dct = graph.nodes[node]
            for key, value in settings.filt.items():
                if node_dct[key] == value:
                    result = True
            if result is False:
                remove_lst.append(node)
        for node in remove_lst:
            graph.remove_node(node)

    # figure out the color of the nodes based on the bitmask.
    nodelist = []
    node_color = []
    base_color = 0x66
    color_increment = 0x11
    max_color = 0xFF
    lookup = {}
    for node in graph.nodes:
        node_dct = graph.nodes[node]
        if "bitmask" in node_dct:
            bitmask = node_dct["bitmask"]
            lst = bitmask.tolist()
            ones = [1 for _ in lst if _ == 1]
            node_dct["bit_count"] = len(ones)
            color_number = base_color + round(color_increment * math.log(len(ones), 2))
            color_number_offset = color_number + color_increment
            color_number = color_number if color_number < max_color else max_color
            color_number_offset = color_number_offset if color_number_offset < max_color else max_color
            color = hex(color_number).replace("0x", "")
            coloroff = hex(color_number_offset).replace("0x", "")
            if node_dct["obj_type"] == "link":
                color = f"#{color}{coloroff}{color}"
            elif node_dct["obj_type"] == "switch":
                color = f"#{coloroff}{color}{color}"
            else:
                color = f"#{color}{color}{coloroff}"
        else:
            node_dct["bit_count"] = None
            color = "#aaaadd"
        if node_dct['bit_count'] == 1:
            new_node_name = f"{node}:{node_dct['bitmask'].intervals}"
        else:
            new_node_name = node
        nodelist.append(new_node_name)
        # nodelist.append(node)
        lookup[node] = new_node_name
        node_color.append(color)
    node_dct = dict(nodelist=nodelist, node_color=node_color, node_size=960, alpha=0.7)
    edge_dct = dict(node_size=960)
    graph = nx.relabel_nodes(graph, lookup)
    show_nxgraph(graph, node_dct=node_dct, edge_dct=edge_dct)


@cli.command()
@click.argument("machine-name", type=click.STRING, required=True)
@click.option(
    "-s",
    "--range-start",
    type=click.DateTime(formats=[FORMAT_DATE_DAY, FORMAT_DATE_SECOND_T]),
    required=True,
)
def export_leaves_to_node(machine_name, range_start):
    """given a machine and start, return a lookup of leaves to node"""
    sltm = get_Super_Location_to_Mask(filt={"name": machine_name, "pit": range_start})
    ltm_obj = sltm.ltm_object_factory(machine_name, range_start)
    graph = ltm_obj.graph
    # leaves <- node, if a leaf, leaf == node
    block_lookup = {}
    descendants = sorted(nx.algorithms.dag.descendants(graph, machine_name))
    for node in descendants:
        leaves = get_leaves(graph, node)
        if is_leaf(graph, node):
            block_lookup[node] = node
        else:
            block_lookup[",".join(sorted(leaves))] = node
    print(pformat(block_lookup, width=256))


@cli.command()
@click.argument("machine-name", type=click.STRING, required=True)
@click.option(
    "-s",
    "--range-start",
    type=click.DateTime(formats=[FORMAT_DATE_DAY, FORMAT_DATE_SECOND_T]),
    required=True,
)
def export_node_to_leaves(machine_name, range_start):
    """given a machine and start, return a lookup of leaves to node"""
    sltm = get_Super_Location_to_Mask(filt={"name": machine_name, "pit": range_start})
    ltm_obj = sltm.ltm_object_factory(machine_name, range_start)
    graph = ltm_obj.graph
    # leaves <- node, if a leaf, leaf == node
    block_lookup = {}
    descendants = sorted(nx.algorithms.dag.descendants(graph, machine_name))
    for node in descendants:
        leaves = get_leaves(graph, node)
        if is_leaf(graph, node):
            block_lookup[node] = node
        else:
            block_lookup[node] = leaves
    print(pformat(block_lookup, width=256))


if __name__ == "__main__":
    cli()
