#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import gc
import time
import functools
from typing import List

import click
import pandas as pd
import matplotlib.pyplot as plt
import memory_profiler
import random
import seaborn as sns

from Octeres.bitmask.bitmask_lite import logical_xnor_v0, logical_xnor_v1, logical_xnor_v2, logical_xnor_v3, logical_xnor_v4, \
    logical_xnor_v5, logical_xnor_v6, convert_intervals_to_LiteBitmaskSlots, LiteBitmask, liteBitmaskSlotsZero, LiteBitmaskSlots, \
    logical_or_v7, get_combined_intervals_until_all_consumed, get_combined_intervals_until_all_consumed_v1, \
    get_combined_intervals_until_all_consumed_v2, join_bitmasks_v0, join_bitmasks_v2, join_bitmasks_v6, logical_or_v0, \
    logical_or_v4, logical_or_v5, logical_or_v6, logical_or_v8, logical_and_v0, logical_and_v1, logical_and_v2, logical_and_v3, \
    get_combined_intervals_until_all_consumed_v3

pd.set_option("display.max_rows", None)
pd.set_option("display.max_columns", None)
pd.set_option("display.width", 512)
pd.set_option("display.max_colwidth", None)
pd.set_option("display.float_format", "{:.6f}".format)


class Settings:
    def __init__(self) -> None:
        self.topic_lst = []


pass_settings = click.make_pass_decorator(Settings, ensure=True)


@click.group()
@pass_settings
def cli(settings: Settings):
    pass


"""A big set of intervals pair test sample."""
interval_pair_input_deck = [
    ([[i, i + 1] for i in range(0, 99999, 4)], [[i, i + 1] for i in range(2, 99999, 4)], "no_overlap_even"),
    ([[i, i + 1] for i in range(0, 99999)], [[i, i + 1] for i in range(0, 99999)], "full_overlap"),
    ([[i, i + 1] for i in range(0, 99999, 10)], [[i, i + 1] for i in range(5, 99999, 10)], "sparse_overlap"),
    (
        sorted([[i, i + 1] for i in random.sample(range(0, 99999), 50000)]),
        sorted([[i, i + 1] for i in random.sample(range(0, 99999), 50000)]),
        "random",
    ),
    ([[i, i + 1] for i in range(0, 99999, 2)], [[i, i + 1] for i in range(1, 99999, 2)], "alternating"),
    ([[i, i + 5] for i in range(0, 99999, 10)], [[i + 2, i + 3] for i in range(0, 99999, 10)], "nested"),
    ([[i, i + 1] for i in range(0, 99999, 3)], [[i, i + 1] for i in range(1, 99999, 4)], "mixed"),
    ([[i, i + 100] for i in range(0, 99999, 500)], [[i, i + 100] for i in range(250, 99999, 500)], "large_intervals"),
    ([[i, i + 1] for i in range(0, 99999, 3)], [[i, i + 1] for i in range(500, 99999, 3)], "3x3"),
    ([[i, i + 1] for i in range(0, 99999, 3)], [[i, i + 1] for i in range(500, 99999, 2)], "3x2"),
    ([[i, i + 1] for i in range(0, 99999)], [[i, i + 1] for i in range(1, 99999, 2)], "1x2"),
]


class IntervalTestCases:
    """Convert test_cases into bitintervals and bitmasks for later use."""

    def __init__(self, test_cases):
        self.test_cases = test_cases
        self.bitmask_cases = []
        self.bitmask_slots_cases = []
        for alst, blst, description in test_cases:
            bitmask_a, bitmask_b = convert_intervals_into_bitmask([alst, blst])
            self.bitmask_slots_cases.append((bitmask_a.to_slots(), bitmask_b.to_slots(), description))
            self.bitmask_cases.append((bitmask_a, bitmask_b, description))

    def get_bitmask_cases(self):
        return self.bitmask_cases

    def get_bitmask_slots_cases(self):
        return self.bitmask_slots_cases


def decorate_memory(func):
    @functools.wraps(func)
    def decorate(*args, **kwargs):
        gc.disable()
        m_lst, result = memory_profiler.memory_usage((func, args, kwargs), interval=0.05, retval=True)
        gc.enable()
        assert result is not None
        return m_lst

    return decorate


@cli.command()
@pass_settings
def intervals_performance(settings: Settings):
    """Test the performance of interval and bitmask operations with different test cases."""
    profile_memory = False  # Parameter controlling whether to have memory test
    bit_mask = True  # Parameter controlling whether testing with intervals or bitmask
    gc.disable()

    def lots_of_memory(a, b):
        x = [0] * 1000000
        y = [1] * 1000000
        time.sleep(0.1)
        return True

    """A lot of test functions"""
    funcs = [
        # intervals_xor_set_v0,
        # intervals_xor_v0,
        # intervals_xor_v4,
        # intervals_xor_v5,
        # intervals_xor_v6,
        # intervals_xor_v7,
        # logical_xnor_v0,
        # logical_xnor_v1,
        # logical_xnor_v2,
        # logical_xnor_v3,
        # logical_xnor_v4,
        # logical_xnor_v5,
        # logical_xnor_v6,
        # intervals_xor_faster,
        # logical_and_v0,
        # logical_and_v1,
        # logical_and_v2,
        # logical_and_v3,
        # logical_or_v0,
        # logical_or_v4,
        logical_or_v5,
        logical_or_v6,
        logical_or_v7,
        logical_or_v8,
        # intervals_and_fast,
        # intervals_subtract_fast4,
        # intervals_subtract_fast,
        # intervals_subtract_fast3
    ]
    # if profile_memory:
    #     funcs.append(lots_of_memory)

    count = 20
    results = []
    test_cases_instance = IntervalTestCases(interval_pair_input_deck)
    if bit_mask:
        itc_cases = test_cases_instance.get_bitmask_cases()
    else:
        itc_cases = test_cases_instance.get_bitmask_slots_cases()

    for bitmask_slots_a, bitmask_slots_b, speedcase in itc_cases:
        for func in funcs:
            warmup = True
            for i in range(count):
                ts = time.perf_counter()
                func(bitmask_slots_a, bitmask_slots_b)
                te = time.perf_counter()
                if profile_memory:
                    mfunc = decorate_memory(func)
                    m = mfunc(bitmask_slots_a, bitmask_slots_b)
                    memory = max(m) - min(m)
                else:
                    memory = 0
                if warmup:
                    warmup = False
                    continue
                results.append(
                    {
                        "speedcase": speedcase,
                        "func": func.__name__,
                        "time(s)": te - ts,
                        "mem(Mb)": memory,
                        "i": i,
                    }
                )
        gc.collect()

    df = pd.DataFrame(results)

    # Create a boxplot for the time
    plt.figure(figsize=(15, 8))
    sns.boxplot(x="speedcase", y="time(s)", hue="func", data=df)
    plt.xticks(rotation=45)
    plt.title("Execution Time by Function and Test Case")
    plt.tight_layout()
    plt.show()

    if profile_memory:
        # Create a boxplot for memory usage if profiling memory
        plt.figure(figsize=(15, 8))
        sns.boxplot(x="speedcase", y="mem(Mb)", hue="func", data=df)
        plt.xticks(rotation=45)
        plt.title("Memory Usage by Function and Test Case")
        plt.tight_layout()
        plt.show()


def generate_random_intervals(num_intervals, max_start=10000, max_length=100):
    """Generate a random interval list."""
    intervals = []
    for _ in range(num_intervals):
        start = random.randint(0, max_start)
        length = random.randint(1, max_length)
        end = start + length - 1
        intervals.append([start, end])
    return intervals


"""A big set of intervals list test sample. Running this test sample with 100K sizes would be very slow.
   We would need a function to control the test size in the future."""
intervals_list_input_deck = [
    ([[interval] for interval in generate_random_intervals(100000)], "random"),
    ([[[i, i + 1]] for i in range(0, 200000, 2)], "densely non-overlap"),
    ([[[i, i + 1000]] for i in range(0, 200000, 100)], "Partially overlap"),
    ([[[i, i + 1]] for i in range(0, 200000, 200)], "Sparsely non-overlap"),
    ([[[i, i + 10]] for i in range(0, 100000, 12)], "Small gaps"),
    ([[[i, i + 50]] if i % 100 == 0 else [[i, i + 1]] for i in range(0, 100000)], "Alternating Intervals"),
    ([[[i, i + 100000 - (j * 10000)]] for i, j in zip(range(0, 100000, 1000), range(10))], "Nested intervals"),
    ([[[i, i + random.randint(1, 100)]] for i in range(0, 100000, 500)], "Mixed size intervals"),
]


def convert_intervals_lst_to_bitmask_slots_lst(length, intervals_list):
    """Convert interval_list into bitmask slots."""
    lst = []
    for intervals in intervals_list:
        bit_count = sum(end - start + 1 for start, end in intervals),
        bit_idx_first = intervals[0][0],
        bit_idx_last = intervals[-1][1],
        next_intervals = LiteBitmaskSlots(
            length,
            bit_count,
            intervals,
            bit_idx_first,
            bit_idx_last,
        )
        lst.append(next_intervals)
    return lst


def convert_intervals_into_bitmask(intervals_list, length=1145141) -> List[LiteBitmask]:
    """Convert interval_list into bitmask_list."""
    bitmask_list = []
    for intervals in intervals_list:
        bit_count = sum(end - start + 1 for start, end in intervals)
        bit_idx_first = intervals[0][0] if intervals else -1
        bit_idx_last = intervals[-1][1] if intervals else -1
        bitmask = LiteBitmask(length, bit_count, intervals, bit_idx_first, bit_idx_last)
        bitmask_list.append(bitmask)
    return bitmask_list


def combine_intervals_sequentially(intervals_list):
    """Run the intervals_or_v7 a bunch of times to perform the interval combine."""
    if not intervals_list:
        combined_intervals = liteBitmaskSlotsZero
    else:
        length = intervals_list[0].length
        combined_intervals = LiteBitmaskSlots(length, 0, [], -1, -1)
        for intervals in intervals_list:
            combined_intervals = logical_or_v7(combined_intervals, intervals)
    return combined_intervals


class IntervalTestCases2:
    """Convert intervals list into bitintervals and bitmasks for later use."""

    def __init__(self, interval_test_cases):
        self.interval_test_cases = interval_test_cases
        self.bitmask_cases = []
        self.bitmask_slots_cases = []

        for interval_list, description in interval_test_cases:
            bitmask_lst = convert_intervals_into_bitmask(interval_list)
            bitmask_slots_lst = convert_intervals_lst_to_bitmask_slots_lst(2**20, interval_list)
            self.bitmask_cases.append((bitmask_lst, description))
            self.bitmask_slots_cases.append((bitmask_slots_lst, description))

    def get_bitmask_cases(self):
        return self.bitmask_cases

    def get_bitmask_slots_cases(self):
        return self.bitmask_slots_cases


@cli.command()
@pass_settings
def combined_intervals_performance_v0(settings: Settings):
    """Run the get_combined_intervals_until_all_consumed several times to test the performance.
    This test only test with one intervals_list, but it could test the correctness and print out overall running time."""
    intervals_lst_10000_raw = [[interval] for interval in generate_random_intervals(100)]
    bitmask_slots_lst_10000_nested = convert_intervals_lst_to_bitmask_slots_lst(2 ** 20, intervals_lst_10000_raw)

    bitmask = None
    combined_intervals_optimized = None
    combined_intervals_optimized_v1 = None
    combined_intervals_optimized_v2 = None
    sequential_time = 0
    optimized_time = 0
    optimized_time_v1 = 0
    optimized_time_v2 = 0
    bitmask_time = 0
    results = []
    count = 50

    for i in range(count):
        start_time = time.time()
        combined_intervals_sequential = combine_intervals_sequentially(bitmask_slots_lst_10000_nested)
        sequential_time = time.time() - start_time
        results.append({"iteration": i, "method": "combine_intervals_sequentially", "time(s)": sequential_time})

        start_time = time.time()
        combined_intervals_optimized = get_combined_intervals_until_all_consumed(bitmask_slots_lst_10000_nested)
        optimized_time = time.time() - start_time
        results.append({"iteration": i, "method": "get_combined_intervals_until_all_consumed", "time(s)": optimized_time})

        start_time = time.time()
        combined_intervals_optimized_v1 = get_combined_intervals_until_all_consumed_v1(bitmask_slots_lst_10000_nested)
        optimized_time_v1 = time.time() - start_time
        results.append({"iteration": i, "method": "get_combined_intervals_until_all_consumed_v1", "time(s)": optimized_time_v1})

        start_time = time.time()
        combined_intervals_optimized_v2 = get_combined_intervals_until_all_consumed_v2(bitmask_slots_lst_10000_nested)
        optimized_time_v2 = time.time() - start_time
        results.append({"iteration": i, "method": "get_combined_intervals_until_all_consumed_v2", "time(s)": optimized_time_v2})

        start_time = time.time()
        combined_intervals_optimized_v3 = get_combined_intervals_until_all_consumed_v3(bitmask_slots_lst_10000_nested)
        optimized_time_v3 = time.time() - start_time
        results.append({"iteration": i, "method": "get_combined_intervals_until_all_consumed_v2", "time(s)": optimized_time_v3})

        start_time = time.time()
        bitmask = (
            functools.reduce(lambda a, b: logical_or_v7(a, b), bitmask_slots_lst_10000_nested)
            if len(bitmask_slots_lst_10000_nested) > 1
            else (bitmask_slots_lst_10000_nested[0] if len(bitmask_slots_lst_10000_nested) == 1 else None)
        )
        bitmask_time = time.time() - start_time
        results.append({"iteration": i, "method": "bitmask_reduce", "time(s)": bitmask_time})

    if bitmask is not None and combined_intervals_optimized is not None and bitmask.intervals == combined_intervals_optimized.intervals:
        print("bitmask and combined_intervals_optimized have the same content.")
    else:
        print("bitmask and combined_intervals_optimized have different content.")

    if bitmask is not None and combined_intervals_optimized_v1 is not None and bitmask.intervals == combined_intervals_optimized_v1.intervals:
        print("bitmask and combined_intervals_optimized_v1 have the same content.")
    else:
        print("bitmask and combined_intervals_optimized_v1 have different content.")

    if bitmask is not None and combined_intervals_optimized_v2 is not None and bitmask.intervals == combined_intervals_optimized_v2.intervals:
        print("bitmask and combined_intervals_optimized_v2 have the same content.")
    else:
        print("bitmask and combined_intervals_optimized_v2 have different content.")

    if bitmask is not None and combined_intervals_optimized_v3 is not None and bitmask.intervals == combined_intervals_optimized_v3.intervals:
        print("bitmask and combined_intervals_optimized_v3 have the same content.")
    else:
        print("bitmask and combined_intervals_optimized_v3 have different content.")

    print("sequential_time:", sequential_time)
    print("optimized_time:", optimized_time)
    print("optimized_time_v1:", optimized_time_v1)
    print("optimized_time_v2:", optimized_time_v2)
    print("optimized_time_v2:", optimized_time_v3)
    print("bitmask time:", bitmask_time)

    df = pd.DataFrame(results)

    plt.figure(figsize=(15, 8))
    sns.violinplot(x="method", y="time(s)", data=df)
    plt.xticks(rotation=45)
    plt.title("Execution Time by Method for Combined Intervals")
    plt.tight_layout()
    plt.show()


def reduce_intervals(intervals_list):
    """Reduce the intervals using intervals_or_v7 to perform interval combine."""
    if len(intervals_list) > 1:
        return functools.reduce(lambda a, b: logical_or_v7(a, b), intervals_list)
    elif len(intervals_list) == 1:
        return intervals_list[0]
    else:
        return None


@cli.command()
@pass_settings
def combined_intervals_performance(settings: Settings):
    """Test various functions for combining intervals using multiple test cases."""
    # Define the functions to be tested
    funcs = [
        # combine_intervals_sequentially,  # Very slow on one
        # get_combined_intervals_until_all_consumed,
        get_combined_intervals_until_all_consumed_v1,
        get_combined_intervals_until_all_consumed_v2,
        get_combined_intervals_until_all_consumed_v3,
        # reduce_intervals  # Very SLOW!
    ]

    test_cases_instance = IntervalTestCases2(intervals_list_input_deck)

    bitmask_slots_cases = test_cases_instance.get_bitmask_slots_cases()

    count = 20
    results = []

    for bitmask_slots_list, description in bitmask_slots_cases:
        for func in funcs:
            warmup = True
            for i in range(count):
                start_time = time.time()
                combined_bitmask_slots = func(bitmask_slots_list)
                elapsed_time = time.time() - start_time
                if warmup:
                    warmup = False
                    continue
                results.append({
                    "description": description,
                    "func": func.__name__,
                    "time(s)": elapsed_time,
                    "iteration": i
                })

    df = pd.DataFrame(results)

    plt.figure(figsize=(15, 8))
    sns.boxplot(x="description", y="time(s)", hue="func", data=df)
    plt.xticks(rotation=45)
    plt.title("Execution Time by Function and Test Case for Combined Intervals")
    plt.tight_layout()
    plt.show()


@cli.command()
@pass_settings
def join_bitmasks_performance_v0(settings: Settings):
    """Run the join_bitmask process several times to test the performance.
    This test only test with one intervals_list, but it could test the correctness and print out overall running time."""
    intervals_list_10000_raw = [[interval] for interval in generate_random_intervals(10000)]
    bitmask_list_10000_nested = convert_intervals_into_bitmask(intervals_list_10000_raw)

    bitmask = None
    combined_bitmask_optimized1 = None
    combined_bitmask_optimized2 = None
    bitmask_time = 0
    optimized_time1 = 0
    optimized_time2 = 0
    results = []
    count = 500

    for i in range(count):
        start_time = time.time()
        bitmask = join_bitmasks_v0(bitmask_list_10000_nested, LiteBitmask.logical_or)
        bitmask_time = time.time() - start_time
        results.append({"iteration": i, "method": "join_bitmasks_v0", "time(s)": bitmask_time})

        start_time = time.time()
        combined_bitmask_optimized1 = join_bitmasks_v2(bitmask_list_10000_nested, LiteBitmask.logical_or)
        optimized_time1 = time.time() - start_time
        results.append({"iteration": i, "method": "join_bitmasks_v2", "time(s)": optimized_time1})

        start_time = time.time()
        combined_bitmask_optimized2 = join_bitmasks_v6(bitmask_list_10000_nested, LiteBitmask.logical_or)
        optimized_time2 = time.time() - start_time
        results.append({"iteration": i, "method": "join_bitmasks_v6", "time(s)": optimized_time2})

    if bitmask == combined_bitmask_optimized1:
        print("bitmask and combined_bitmask_optimized1 have the same content.")
    else:
        print("bitmask and combined_bitmask_optimized1 have different content.")

    if bitmask == combined_bitmask_optimized2:
        print("bitmask and combined_bitmask_optimized2 have the same content.")
    else:
        print("bitmask and combined_bitmask_optimized2 have different content.")

    print("bitmask time:", bitmask_time)
    print("optimized_time1:", optimized_time1)
    print("optimized_time2:", optimized_time2)

    df = pd.DataFrame(results)

    plt.figure(figsize=(15, 8))
    sns.violinplot(x="method", y="time(s)", data=df)
    plt.xticks(rotation=45)
    plt.title("Execution Time Distribution by Method for join_bitmasks")
    plt.tight_layout()
    plt.show()


@cli.command()
@pass_settings
def join_bitmasks_performance(settings: Settings):
    """Test various join_bitmasks functions using multiple test cases."""
    funcs = [
        join_bitmasks_v0,
        join_bitmasks_v2,
        join_bitmasks_v6,
    ]

    test_cases_instance = IntervalTestCases2(intervals_list_input_deck)
    bitmask_cases = test_cases_instance.get_bitmask_cases()
    count = 50
    results = []

    for bitmask_list, description in bitmask_cases:
        for func in funcs:
            warmup = True
            for i in range(count):
                start_time = time.time()
                combined_bitmask = func(bitmask_list, LiteBitmask.logical_or)
                elapsed_time = time.time() - start_time
                if warmup:
                    warmup = False
                    continue
                results.append({
                    "description": description,
                    "func": func.__name__,
                    "time(s)": elapsed_time,
                    "iteration": i
                })

    df = pd.DataFrame(results)

    plt.figure(figsize=(15, 8))
    sns.boxplot(x="description", y="time(s)", hue="func", data=df)
    plt.xticks(rotation=45)
    plt.title("Execution Time Distribution by Method for join_bitmasks")
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    cli()
