#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

""" example usage:
python tcut.py --range_start "2024-12-12 00:00:00" --range_end "2024-12-12 00:15:00"
--range_format "|%Y-%m-%dT%H:%M:%S.%fZ|" --range_length 26 -v data.csv"""

import os
import sys
import argparse
import datetime
from dateutil.parser import parse as dp
import logging

from Octeres.file_utils import read_cut, binary_search_file, read_cut_generator

parser = argparse.ArgumentParser(description="Search a log file for a section of time.")
parser.add_argument(
    "--range_format",
    dest="range_format",
    help=r"Range format Ex: 2018-03-05 15:26:21 -> %%Y-%%m-%%d %%H:%%M:%%S",
)
parser.add_argument("--range_start", dest="range_start", help="Start of the time cut.")
parser.add_argument("--range_end", dest="range_end", help="End of the time cut.")
parser.add_argument("--range_length", dest="range_length", help="length of the timestamp in the file.\n"
                                                                "If csv, include length with quoting.")
parser.add_argument("file_names", metavar="F", type=str, nargs="+", help="File Name ")
parser.add_argument("--verbose", "-v", action="count")


def time_binary_search(log, file_name, range_length, range_format, range_start, range_end):
    """entry point for the recursive binary search"""
    file_stat = os.stat(file_name)
    file_size = file_stat.st_size
    log.debug("searching file %s" % file_name)
    log.debug("st_size:%s" % file_size)
    log.debug("%s Megabytes" % (file_size / (1024 * 1024)))

    with open(file_name, "r") as fo:
        pos_start = binary_search_file(log, fo, range_length, range_format, range_start, 0, file_size)
        pos_end = binary_search_file(log, fo, range_length, range_format, range_end, pos_start, file_size)

    log.debug("start:%s, end:%s" % (pos_start, pos_end))
    return pos_start, pos_end


def main(args):
    logging.basicConfig()
    log = logging.getLogger("tcut")
    verbosity = args.verbose
    if verbosity is None:
        log.setLevel(logging.ERROR)
    if verbosity == 1:
        log.setLevel(logging.WARN)
    elif verbosity == 2:
        log.setLevel(logging.INFO)
    elif verbosity == 3:
        log.setLevel(logging.DEBUG)
    range_start = dp(args.range_start)
    range_end = dp(args.range_end)
    range_format = args.range_format
    range_length = int(args.range_length)
    file_names = args.file_names
    log.debug("Processing logs:\n")
    for file_name in file_names:
        log.debug("    %s" % file_name)
    log.debug("Cutting log from %s to %s" % (range_start, range_end))
    log.debug("Delta:%s hours" % ((range_end - range_start).total_seconds() / 3600.0))
    log.debug("Using format %s" % range_format)
    for file_name in file_names:
        pos_start, pos_end = time_binary_search(log, file_name, range_length, range_format, range_start, range_end)
        for line in read_cut_generator(file_name, pos_start, pos_end):
            print(line)
    log.debug("Cut complete.")


if __name__ == "__main__":
    args = parser.parse_args()
    main(args)
