# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""typesafe decorator for enforcing types based upon type hinting"""
import types
import inspect
from inspect import Signature

inspect_emtpy_type = Signature.empty
from functools import wraps
from collections import OrderedDict


def _isinstance_deep(obj, type_tup):
    """This will handle the recursive nature of complex types such
    as List[List[str]]"""
    # FIXME, implement
    raise NotImplemented("Not Implemented")


def get_func_typesafe_spec(func, remove_self=True):
    """assuming the function has already been decorated with
    typesafe, get the required types dictionary and return it"""
    try:
        # function
        required_types = func.required_types
    except AttributeError:
        try:
            # method
            required_types = func.__func__.required_types
        except AttributeError:
            # doesn't exist
            required_types = OrderedDict()
    if remove_self is True:
        result = [[key, value] for key, value in required_types.items() if key != "self"]
    else:
        result = [[key, value] for key, value in required_types.items()]
    return result


def set_types(func):
    """given a function set the variable "required_types"
    on the function.
    This will also include the return type as the key "return"
    The return is a OrderedDict()
    """
    sig = inspect.signature(func)
    odct = OrderedDict()
    sig_parameters = sig.parameters
    # let's be careful to loop in order
    for key in sig_parameters.keys():
        parameter = sig_parameters[key]
        name = parameter.name
        default = parameter.default
        annotation = parameter.annotation
        if default is not inspect_emtpy_type:  # FIXME: cannot have annotation and default..
            odct[key] = {
                "name": name,
                "required_type": type(default),
                "default": default,
                "param_type": "kwarg",
            }
        else:
            if annotation is not inspect_emtpy_type:
                odct[key] = {
                    "name": name,
                    "required_type": annotation,
                    "param_type": "arg",
                }
            else:
                # there is a linkage here to the main typesafe decorator
                odct[key] = {
                    "name": name,
                    "required_type": None,
                    "param_type": "arg",
                }  # <- key on this #1

    # go get the return code
    if sig.return_annotation is not inspect_emtpy_type:
        odct["return"] = {
            "name": "return",
            "required_type": sig.return_annotation,
            "param_type": "return",
        }
    else:
        odct["return"] = {
            "name": "return",
            "required_type": None,
            "param_type": "return",
        }  # again, <- key on this #2
    return odct


def typesafe(func):
    """type checking decorator for python 3.5"""
    sig = inspect.signature(func)
    # these are ordered
    # they also contain both args and kwargs
    parameter_keys = list(sig.parameters.keys())
    return_annotation = sig.return_annotation
    # set the types so we can check them
    required_types = set_types(func)
    try:
        # function
        func.required_types = required_types  # HACKz
    except AttributeError:
        # method
        func.__func__.required_types = required_types  # HACKz

    @wraps(func)
    def wrapper(*args, **kwargs):
        """decorator"""
        # reorder args, kwargs. kwargs could be out of order
        pos_values = OrderedDict()
        pos_keys = OrderedDict()
        if len(parameter_keys) < len(args):
            raise TypeError("Incorrect number of arguments. required(%s) != provided(%s)" % (len(parameter_keys), len(args)))
        for idx, value in enumerate(args):
            pos_values[idx] = value
            pos_keys[idx] = parameter_keys[idx]
        for key, value in kwargs.items():
            # go get the index for this key
            try:
                idx = parameter_keys.index(key)
            except ValueError:  # ValueError: 'xyz' is not in list
                raise TypeError("%s is not a valid kwarg." % key)
            else:
                pos_values[idx] = value
                pos_keys[idx] = key

        bad_args = []
        # now since we have order, we can check types
        for idx in pos_values.keys():
            value = pos_values[idx]
            key = pos_keys[idx]
            required_type = required_types[key]["required_type"]
            if required_type is not None:  # pivot here from above #1
                if not isinstance(value, required_type):
                    bad_args.append((idx, key, type(value), required_type, value))

        # throw an error.  We wanted to gather them all first
        if bad_args:
            error_lst = ["Function %s" % func.__name__]
            for params in bad_args:
                error_lst.append("    Argument %s:%s is type %s and should be type %s, value is %s" % (params))
            raise TypeError("\n".join(error_lst))

        # execute the function
        ret = func(*args, **kwargs)

        # now check the return code
        return_param = required_types["return"]
        if return_param["required_type"] is not None:  # pivot here #2
            if not isinstance(ret, return_param["required_type"]):
                bad_ret = "    Return is type %s, value: %s and should be type %s" % (
                    type(ret),
                    ret,
                    return_annotation,
                )
                raise TypeError(bad_ret)

        return ret

    return wrapper


def get_func_spec(func_or_method, remove_self=True):
    """given a function or func, return a dictionary with keys containing
    information about that function.

    func: name of the func
    doc: doc string of the func
    args: args
    kwargs: kwargs
    return: return
    """
    if isinstance(func_or_method, types.MethodType):
        func = func_or_method.__func__
    else:
        func = func_or_method

    dct = {}
    dct["function"] = func.__name__
    if func.__doc__ is None:
        raise Exception("All api functions must have doc strings. %s" % func)
    dct["doc"] = func.__doc__  # textwrap.dedent(func.__doc__).replace('\n',' ')
    dct["sig"] = "%s" % inspect.signature(func)
    try:
        required_types = func.required_types
    except AttributeError:
        print("Function is not typesafe")
        func = typesafe(func)
        required_types = func.required_types

    args = []
    kwargs = []
    returns = []

    for key, value in required_types.items():
        if remove_self is True and key == "self":
            continue
        if value.get("param_type", None) == "arg":
            args.append(value)
        elif value.get("param_type", None) == "kwarg":
            kwargs.append(value)
        elif value.get("param_type", None) == "return":
            returns.append(value)
        else:
            raise Exception("Unknown Parameter:%s:%s" % (key, value))

    dct["args"] = args
    dct["kwargs"] = kwargs
    if len(returns) == 1:
        dct["return"] = returns[0]
    else:
        raise Exception("More than one return! %s" % ("\n".join(returns)))
    return dct
