# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""
This module is a series of tests to check over and under flow.  Numpy will eventually
do this for us, but for now we have some tests.

These tests only work when __boxed_dtype__ == int8
"""

import sys
import warnings
from _decimal import Underflow, Overflow

import numpy as np
import pytest

from Octeres.bitmask import BitmaskSerializer
from Octeres.bitmask import EnhancedBitmask as eb
from Octeres.bitmask import check_eb_overunderflow, BASE_DTYPE
from Octeres.bitmask.bitmask_globals import __boxed_max_base_dtype_value__, __boxed_min_base_dtype_value__, __boxed_dtype__
from Octeres.util import superprint

bdtype = __boxed_dtype__
np.seterr(over="raise", under="raise")
np.set_printoptions(threshold=sys.maxsize, linewidth=132)
BitS = BitmaskSerializer()
warnings.filterwarnings("ignore", category=DeprecationWarning)


@pytest.mark.skipif(bdtype != np.int8, reason="boxed dtype is not np.int8")
def test_eb_view_01():
    eb.__flow_check__ = True
    try:
        e0 = eb.zeros(10)
        e0[1] = 1
        with pytest.raises(Underflow):
            e0[2] = -129
    finally:
        eb.__flow_check__ = False


@pytest.mark.skipif(bdtype != np.int8, reason="boxed dtype is not np.int8")
def test_eb_sign_01():
    eb.__flow_check__ = True
    try:
        e0 = eb.zeros(10)
        e0[1] = 1
        with pytest.raises(Overflow):
            e0[2] = 256
        e0s = eb.sign(e0)
        ncorr = np.array([0, 1, 0, 0, 0, 0, 0, 0, 0, 0], dtype=BASE_DTYPE)
        superprint(e0s)
        superprint(ncorr)
        assert (e0s == ncorr).all()

        n1 = np.zeros(10)
        n1[1] = 12
        n1[2] = -123
        with pytest.raises(Underflow):
            e1s = eb.sign(n1)
    finally:
        eb.__flow_check__ = False


@pytest.mark.skipif(bdtype != np.int8, reason="boxed dtype is not np.int8")
def test_eb_underflow_00():
    e2 = eb.zeros(10)
    e3 = eb.zeros(10)
    e2[1] = -10
    e3[1] = __boxed_max_base_dtype_value__
    with pytest.raises(Underflow):
        eb.subtract(e2, e3)


@pytest.mark.skipif(bdtype != np.int8, reason="boxed dtype is not np.int8")
def test_eb_overflow_00():
    lst = []
    lst.append((0, 0, True))
    lst.append((1, __boxed_max_base_dtype_value__ - 1, True))
    lst.append((100, 25, True))
    lst.append((100, 100, False))
    lst.append((100, 27, True))
    lst.append((100, 28, False))
    lst.append((__boxed_max_base_dtype_value__, 1, False))

    e2 = eb.zeros(10)
    e3 = eb.zeros(10)
    for a, b, good in lst:
        e2[1] = a
        e3[1] = b
        if good is True:
            result = eb.add(e2, e3)
        else:
            with pytest.raises(Overflow):
                eb.add(e2, e3)


@pytest.mark.skipif(bdtype != np.int8, reason="boxed dtype is not np.int8")
def test_eb___array_wrap__01():
    out_arr = np.array(2345, dtype=np.uint64)
    with pytest.raises(Overflow):
        obj = eb.array(out_arr)


@pytest.mark.skipif(bdtype != np.int8, reason="boxed dtype is not np.int8")
def test_overflow_00():
    bitmask = np.zeros(10, dtype=np.int64)
    bitmask[1] = 76543
    with pytest.raises(Overflow):
        check_eb_overunderflow(bitmask, debug=eb.__debug_flag__)


@pytest.mark.skipif(bdtype in (np.int8, np.int16), reason="boxed dtype is not np.int8")
def test_overflow_01():
    bitmask = np.zeros(10, dtype=BASE_DTYPE)
    bitmask[1] = 76543
    with pytest.raises(Overflow):
        check_eb_overunderflow(bitmask, debug=eb.__debug_flag__)


def test_overflow_02():
    bitmask = eb.zeros(10)
    with pytest.raises(Overflow):
        bitmask[1] = 76543


@pytest.mark.skipif(bdtype != np.int8, reason="boxed dtype is not np.int8")
def test_overflow_03():
    l0 = [12345, 0, -1234]
    with pytest.raises(Overflow):
        eb.array(l0)
    l0.reverse()
    with pytest.raises(Underflow):
        eb.array(l0)

    a0 = np.array([12345, 0, -1234])
    with pytest.raises(Overflow):
        eb.array(a0)


@pytest.mark.skipif(bdtype != np.int8, reason="boxed dtype is not np.int8")
def test_overflow_04():
    e0 = eb.ones(10)
    er = eb.zeros(10)
    superprint(e0)
    assertion_checked = False
    for x in range(__boxed_max_base_dtype_value__ + 1):
        # er += e0  # don't even know what this calls!
        # er = er + e0  # no overflow...
        if x == __boxed_max_base_dtype_value__:
            with pytest.raises(Overflow):
                assertion_checked = True
                er = eb.add(er, e0)
        else:
            er = eb.add(er, e0)
    assert assertion_checked
