# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import random
from functools import reduce
from itertools import chain

from Octeres.data_transform import *
from Octeres.data_generation import generate_random_dataframe
import pytest
from datetime import datetime
from typing import Callable
from contextlib import nullcontext as nc

fixtures_data = (
    pytest.param(
        ts_to_xminute,
        (2, datetime(2024, 1, 1, 23, 59, 59)),
        nc(enter_result=datetime(2024, 1, 1, 23, 58)),
        id="",
    ),
    pytest.param(
        ts_to_xsecond,
        (2, datetime(2024, 1, 1, 23, 59, 59)),
        nc(enter_result=datetime(2024, 1, 1, 23, 59, 58)),
        id="",
    ),
    pytest.param(
        ts_to_xsecond,
        (5, datetime(2024, 1, 1, 23, 59, 59)),
        nc(enter_result=datetime(2024, 1, 1, 23, 59, 55)),
        id="",
    ),
)


@pytest.mark.parametrize("method,args,expectation", fixtures_data)
def test_ts_to_x(method: Callable, args, expectation):
    with expectation:
        result = method(*args)
        assert result == expectation.enter_result


def test_split_join_00():
    columns = {
        "cola": {"a": ["cat", "dog", "cow"], "pfunc": "".join},
        "colb": {"a": tuple("abcdefghijklmnopqrstuvwxyz"), "size": 4, "pfunc": "".join},
    }
    df = generate_random_dataframe(columns, 117)
    # print(df.head())
    df_lst = df_split(df, chunk_size=13)
    df2 = df_join(df_lst)
    print()
    print(df.head())
    print(df2.head())
    assert len(df) == len(df2)
    for col in columns.keys():
        assert df[col].to_list() == df2[col].to_list()


def test_df_groupby_agg_00():
    row_count = 117000
    columns = {
        "dimension": {
            "a": ["cat", "dog", "cow"],
            "p": (0.1, 0.6, 0.3),
            "pfunc": "".join,
        },
        "value": lambda: random.random() * 10,
    }
    df = generate_random_dataframe(columns, row_count)
    func_int_flatten = lambda i: sum(i)
    func_str_flatten = lambda s: "".join(list(set(chain.from_iterable([s]))))
    df["ct"] = 0
    # FIXME: evaluate adding a dt for min and max agg_ts grouping.
    agg_dct = {"dimension": func_str_flatten, "value": func_int_flatten, "ct": "count"}
    dfg = df.groupby(["dimension"], as_index=False).agg(agg_dct)

    agg_c_dct = {
        "dimension": func_str_flatten,
        "value": func_int_flatten,
        "ct": "count",
    }
    agg_f_dct = {"dimension": func_str_flatten, "value": func_int_flatten, "ct": "sum"}
    dfg2 = df_groupby_agg(df, ["dimension"], agg_c_dct=agg_c_dct, agg_f_dct=agg_f_dct, chunk_size=16384)
    dfg3 = df_groupby_agg(df, ["dimension"], agg_c_dct=agg_c_dct, agg_f_dct=agg_f_dct, chunk_size=32)
    assert sum(dfg["ct"].to_list()) == row_count
    assert sum(dfg2["ct"].to_list()) == row_count
    assert sum(dfg3["ct"].to_list()) == row_count
    print(dfg.head())
    print(dfg2.head())


def test_df_groupby_agg_01():
    row_count = 117000
    columns = {
        "dimension": {
            "a": ["cat", "dog", "cow"],
            "p": (0.1, 0.6, 0.3),
            "pfunc": "".join,
        },
        "value": lambda: random.random() * 10,
    }
    df = generate_random_dataframe(columns, row_count)
    func_int_flatten = lambda i: sum(i)
    func_str_flatten = lambda s: "".join(list(set(chain.from_iterable([s]))))
    df["ct"] = 0
    agg_dct = {"dimension": func_str_flatten, "value": func_int_flatten, "ct": "count"}
    dfg = df.groupby(["dimension"], as_index=False).agg(agg_dct)
    agg_c_dct = {
        "dimension": func_str_flatten,
        "value": func_int_flatten,
        "ct": "count",
    }
    agg_f_dct = {"dimension": func_str_flatten, "value": func_int_flatten, "ct": "sum"}
    dfg3 = df_groupby_agg(df, ["dimension"], agg_c_dct=agg_c_dct, agg_f_dct=agg_f_dct, chunk_size=32)
    assert sum(dfg["ct"].to_list()) == row_count
    assert sum(dfg3["ct"].to_list()) == row_count
    print(dfg.head())

    import trace, sys

    tracer = trace.Trace(ignoredirs=[sys.prefix, sys.exec_prefix], trace=0, count=1)
    dfg3 = tracer.runfunc(
        df_groupby_agg,
        df,
        by=["dimension"],
        agg_c_dct=agg_c_dct,
        agg_f_dct=agg_f_dct,
        chunk_size=32,
    )
    tr = tracer.results()
    from pprint import pformat

    print(pformat(tr.counts))
