# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import datetime
from Octeres.SL.abstract_location_to_mask import get_bitmask_from_locations
from Octeres.SL.modules.ltm import get_Super_Location_to_Mask
from Octeres.bitmask import BitmaskSerializer
from Octeres.util import superprint

BitS = BitmaskSerializer()


def join_on_comma(names: list) -> str:
    return ",".join(names)


class Test_LTM_Salad:
    @classmethod
    def setup_class(cls):
        cls.machine_name = "salad"
        cls.sltm = get_Super_Location_to_Mask(filt={"name": cls.machine_name})

    @classmethod
    def teardown_class(cls):
        pass

    def test_00(self):
        range_start = datetime.datetime(2020, 5, 29)
        range_end = range_start + datetime.timedelta(days=1)
        ltm_obj = self.sltm.ltm_object_factory(self.machine_name, range_start, range_end=range_end)
        node_lstofdct = [(node, ltm_obj.graph.nodes[node]) for node in ltm_obj.graph.nodes]
        computes = [node for node, dct in node_lstofdct if dct["bitmask"].bit_count == 1]

    def test_01(self):
        range_start = datetime.datetime(2020, 7, 20)
        range_end = datetime.datetime(2020, 7, 21)
        ltm_obj = self.sltm.ltm_object_factory(self.machine_name, range_start, range_end=range_end)
        graph = ltm_obj.graph

        leaves = [n for n in graph.nodes() if graph.out_degree(n) == 0]
        assert len(leaves) == 192

        bitmask = get_bitmask_from_locations(ltm_obj, "gpu_name", ["salad01-gpu0", "salad01-gpu7"])
        assert bitmask.bit_count == 2

        bitmask = get_bitmask_from_locations(ltm_obj, "", ["salad01-gpu0", "salad01-gpu7"])
        assert bitmask.bit_count == 2

        bitmask = get_bitmask_from_locations(ltm_obj, "", ["salad01"])
        assert bitmask.bit_count == 8

        bitmask = get_bitmask_from_locations(ltm_obj, "", ["salad01", "salad01-gpu7"])
        assert bitmask.bit_count == 8

        bitmask = get_bitmask_from_locations(ltm_obj, "", ["salad01", "salad02-gpu7"])
        assert bitmask.bit_count == 9

        bitmask = get_bitmask_from_locations(ltm_obj, "", ["salad01", "salad02-gpu7", "salad01-gpu7"])
        assert bitmask.bit_count == 9

    def test_02(self):
        range_start = datetime.datetime(2020, 7, 20)
        range_end = datetime.datetime(2020, 7, 21)
        ltm_obj = self.sltm.ltm_object_factory(self.machine_name, range_start, range_end=range_end)

        location_type = "gpu"
        bitmask = get_bitmask_from_locations(ltm_obj, location_type, ["salad01-gpu0"])
        superprint(f"searching for bitmask {BitS.encode_bitmask_to_hex(bitmask)}")
        location_lst = ltm_obj.get_locations_from_bitmask(bitmask, location_type)
        assert location_lst == ["salad01-gpu0"]
        location_type = ""
        bitmask = get_bitmask_from_locations(ltm_obj, location_type, ["salad01"])
        assert bitmask.bit_count == 8
        superprint(f"searching for bitmask {BitS.encode_bitmask_to_hex(bitmask)}")
        location_lst = ltm_obj.get_locations_from_bitmask(bitmask, "node")
        location_lst.extend(ltm_obj.get_locations_from_bitmask(bitmask, "gpu"))
        assert location_lst == [
            "salad01",
            "salad01-gpu0",
            "salad01-gpu1",
            "salad01-gpu2",
            "salad01-gpu3",
            "salad01-gpu4",
            "salad01-gpu5",
            "salad01-gpu6",
            "salad01-gpu7",
        ]

        location_type = ""
        bitmask = ltm_obj.base_mask.copy()
        bitmask[0:8] = 1
        bitmask[15] = 1
        assert bitmask.bit_count == 9
        graph_location_lst = ltm_obj.get_locations_from_bitmask(bitmask, "node")
        graph_location_lst.extend(ltm_obj.get_locations_from_bitmask(bitmask, "gpu"))
        location_lst = [
            "salad01",
            "salad01-gpu0",
            "salad01-gpu1",
            "salad01-gpu2",
            "salad01-gpu3",
            "salad01-gpu4",
            "salad01-gpu5",
            "salad01-gpu6",
            "salad01-gpu7",
            "salad02-gpu7",
        ]
        assert location_lst == graph_location_lst
        assert join_on_comma(["salad01", "salad02-gpu7"]) == "salad01,salad02-gpu7"

    def test_03(self):
        range_start = datetime.datetime(2020, 7, 21)
        ltm_obj = self.sltm.ltm_object_factory(self.machine_name, range_start)
        # since salad01-gpu0 is not a slot_name, but a name, it will find it.
        location_type = "slot_name"
        bitmask = get_bitmask_from_locations(ltm_obj, location_type, ["salad01-gpu0"])
        location_lst = ltm_obj.get_locations_from_bitmask(bitmask, "gpu")
        location_lst.extend(ltm_obj.get_locations_from_bitmask(bitmask, "slot"))
        assert location_lst == [
            "salad01-gpu0",
            "n01g0s0",
            "n01g0s1",
            "n01g0s2",
            "n01g0s3",
            "n01g0s4",
            "n01g0s5",
            "n01g0s6",
        ]
        location_type = ""
        bitmask = get_bitmask_from_locations(ltm_obj, location_type, ["salad01"])
        assert bitmask.bit_count == 56
        bitmask = ltm_obj.base_mask.copy()
        bitmask[0:8] = 1
        bitmask[15] = 1
        assert bitmask.bit_count == 9
        graph_location_lst = ltm_obj.get_locations_from_bitmask(bitmask, "gpu")
        graph_location_lst.extend(ltm_obj.get_locations_from_bitmask(bitmask, "slot"))
        location_lst = [
            "salad01-gpu0",
            "n01g0s0",
            "n01g0s1",
            "n01g0s2",
            "n01g0s3",
            "n01g0s4",
            "n01g0s5",
            "n01g0s6",
            "n01g1s0",
            "n01g2s1",
        ]
        assert location_lst == graph_location_lst
        assert join_on_comma(["salad01", "salad02-gpu7"]) == "salad01,salad02-gpu7"

    def test_04(self):
        range_start = datetime.datetime(2020, 7, 20)
        range_end = datetime.datetime(2020, 7, 20) + datetime.timedelta(hours=12)
        ltm_obj = self.sltm.ltm_object_factory(self.machine_name, range_start, range_end=range_end)
        location_type = "gpu"
        bitmask = ltm_obj.base_mask.copy()
        bitmask[15] = 1
        location_lst = ltm_obj.get_locations_from_bitmask(bitmask, location_type)
        assert location_lst == ["salad02-gpu7"]

    def test_filter_node_names_00(self):
        range_start = datetime.datetime(2020, 7, 20)
        range_end = datetime.datetime(2020, 7, 20) + datetime.timedelta(hours=12)
        ltm_obj = self.sltm.ltm_object_factory(self.machine_name, range_start, range_end=range_end)
        nodes = [node for node in ltm_obj.graph.nodes if ltm_obj.graph.nodes[node]["is_schedulable"] == 1]
        correct_data = set()
        for node in ltm_obj.graph.nodes:
            is_schedulable = ltm_obj.get_is_schedulable(self.machine_name, node)
            if is_schedulable == 1:
                correct_data.add(node)
        assert set(nodes) == correct_data
