# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""
Unit tests for typesafe.
"""
import inspect
import sys
from _collections import OrderedDict
from functools import wraps
from pprint import pformat
from typing import Union
from contextlib import nullcontext

import pytest
from Octeres.typesafe import get_func_typesafe_spec
from Octeres.typesafe import typesafe
from Octeres.util import superprint


@typesafe
def func_untyped(a, b):
    """func_untyped"""
    return a + b


@typesafe
def func_typed(a: int, b: int) -> int:
    """func_typed"""
    return a + b


@typesafe
def func_partial(a: int, b):
    """func_partial"""
    return a + b


@typesafe
def func_typed_rt(a, b) -> int:
    """func_typed_rtonly"""
    return a + b


#
# return of List[int] doesn't work because of
# https://github.com/python/typing/commit/7baf6fe2a6d96ab09c807aac174650baa35d15dd
# https://github.com/python/typing/pull/283
#
@typesafe
def func_typed_complex(a: int, b: int) -> list:
    """func_typed_rt"""
    return [a, b]


@typesafe
def func_typed_kwargs(a: int, b=100) -> int:
    """func_typed_kwargs"""
    return a + b


@typesafe
def func_typed_useful_kwargs(a: int, *, b=100) -> int:
    """func_typed_kwargs"""
    return a + b


@pytest.mark.parametrize(
    "func",
    (
        pytest.param(func_untyped, id="untyped"),
        pytest.param(func_typed, id="typed"),
        pytest.param(func_partial, id="partial"),
        pytest.param(func_typed_rt, id="rt"),
        pytest.param(func_typed_complex, id="complex"),
        pytest.param(func_typed_kwargs, id="kwargs"),
        pytest.param(func_typed_useful_kwargs, id="useful_kwargs"),
    ),
)
def test_decorate(func):
    sig = inspect.signature(func)
    parameters = sig.parameters
    assert len(parameters) == 2


def test_check_required_types():
    correct = [
        ("a", {"param_type": "arg", "name": "a", "required_type": None}),
        ("b", {"param_type": "arg", "name": "b", "required_type": None}),
        ("return", {"param_type": "return", "name": "return", "required_type": None}),
    ]
    correct = OrderedDict(correct)
    assert func_untyped.required_types == correct


@pytest.mark.parametrize(
    "func,args,expectation",
    (
        pytest.param(func_untyped, ("1", "2"), nullcontext(enter_result="12"), id="func_untyped"),
        pytest.param(func_typed, ("1", "2"), pytest.raises(TypeError), id="func_typed"),
        pytest.param(func_partial, (1, "2"), pytest.raises(TypeError), id="func_partial"),
        pytest.param(func_typed_rt, (1, 2), nullcontext(enter_result=3), id="func_typed_rt"),
        pytest.param(
            func_typed_complex,
            (1, 2),
            nullcontext(enter_result=[1, 2]),
            id="func_typed_complex",
        ),
    ),
)
def test_args(func, args, expectation):
    with expectation:
        result = func(*args)
        correct = expectation.enter_result
        assert result == correct, "result:%s correct:%s" % (result, correct)


@pytest.mark.parametrize(
    "func,args,kwargs,expectation",
    (
        pytest.param(
            func_typed_kwargs,
            (1, "3"),
            dict(),
            pytest.raises(TypeError),
            id="func_typed_kwargs",
        ),
        pytest.param(
            func_typed_kwargs,
            (1,),
            dict(b="3"),
            pytest.raises(TypeError),
            id="func_typed_kwargs",
        ),
        pytest.param(func_typed_kwargs, (), dict(a=1), nullcontext(101), id="func_typed_kwargs"),
        pytest.param(func_typed_kwargs, (1,), dict(), nullcontext(101), id="func_typed_kwargs"),
        pytest.param(func_typed_kwargs, (1, 3), dict(), nullcontext(4), id="func_typed_kwargs"),
        pytest.param(
            func_typed_kwargs,
            (1, "3"),
            dict(),
            pytest.raises(TypeError),
            id="func_typed_kwargs",
        ),
    ),
)
def test_kwargs(func, args, kwargs, expectation):
    with expectation:
        result = func(*args, **kwargs)
        correct = expectation.enter_result
        assert result == correct, "result:%s correct:%s" % (result, correct)


def test_kwargs_star():
    with pytest.raises(TypeError):
        result = func_typed_useful_kwargs(1, 2)  # will not work with * in between

    result = func_typed_useful_kwargs(1, b=2)  # will not work with * in between
    correct = 3
    assert result == correct, "result:%s correct:%s" % (result, correct)

    result = func_typed_useful_kwargs(1)  # will not work with * in between
    correct = 101
    assert result == correct, "result:%s correct:%s" % (result, correct)


def test_kwargs_01():
    def func(a: int, b: str, c={}) -> tuple:
        """func_typed_kwargs"""
        return c, a, b

    func = typesafe(func)

    a = 1
    b = "string"
    c = {"filters": [{"name": "location", "val": "%5%", "op": "like"}]}
    result_tup = func(a, b, c=c)
    superprint(pformat(result_tup))
    assert result_tup == (c, a, b)


def test_bad_function():
    def func(a: int, b: int) -> int:
        """this function returns a string when it should be an int"""
        return "%s%s" % (a, b)  # type: ignore

    func = typesafe(func)
    with pytest.raises(TypeError):
        func(1, 2)


def test_generic_func():
    def func(a: int, b: int) -> Union[object]:
        return "%s%s" % (a, b)

    func = typesafe(func)
    result = func(1, 2)
    assert result == "12"

    def func2(a: int, b: int) -> Union[object]:
        return None

    func = typesafe(func)
    result = func2(1, 2)
    assert result == None

    def func3(a: int, b: int) -> Union[str, int]:
        return None

    func = typesafe(func)
    result = func3(1, 2)
    assert result == None

    def func4(a: Union[str, int], b: int) -> Union[str, int]:
        return None

    func = typesafe(func)
    result = func4("1", 2)
    assert result == None


# def test_generic_subscripted_func_00():
#     """
#     PROBLEMS WITH PYTHON 3.5, 3.6, 3.7
#
#     Type checking will work, but then it must serialize List[str] which is a _GenericAlias
#     Checking the type of _GenericAlias is difficult as it's different in Python 3.5, 3.6, 3.7
#
#     TypeError: Object of type _GenericAlias is not JSON serializable
#
#     pydantic.error_wrappers.ValidationError: 1 validation error for GetMagnitudeTimeline
#     range_start
#     invalid datetime format (type=value_error.datetime)
#     :return:
#     """
#     def func(a: Union[str, int], b: int) -> Union[object]:
#         return "%s%s" % (a, b)
#
#     func = typesafe(func)
#     result = func(1, 2)
#     self.assertEqual(result, "12")

# def test_generic_subscripted_func_01():
#     """
#     PROBLEMS WITH PYTHON 3.5, 3.6, 3.7
#
#     Type checking will work, but then it must serialize List[str] which is a _GenericAlias
#     Checking the type of _GenericAlias is difficult as it's different in Python 3.5, 3.6, 3.7
#
#     TypeError: Object of type _GenericAlias is not JSON serializable
#
#     pydantic.error_wrappers.ValidationError: 1 validation error for GetMagnitudeTimeline
#     range_start
#     invalid datetime format (type=value_error.datetime)
#     :return:
#     """
#     def func(a: Optional[str], b: int) -> Union[object]:
#         return "%s%s" % (a, b)
#
#     func = typesafe(func)
#     result = func(1, 2)
#     self.assertEqual(result, "12")


def test_multicall():
    def func(a: int, b: int, c=10, d="2", e=[]) -> tuple:
        """this function returns the signature but
        we will call it in different orders and with
        a mix of args to kwargs"""
        return (a, b, c, d, e)

    func = typesafe(func)
    result = func(1, 2, d="3", c=9)
    assert result == (1, 2, 9, "3", [])

    result = func(1, 2, 3, e=[1, 2], d="9")
    assert result == (1, 2, 3, "9", [1, 2])

    with pytest.raises(TypeError):
        func(1, 2, "4")


def test_decorate2():
    def func(a: int, b: int, c=10, d="2", e=[]) -> tuple:
        """this function returns the signature but
        we will call it in different orders and with
        a mix of args to kwargs"""
        return (a, b, c, d, e)

    def funcwrap(func):
        @wraps(func)
        def decorate(*args, **kwargs):
            return func(*args, **kwargs)

        return func

    def expose(func):
        """sets a variable, "exposed" on the object to True"""
        func.exposed = True
        return func

    func = funcwrap(func)
    func = typesafe(func)
    func = expose(func)
    print(func.required_types)
    sig = inspect.signature(func)

    correct = [
        ("a", {"param_type": "arg", "required_type": int, "name": "a"}),
        ("b", {"param_type": "arg", "required_type": int, "name": "b"}),
        (
            "c",
            {"param_type": "kwarg", "required_type": int, "default": 10, "name": "c"},
        ),
        (
            "d",
            {"param_type": "kwarg", "required_type": str, "default": "2", "name": "d"},
        ),
        (
            "e",
            {"param_type": "kwarg", "required_type": list, "default": [], "name": "e"},
        ),
        ("return", {"param_type": "return", "required_type": tuple, "name": "return"}),
    ]
    correct = OrderedDict(correct)

    if sys.version.startswith("3.5") or sys.version.startswith("3.6"):
        assert "%s" % sig == "(a:int, b:int, c=10, d='2', e=[]) -> tuple"
    else:
        assert "%s" % sig == "(a: int, b: int, c=10, d='2', e=[]) -> tuple"
    assert func.required_types == correct


def test_get_func_typesafe_spec_00():
    def func(a: int, b: int, c=10, d="2", e=[]) -> tuple:
        """this function returns the signature but
        we will call it in different orders and with
        a mix of args to kwargs"""
        return (a, b, c, d, e)

    spec = get_func_typesafe_spec(func, remove_self=False)
    assert spec == []
    spec = get_func_typesafe_spec(func, remove_self=True)
    assert spec == []

    func = typesafe(func)
    spec = get_func_typesafe_spec(func, remove_self=False)
    correct_spec = [
        ["a", {"param_type": "arg", "name": "a", "required_type": int}],
        ["b", {"param_type": "arg", "name": "b", "required_type": int}],
        [
            "c",
            {"default": 10, "name": "c", "required_type": int, "param_type": "kwarg"},
        ],
        [
            "d",
            {"default": "2", "name": "d", "required_type": str, "param_type": "kwarg"},
        ],
        [
            "e",
            {"default": [], "name": "e", "required_type": list, "param_type": "kwarg"},
        ],
        ["return", {"param_type": "return", "name": "return", "required_type": tuple}],
    ]
    assert spec == correct_spec

    correct_spec = [
        ["a", {"param_type": "arg", "name": "a", "required_type": int}],
        ["b", {"param_type": "arg", "name": "b", "required_type": int}],
        [
            "c",
            {"default": 10, "name": "c", "required_type": int, "param_type": "kwarg"},
        ],
        [
            "d",
            {"default": "2", "name": "d", "required_type": str, "param_type": "kwarg"},
        ],
        [
            "e",
            {"default": [], "name": "e", "required_type": list, "param_type": "kwarg"},
        ],
        ["return", {"param_type": "return", "name": "return", "required_type": tuple}],
    ]
    spec = get_func_typesafe_spec(func, remove_self=True)
    assert spec == correct_spec


def test_get_func_typesafe_spec_01():
    class Container:
        def func_plain(self, a: int, b: int, c=10, d="2", e=[]) -> tuple:
            """this function returns the signature but
            we will call it in different orders and with
            a mix of args to kwargs"""
            return (a, b, c, d, e)

        @typesafe
        def func(self, a: int, b: int, c=10, d="2", e=[]) -> tuple:
            """this function returns the signature but
            we will call it in different orders and with
            a mix of args to kwargs"""
            return (a, b, c, d, e)

    container = Container()
    spec = get_func_typesafe_spec(container.func_plain, remove_self=False)
    assert spec == []
    spec = get_func_typesafe_spec(container.func_plain, remove_self=True)
    assert spec == []

    spec = get_func_typesafe_spec(container.func, remove_self=False)
    correct_spec = [
        ["self", {"required_type": None, "param_type": "arg", "name": "self"}],
        ["a", {"required_type": int, "param_type": "arg", "name": "a"}],
        ["b", {"required_type": int, "param_type": "arg", "name": "b"}],
        [
            "c",
            {"param_type": "kwarg", "required_type": int, "default": 10, "name": "c"},
        ],
        [
            "d",
            {"param_type": "kwarg", "required_type": str, "default": "2", "name": "d"},
        ],
        [
            "e",
            {"param_type": "kwarg", "required_type": list, "default": [], "name": "e"},
        ],
        ["return", {"required_type": tuple, "param_type": "return", "name": "return"}],
    ]
    assert spec == correct_spec
    spec = get_func_typesafe_spec(container.func, remove_self=True)
    correct_spec = [
        ["a", {"required_type": int, "param_type": "arg", "name": "a"}],
        ["b", {"required_type": int, "param_type": "arg", "name": "b"}],
        [
            "c",
            {"param_type": "kwarg", "required_type": int, "default": 10, "name": "c"},
        ],
        [
            "d",
            {"param_type": "kwarg", "required_type": str, "default": "2", "name": "d"},
        ],
        [
            "e",
            {"param_type": "kwarg", "required_type": list, "default": [], "name": "e"},
        ],
        ["return", {"required_type": tuple, "param_type": "return", "name": "return"}],
    ]
    assert spec == correct_spec


def test_get_func_typesafe_spec_02():
    # this will require more to look at.  The difference between the decorator
    # on the method or after is quite different.
    class Container:
        def func_plain(self, a: int, b: int, c=10, d="2", e=[]) -> tuple:
            """this function returns the signature but
            we will call it in different orders and with
            a mix of args to kwargs"""
            return (a, b, c, d, e)

        @typesafe
        def func(self, a: int, b: int, c=10, d="2", e=[]) -> tuple:
            """this function returns the signature but
            we will call it in different orders and with
            a mix of args to kwargs"""
            return (a, b, c, d, e)

    container = Container()
    container.func_plain = typesafe(container.func_plain)
    # requires func.__func__.required_types = required_types
    # you cannot assign attributes to methods
    correct = [
        ("a", {"name": "a", "param_type": "arg", "required_type": int}),
        ("b", {"name": "b", "param_type": "arg", "required_type": int}),
        (
            "c",
            {"default": 10, "name": "c", "param_type": "kwarg", "required_type": int},
        ),
        (
            "d",
            {"default": "2", "name": "d", "param_type": "kwarg", "required_type": str},
        ),
        (
            "e",
            {"default": [], "name": "e", "param_type": "kwarg", "required_type": list},
        ),
        ("return", {"name": "return", "param_type": "return", "required_type": tuple}),
    ]
    correct = OrderedDict(correct)
    assert container.func_plain.required_types == correct

    correct = [
        ("self", {"required_type": None, "param_type": "arg", "name": "self"}),
        ("a", {"required_type": int, "param_type": "arg", "name": "a"}),
        ("b", {"required_type": int, "param_type": "arg", "name": "b"}),
        (
            "c",
            {"required_type": int, "default": 10, "param_type": "kwarg", "name": "c"},
        ),
        (
            "d",
            {"required_type": str, "default": "2", "param_type": "kwarg", "name": "d"},
        ),
        (
            "e",
            {"required_type": list, "default": [], "param_type": "kwarg", "name": "e"},
        ),
        ("return", {"required_type": tuple, "param_type": "return", "name": "return"}),
    ]
    correct = OrderedDict(correct)
    assert container.func.required_types == correct


# FIXME List[str]
# def test_complex_type():
#     def func() -> List[str]:
#         return ['a','b','c']
#
#     def func_fail_00() -> List[str]:
#         return [1,2,3]
#
#     def func_fail_01() -> List[str]:
#         return [1,2,'a']
#
#     func = typesafe(func)
#     func_fail_00 = typesafe(func_fail_00)
#     func_fail_01 = typesafe(func_fail_01)
#
#     print(func_fail_00.required_types)
#
#     correct = ['a','b','c']
#     self.assertEqual(func(), correct)
#
#     with self.assertRaises(TypeError):
#         func_fail_00()
#
#     with self.assertRaises(TypeError):
#         func_fail_01()
